//Detecta circulos e os devolve num vetor, junto com valor acumulado.
//Devolve circulos com acumulador<limiar ou max_circulos se houver muitos circulos
#include <cekeikon.h>

void subtrai(Mat_<FLT>& b, int l, int c, int borraoespacial, float magn) {
  for (int l2=l-borraoespacial; l2<=l+borraoespacial; l2++)
    for (int c2=c-borraoespacial; c2<=c+borraoespacial; c2++)
      if (0<=l2 && l2<b.rows && 0<=c2 && c2<b.cols)
        b(l2,c2) -= magn;
}

void houghCir(Mat_<COR> _ent,
              vector<Vec3i>& circulos, // l, c, raio
              vector<float>& acumulado,
              float limiar, int max_circulos, // Devolve circulos com acumulador<limiar ou
                                   // max_circulos se houver muitos circulos.
              double min_dist,
              int rmin,
              int rmax,
              int borraoespacial=1,
              double GaussDev=2.0,
              double pesormax=1.0,
              double pesormin=1.0) {
  Mat_<COR> ent=_ent.clone();
  int nl=ent.rows;
  int nc=ent.cols;
  int ns=rmax-rmin+1;

  vector< Mat_<FLT> > acc(ns);
  for (int i=0; i<ns; i++) {
    acc[i].create(nl,nc);
    acc[i].setTo(0.0); //comecou com zero
  }

  GaussianBlur(ent,ent,Size(0,0),GaussDev,GaussDev);

  Mat_<CPX> grad=gradienteScharr(ent,true); // y para cima
  Mat_<FLT> magn=modulo(grad);
  Mat_<CPX> vers=versor(grad);

  for (int l=0; l<ent.rows; l++) {
    for (int c=0; c<ent.cols; c++)
      if (magn(l,c)>0.001) {
        // Traca reta em acc, a partir do centro (l,c).
        for (unsigned s=0; s<acc.size(); s++) {
          int r=s+rmin;
          int l2=round(l-imag(vers(l,c))*r);
          int c2=round(c+real(vers(l,c))*r);
          subtrai(acc[s],l2,c2,borraoespacial,magn(l,c));
          l2=round(l+imag(vers(l,c))*r);
          c2=round(c-real(vers(l,c))*r);
          subtrai(acc[s],l2,c2,borraoespacial,magn(l,c));
        }
      }
  }

  // Para cada pixel, acha a escala (raio) que tem o menor valor no acc.
  // Preenche minimo e escala
  Mat_<FLT> menoracc(ent.size());
  Mat_<int> menorescala(ent.size());
  for (unsigned i=0; i<ent.total(); i++) {
    float minval=0; int minind=0;
    for (int s=0; s<ns; s++)
      if (acc[s](i)<minval) { minval=acc[s](i); minind=s; }
    menoracc(i)=minval;
    menorescala(i)=minind;
  }

  //Pode dar mais peso para raios menores ou maiores
  if (pesormax!=1.0 || pesormin!=1.0)
    for (unsigned i=0; i<ent.total(); i++) {
      double peso = pesormin + ( (pesormax-pesormin) * (ns-1-menorescala(i)) / (ns-1) );
      // ns-1              pesormax
      // menorescala       peso
      // 0                 pesormin
      //
      // ns-1              pesormax-pesormin
      // ns-1-menorescala  peso-pesormin
      //
      // (peso-pesormin)*(ns-1) = (pesormax-pesormin)*(ns-1-menorescala)
      //
      // peso = (pesormax-pesormin)*(ns-1-menorescala)/(ns-1) + pesormin
      menoracc(i)*=peso;
    }

  // Normaliza menoracc para 0..1
  Mat_<FLT> menoraccnormal=normaliza(menoracc);

  // Chama vector<Point> kmin(Mat_<FLT> a, int k, double d=0.0) para achar os k minimos.
  // Os pontos de a com valor 1 nao sao levados em conta.
  vector<Point> centros=kmin(menoraccnormal,max_circulos,min_dist);

  for (int i=0; i<int(centros.size()); i++) {
    Point c=centros[i];
    float a=menoracc(c);
    if (menoracc(c)<limiar) {
      int r=menorescala(c)+rmin;
      Vec3f v(c.x,c.y,r); circulos.push_back(v);
      acumulado.push_back(a);
    }
  }
}

void houcirc(int argc, char** argv) {
  const char* keys =
    " curto| longo    |default| explicacao\n"
    " -l   | -limiar  | -15   | Devolve circulos com acc<limiar\n"
    " -c   | -maxcirc | 20    | Numero maximo de circulos\n"
    " -d   | -mindist | 20    | Distancia minima entre dois circulos\n"
    " -min | -rmin    | 20    | Raio minimo do circulo\n"
    " -max | -rmax    | 80    | Raio maximo do circulo\n"
    " -b   | -borrao  |  1    | Borrao espacial no espaco de Hough\n"
    " -g   | -gaussd  |  2    | Desvio padrao do filtro gaussiano\n"
    " -pmin| -pmin    |  1    | Peso do circulo de raio minimo\n"
    " -pmax| -pmax    |  1    | Peso do circulo de raio maximo\n"
    " -nl  | -nlinhas |  0    | Redimensiona para nl. 0=nao redimensiona\n"
    " -nc  | -ncolunas|  0    | Redimensiona para nc. 0=nao redimensiona\n";

  if (argc<3) {
    printf("HouCirC ent.ppm sai.ppm [opcoes]\n");
    printf("%s",keys);
    erro("Erro: Numero de argumentos invalido");
  }

  ArgComando cmd(argc,argv);
  string nomeent=cmd.getCommand(0);
  if (nomeent=="") erro("Erro: Nao especificou imagem de entrada");
  string nomesai=cmd.getCommand(1);
  if (nomesai=="") erro("Erro: Nao especificou imagem de saida");

  float limiar=cmd.getFloat("-l","-limiar",-15);
  int max_circulos=cmd.getInt("-c","-maxcirc",20);
  double min_dist=cmd.getDouble("-d","-mindist",20);
  int rmin=cmd.getInt("-min","-rmin",20);
  int rmax=cmd.getInt("-max","-rmax",80);
  int borrao=cmd.getInt("-b","-borrao",1);
  double gaussd=cmd.getDouble("-g","-gaussd",2);
  double pesormin=cmd.getDouble("-pmin","-pmin",1);
  double pesormax=cmd.getDouble("-pmax","-pmax",1);
  int nl=cmd.getInt("-nl","-nlinhas",0);
  int nc=cmd.getInt("-nc","-ncolunas",0);
  cmd.leuTodos();

  Mat_<COR> ent; le(ent,nomeent);
  if (nl>0 || nc>0)
    resize(ent,ent,Size(nc,nl),INTER_AREA);

  vector<Vec3i> circulos; // x, y, raio
  vector<float> acumulado;
  houghCir(ent, circulos, acumulado,
           limiar, max_circulos, min_dist, rmin, rmax,
           borrao, gaussd, pesormin, pesormax);

  for (int i=0; i<int(circulos.size()); i++) {
    Point center(circulos[i][0], circulos[i][1]);
    int radius = circulos[i][2];
    // draw the circle center
    circle( ent, center, 3, Scalar(0,255,0), -1, 8, 0 );
    // draw the circle outline
    circle( ent, center, radius, Scalar(0,0,255), 3, 8, 0 );
  }
  imp(ent,nomesai);
}
