// CLAHEC = CLAHE para imagens coloridas
// Linke com OpenCV3
// compila clahec -c -v3
#include <cekeikon.h>

int main(int argc, char** argv) {
  if (argc<3) {
    printf("clahec: Contrast Limited Adaptive Histogram Equalization (COR)\n");
    printf("clahec ent.ppm sai.ppm [opcoes]\n");
    printf("%s",
      " curto| longo    |default| explicacao\n"
      " -c   | -clip    | 40    | Clip histogram limit\n"
      " -t   | -tile    |  8    | Use 8x8 tile grid size\n"
      " -r   | -rgb     | false | Equaliza cada banda RGB\n"
      "   Se nao especificar -r, equaliza L de CieLab\n");
    erro("Erro: Numero de argumentos invalido");
  }
  ArgComando cmd(argc,argv);
  string nomeent=cmd.getCommand(0);
  string nomesai=cmd.getCommand(1);
  double clip=cmd.getDouble("-c","-clip",40.0);
  int tile=cmd.getInt("-t","-tile",8);
  bool rgb=cmd.getBool("-r","-rgb",false);
  cmd.leuTodos();

  Mat_<COR> ent; le(ent,nomeent);
  Mat_<COR> sai;

  if (rgb==false) { // Processa em CieLab  
    Mat_<COR> lab; cvtColor(ent,lab,CV_BGR2Lab);
    vector< Mat_<GRY> > lab_planes(3);
    split(lab, lab_planes);  // now we have the L image in lab_planes[0]
  
    Ptr<CLAHE> clahe = createCLAHE();
    clahe->setClipLimit(clip);
    clahe->setTilesGridSize(Size(tile,tile));
    
    clahe->apply(lab_planes[0],lab_planes[0]);
    merge(lab_planes,lab);
  
    cvtColor(lab,sai,CV_Lab2BGR);
  } else { // Processa cada banda
    vector< Mat_<GRY> > bandas(3);
    split(ent, bandas);  // Aqui, temos BGR em bandas[0], bandas[1] e bandas[2]
  
    Ptr<CLAHE> clahe = createCLAHE();
    clahe->setClipLimit(clip);
    clahe->setTilesGridSize(Size(tile,tile));
    
    clahe->apply(bandas[0],bandas[0]);
    clahe->apply(bandas[1],bandas[1]);
    clahe->apply(bandas[2],bandas[2]);
    
    merge(bandas,sai);
  }  
  imp(sai,nomesai);
}
