#include <cekeikon.h>

/*
//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< RIHOG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class RIHOG {
 public:
  Mat_<Vec2f> kernel; // angulo, anel (-1 indica fora de circulo ou no centro)
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 5
  int raio; // raio = naneis*larguraAnel = 40
  double raioInic; // raioInic = raio / sqrt(naneis)
  vector<double> raios; // Onde termina cada anel
  int diametro; // diametro = 2*raio+1 = 81
  int nbins; // 20
  int ngauss; // 37
  int descriptorSize; // descriptorSize=nbins*naneis=100
  int tnl; // tnl=diametro+2*(ngauss/2)
  double tuso; // tuso=double(diametro)/double(tnl)
  bool bcinv; // brightness-contrast invariant
  double denominador;
  double denominador2;

  void inicializa(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true);
  RIHOG(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true) { inicializa(naneisp,raiop,nbinsp,ngaussp,bcinvp); }
  int quantiza(double d);
  void rihog(Mat_<GRY> a, M3d_<FLT>& features);

  //void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  //void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void RIHOG::inicializa(int naneisp, int raiop, int nbinsp, int ngaussp, bool bcinvp)
{ naneis=naneisp;
  raio=raiop;
  raioInic = double(raio) / sqrt(double(naneis));
  raios.resize(naneis);
  for (int i=0; i<naneis; i++)
    raios[i] = sqrt(double(i)) * raioInic;
  diametro=2*raio+1;
  nbins=nbinsp;
  ngauss=ngaussp;
  descriptorSize=nbins*naneis;
  tnl=diametro+2*(ngauss/2);
  tuso=double(diametro)/double(tnl);
  bcinv=bcinvp;
  denominador=360.0/nbins;
  denominador2=denominador/2;

  kernel.create(tnl,tnl);
  IMG_<Vec2f> kernelp(kernel);
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec2f& p=kernelp.atC(x,y);

      double r=sqrt(double(x*x+y*y));
      int q=0; 
      for (int i=0; i<naneis; i++)
        if (r<raios[i]) { q=i; break; }
      if (q>=naneis || r==0.0) { 
        p[0]=0; p[1]=-1; continue; 
      } else { 
        p[0]=fastAtan2(y,x); // entre 0 e 360
        p[1]=q; 
      }
    }
  
  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[1]>=0) 
        conta[kernel(l,c)[1]]++;
    }
}

int RIHOG::quantiza(double d)
// 0<=d<360
{ 
  //if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-1 fora de 0 a 360");
  d+=denominador2;
  if (d>=360) d-=360; // 0<=d<360
  //if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-2 fora de 0 a 360");
  int q=int(d/denominador);      
  //if (!(0<=q && q<nbins)) erro("Erro RIHOG::quantiza inesperado"); 
  return q;  
}

void RIHOG::rihog(Mat_<GRY> _a, M3d_<FLT>& features)
{ if (_a.rows<kernel.rows || _a.cols<kernel.cols) erro("Erro RIHOG::rihog tamanho imagem");

  Mat_<GRY> a;
  GaussianBlur(_a, a, Size(ngauss,ngauss), 0, 0, BORDER_REPLICATE);

  vector< Mat_<SHT> > g(2); // gradiente 2 direcoes: ->  ^
  Mat_<FLT> k0(1,3, (FLT[]){-1, 0, 1});
  filter2D(a, g[0], CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  Mat_<FLT> k1(3,1, (FLT[]){1, 0, -1});
  filter2D(a, g[1], CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 

  Mat_<FLT> mag(a.size());
  Mat_<FLT> deg(a.size());
  for (int l=0; l<a.rows; l++)
    for (int c=0; c<a.cols; c++) {
      float g0=g[0](l,c);
      float g1=g[1](l,c);
      mag(l,c)=sqrt(g1*g1+g0*g0);
      deg(l,c)=fastAtan2(g1,g0); // fastAtan2(y,x)
    }

  features.create(3,(int[]){a.rows,a.cols,descriptorSize});
  features.setTo(0.0);

  Mat_<FLT> f(naneis,nbins,0.0);
  for (int l=raio; l<a.rows-raio; l++) {
    printf("Linha %d\r",l);
    for (int c=raio; c<a.cols-raio; c++) {
      for (int l2=-raio; l2<=raio; l2++)
        for (int c2=-raio; c2<=raio; c2++) {
          Vec2f k=kernel(l2+raio,c2+raio); // angulo, anel
          if (k[1]==-1) continue; // centro ou fora do maior raio
          //if (k[0]<0 || k[0]>=360) erro("Erro k[0]=angulo fora da faixa"); // Pode tirar, apos debugar
          //if (k[1]<0 || k[1]>=naneis) erro("Erro k[1]=aneis fora da faixa"); // Pode tirar, apos debugar
          float gmag=mag(l+l2,c+c2);
          float gdeg=deg(l+l2,c+c2)-k[0];
          if (gdeg<0) gdeg+=360; // 0<=gdeg<360
          int q=quantiza(gdeg);
          f(arredonda(k[1]),q)+=gmag;
        }
      if (bcinv) {
        double soma=0.0;
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            soma += f(l2,c2);
        double media=soma/f.total();
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            f(l2,c2) -= media;
        double somaq=0.0;
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            somaq += elev2(f(l2,c2));
        double desvio=somaq/f.total();
        if (desvio>epsilon) {
          for (int l2=0; l2<f.rows; l2++)
            for (int c2=0; c2<f.cols; c2++) 
              f(l2,c2) /= desvio;
        } else {
          for (int l2=0; l2<f.rows; l2++)
            for (int c2=0; c2<f.cols; c2++) 
              f(l2,c2) = 0.0;
        }
      } else {
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            f(l2,c2) /= conta[l2]; // l=anel
      }
      int i=0;
      for (int l2=0; l2<f.rows; l2++)
        for (int c2=0; c2<f.cols; c2++) {
          features(l,c,i)=f(l2,c2);
          i++;
        }
    }
  }
  printf("\n");
}
*/

void gerarihog(int argc, char** argv)
{ /*
  if (argc<5) {
    printf("GeraHOG: Extrai HOG de imagens (ex: 128x128) e grava como .IMG\n");
    printf("GeraHOG \"default | nBins=8 | cellSize=16\" hog.img -1|+1 imagens*.jpg\n");
    printf("  Se nao existe hog.img, cria novo\n");
    printf("  Caso contrario, acrescenta novas linhas em pedestre.img\n");
    printf("  -1: exemplo negativo. +1: exemplo positivo.\n");
    erro("Erro: Numero de argumentos invalido");
  }

  int nBins=8;
  int cellSize=16;
  ISTR arq(argv[1]);
  string st=leStr(arq);
  while (st!="eof") {
    if (st=="default") {
    } else if (st=="nbins") {
      leNum(arq,nBins); 
    } else if (st=="cellsize") {
      leNum(arq,cellSize); 
    } else {
      erro("Erro: Palavra desconhecida ",st);
    }
    st=leStr(arq);
  }
  Mat_<COR> a; le(a,argv[4]);
  int rows=a.rows; int cols=a.cols;
  int ns=nBins;
  int nl=rows/cellSize;
  int nc=cols/cellSize;
  int features=ns*nl*nc;

  int n=argc-4;
  Mat_<FLT> pedant; 
  if (existeArq(argv[2])) {
    le(pedant,argv[2]);
    if (pedant.cols!=features+1) erro("Erro: pedestre.img deve ter features+1 colunas. ",format("Features=%d",features));
  }
  Mat_<FLT> ped(pedant.rows+n,features+1,0.0);
  for (int l=0; l<pedant.rows; l++)
    for (int c=0; c<pedant.cols; c++)
      ped(l,c)=pedant(l,c);
  int inicio=pedant.rows;

  FLT rotulo; convArg(rotulo,argv[3]);

  for (int i=0; i<n; i++) {
    le(a,argv[i+4]); 
    printf("Processando %s\r",argv[i+4]);
    HOG hog(a,nBins,cellSize); 
    if (hog.blo.size[0]!=ns || hog.blo.size[1]!=nl || hog.blo.size[2]!=nc) 
      erro("Erro: Dimensoes diferentes",format("ns=%d nl=%d nc=%d",hog.blo.size[0],hog.blo.size[1],hog.blo.size[2]));
    for (int l=0; l<nl; l++)
      for (int c=0; c<nc; c++) {
        for (int s=0; s<ns; s++) 
          ped(i+inicio,l*nc*ns+c*ns+s)=hog.blo(s,l,c);
        ped(i+inicio,features)=rotulo;
      }
  }
  printf("\n");
  imp(ped,argv[2]);
  */
}

void rotaleat(int argc, char** argv)
{ if (argc<3) {
    printf("RotAleat: Rotaciona as imagens em angulo aleatorio\n");
    printf("RotAleat sai ent*.ppm\n");
    printf("  Vai gerar sai12345.ppm para ent12345.ppm\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<COR> ent,sai;
  Mat_<double> rot;
  string st1=argv[1];
  RNG rng;
  for (int i=2; i<argc; i++) {
    string sti=argv[i];
    le(ent,sti);
    int n=arredonda(sqrt(double(elev2(ent.rows)+elev2(ent.cols))));
    sai.create(n,n);

    //double angulo=deg2rad(rng.uniform(0.0,360.0));
    double vdouble1[]={1,0,-ent.cols/2.0,
                      0,1,-ent.rows/2.0,
                      0,0,1};
    Mat_<double> t1(3,3, vdouble1);
    //double c=cos(angulo);
    //double s=sin(angulo);
    double vdouble2[]={1,0,-ent.cols/2.0,
             0,1,-ent.rows/2.0,
             0,0,1
            };
    Mat_<double> r(3,3, vdouble2);
    double vdouble3[]={1,0,n/2.0,
                       0,1,n/2.0,
                       0,0,1
                      };
    Mat_<double> t2(3,3, vdouble3);
    Mat_<double> m=t2*r*t1;
    m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3
  
    warpAffine(ent,sai,m,sai.size(),INTER_LINEAR,BORDER_REPLICATE);
    sti.replace(0,3,st1);
    imp(sai,sti);
  }
}



//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< ClassRihog.exe: Classificacao usando rotation-invariant HOG v1.00>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"  GeraRihog - Extrai HOG de imagens no centro e grava como .IMG\n"
"  RotAleat  - Rotaciona imagens em angulo aleatorio\n"
"...............................................................................\n";

int main(int argc, char** argv)
{ 
  printf(about);
  if (argc<2) {
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="gerarihog") gerarihog(argc-1,&argv[1]);
    else if (comando=="rotaleat")  rotaleat(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto: %d.%02d segundos.\n",t1/100,t1%100);
  }
}
  
  
