#include <cekeikon.h>

M3d_<FLT> serieTemp(M3d_<FLT>& a, int l, int c)
{ Range r0(0,a.size[0]); Range r1(l,l+1); Range r2(c,c+1);
  //return a((Range[]){r0,r1,r2});
  Range ap[]={r0,r1,r2};
  return a(ap);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< POHOG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class POHOG {
 public:
  Mat_<Vec2f> kernel; // angulo, anel (-1 indica fora de circulo ou no centro)
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 5
  int raio; // raio = naneis*larguraAnel = 40
  double raioInic; // raioInic = raio / sqrt(naneis)
  vector<double> raios; //  0    1    2    3    4 
                        //  17.9 25.3 31.0 35.8 40
  int diametro; // diametro = 2*raio+1 = 81
  int nbins; // 20
  int ngauss; // 37
  int descriptorSize; // descriptorSize=nbins*naneis=50
  //int tnl; // tnl=diametro+2*(ngauss/2)
  //double tuso; // tuso=double(diametro)/double(tnl)=0.9759
  bool bcinv;

  void inicializa(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true); // OK
  POHOG(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true) { inicializa(naneisp,raiop,nbinsp,ngaussp,bcinvp); } // OK

  int quantiza(double d);
  void computeDescriptor(Mat_<GRY> a, M3d_<FLT>& b);
  Mat_<FLT> distance(M3d_<FLT> aFeatures, M3d_<FLT> qFeatures);
};

void POHOG::inicializa(int naneisp, int raiop, int nbinsp, int ngaussp, bool bcinvp)
// Nota: Revisar esta rotina de CIHOG - estava errada.
{ naneis=naneisp;
  raio=raiop;
  raioInic = double(raio) / sqrt(double(naneis));
  raios.resize(naneis);
  for (int i=0; i<naneis; i++)
    raios[i] = sqrt(double(i+1)) * raioInic + epsilon; // epsilon foi acrescentado para incluir os pixels no limite

  //cout << raios << endl;

  diametro=2*raio+1;
  nbins=nbinsp;
  ngauss=ngaussp;
  if (ngauss>0 && ngauss%2==0) erro("Erro POHOG::inicializa: ngauss should be odd");
  descriptorSize=nbins*naneis;
  //tnl=diametro+2*(ngauss/2);
  //tuso=double(diametro)/double(tnl);
  bcinv=bcinvp;

  kernel.create(diametro,diametro);
  kernel.setTo(Scalar(-1,-1)); // Pode tirar, depois de debugar
  IMG_<Vec2f> kernelp(kernel); // angulo, anel - o centro (lc,cc) e' setado automaticamente
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec2f& p=kernelp.atC(x,y);
      double r=sqrt(double(x*x+y*y));
      int q=naneis; 
      for (int i=0; i<naneis; i++)
        if (r<=raios[i]) { 
          q=i; break; 
        }

      if (q>=naneis || r==0.0) { 
        p[0]=-1; p[1]=-1; continue; 
      } else { 
        p[0]=fastAtan2(y,x); // entre 0 e 360
        p[1]=q; 
      }
    }

//   for (int l=0; l<kernel.rows; l++) {
//     for (int c=0; c<kernel.cols; c++) 
//       printf("%5.0f ",kernel(l,c)[0]);
//     printf("\n");
//   }
//   printf("\n");
//   
//   for (int l=0; l<kernel.rows; l++) {
//     for (int c=0; c<kernel.cols; c++) 
//       printf("%5.0f ",kernel(l,c)[1]);
//     printf("\n");
//   }
//   printf("\n");
  
  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[1]>=0) 
        conta[kernel(l,c)[1]]++;
    }
}

int POHOG::quantiza(double d)
// 0<=d<360
{ const double denominador=360.0/nbins;
  const double denominador2=denominador/2;
  if (d<0 || 360<=d) erro("Erro POHOG:quantiza-1 fora de 0 a 360");
  d+=denominador2;
  if (d>=360) d-=360; // 0<=d<360
  if (d<0 || 360<=d) erro("Erro POHOG:quantiza-2 fora de 0 a 360");
  int q=int(d/denominador);      
  if (!(0<=q && q<nbins)) erro("Erro POHOG::quantiza inesperado"); 
  return q;  
}

void POHOG::computeDescriptor(Mat_<GRY> a, M3d_<FLT>& b)
// Na saida, b(nbins*anel+q,l,c) contem a media de modulo gradiente daquele bin
// q e' angulo quantizado
// Posicao invalida e' zero
{ if (a.rows<diametro || a.cols<diametro) 
    erro("Erro POHOG::computeDescriptor: imagem muito pequena ou kernel muito grande");

  if (ngauss>1)
    GaussianBlur(a, a, Size(ngauss,ngauss), 0, 0, BORDER_REPLICATE); // In-place filtering is supported.

  // Calcula gradientes gx e gy
  Mat_<SHT> gx,gy; // gradiente 2 direcoes: ->  ^
  //Mat_<FLT> k0(1,3,(FLT[]){-1, 0, 1});
  Mat_<FLT> k0 = (Mat_<FLT>(1,3) << -1, 0, 1);

  filter2D(a, gx, CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  //Mat_<FLT> k1(3,1,(FLT[]){1, 0, -1});
  Mat_<FLT> k1 = (Mat_<FLT>(3,1) << 1, 0, -1);
  filter2D(a, gy, CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 

  // Calcula magnitude mag e angulo ang
  Mat_<FLT> mag(a.size()); Mat_<FLT> ang(a.size());
  for (int l=0; l<a.rows; l++)
    for (int c=0; c<a.cols; c++) {
      ang(l,c)=fastAtan2(gy(l,c),gx(l,c)); // 0<=ang<360
      mag(l,c)=sqrt(double(elev2(gx(l,c))+elev2(gy(l,c))));
    }

  // Aloca b - features de saida
  //b.create(3,(int[]){descriptorSize,a.rows,a.cols});
  int bp[]={descriptorSize,a.rows,a.cols};
  b.create(3,bp);
  b.setTo(0.0); 
  // anel0-ang0 anel0-ang1 etc
  // nbins*anel+q

  // Calcula descriptor b
  IMG_<Vec2f> kernelp(kernel); // angulo, anel - o centro (lc,cc) e' setado automaticamente
  for (int l=raio; l<a.rows-raio; l++)
    for (int c=raio; c<a.cols-raio; c++) {
      for (int l2=kernelp.minl; l2<=kernelp.maxl; l2++)
        for (int c2=kernelp.minc; c2<=kernelp.maxc; c2++) {
          Vec2f k=kernelp.atc(l2,c2); // angulo, anelk
          double degk=k[0];
          int anelk=cvRound(k[1]);
          if (anelk==-1) continue;
          if (degk<0 || degk>=360) erro("Erro degk fora da faixa");   // apagar depois de debugar
          if (anelk<0 || anelk>=naneis) erro("Erro anelk fora da faixa"); // apagar depois de debugar
          double deg=ang(l+l2,c+c2)-degk; // angulo da imagem - angulo do kernel
          if (deg<0) deg+=360; // 0<=deg<360
          int q=quantiza(deg);
          b(nbins*anelk+q,l,c) += mag(l+l2,c+c2);
        }
      for (int l3=0; l3<naneis; l3++)
        for (int c3=0; c3<nbins; c3++)
          b(nbins*l3+c3) /= conta[l3]; // l3=anelk
    } // l,c
}

Mat_<FLT> POHOG::distance(M3d_<FLT> aFeatures, M3d_<FLT> qCentro)
// Na parte invalida, 0.0
{ if (aFeatures.size[0]!=qCentro.size[0])
    erro("Erro POHOG::distance: dimensoes diferentes");
  int ns=aFeatures.size[0];
  if (aFeatures.dims!=3 || qCentro.dims!=3) 
    erro("Erro POHOG::distance: dimensao deve ser 3");
  Mat_<FLT> d(aFeatures.size[1],aFeatures.size[2],0.0);
  for (int l=raio; l<d.rows-raio; l++)
    for (int c=raio; c<d.cols-raio; c++) {
      double soma=0.0;
      for (int s=0; s<ns; s++)
        soma += abs(aFeatures(s,l,c)-qCentro(s));
      d(l,c) = soma/ns;
    }
  return d;
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void pohog(int argc, char** argv)
{ if (argc!=3) {
    printf("POHOG: Compute POHOG features.\n");
    printf("POHOG analyze.tga output.img\n");
    printf("  Output will have the same size as analyze.\n");
    printf("  Invalid pixels at canvas borders will be -infinite.\n");
    erro("Error: Invalid number of arguments");
  }
  
  Mat_<GRY> a; le(a,argv[1]);

  POHOG p(2, 4, 3, 1, false);

  M3d_<FLT> b;
  p.computeDescriptor(a,b);
  
  imp3dImg(b,argv[2]);
}

void dispohog(int argc, char** argv)
{ if (argc!=4) {
    printf("DisPohog: Distance between A and Q using POHOG.\n");
    printf("DisPohog A.PGM Q.PGM P.IMG\n");
    printf("  Output will have the same size as analyze.\n");
    printf("  Invalid pixels at canvas borders will be +infinite.\n");
    erro("Error: Invalid number of arguments");
  }
  
  Mat_<GRY> a; le(a,argv[1]);
  Mat_<GRY> q; le(q,argv[2]);

  POHOG pohog(4, 15, 16, 1, false); //  POHOG(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true)

  printf("Computing features...\n");
  M3d_<FLT> aFeatures;
  pohog.computeDescriptor(a,aFeatures);
  
  //Mat_<FLT> t=fatia(aFeatures,0);
  //mostra(t);

  M3d_<FLT> qFeatures;
  pohog.computeDescriptor(q,qFeatures);

  //t=fatia(qFeatures,0);
  //mostra(t);

  printf("Computing distances...\n");
  int lc=(q.rows-1)/2; int cc=(q.cols-1)/2;
  M3d_<FLT> qCentro=serieTemp(qFeatures,lc,cc);
  Mat_<FLT> p=pohog.distance(aFeatures,qCentro);

  //double maxVal;
  //minMaxLoc(p, 0, &maxVal);
  //p=p/maxVal;
  p=normaliza(p);
  
  imp(p,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< DENSE: Campo denso de descritores inv-rotacao, v1.00>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"Pohog     - Campo denso de POHOG (80)\n"
"DisPohog  - Distancia entre A e Q usando POHOG\n"
"...............................................................................\n";

int main(int argc, char** argv)
{
  if (argc<2) {
    printf(about);
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="pohog")     pohog(argc-1,&argv[1]);
    else if (comando=="dispohog")  dispohog(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto=%d.%02ds\n",t1/100,t1%100);
  }
}


