#include <cekeikon.h>
#include <opencv2/nonfree/nonfree.hpp>

void leTransf(Mat_<DBL>& mat33, string nome)
{ FILE* arq=fopen(nome.c_str(),"rt");
  if (arq==NULL) erro("Erro arquivo inexistente ",nome);
  mat33.create(3,3);
  for (int l=0; l<mat33.rows; l++)
    for (int c=0; c<mat33.rows; c++)
      leNum(arq,mat33(l,c));
  fclose(arq);
}

//<<<<<<<<<<<<<<<<<<<<<<<< key <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
//class KeyPoint
//Point2f pt: coordinates of the keypoint
//flooat size: diameter of the meaningful keypoint neighborhood
//flooat angle: computed orientation of the keypoint (-1 if not applicable). Em graus
//float response: the response by which the most strong keypoints have been selected. 
//int octave: octave (pyramid layer) from which the keypoint has been extracted
//int class_id: object id that can be used to clustered keypoints by an object they belong to

void keysift(int argc, char** argv)
{ if (argc!=3 && argc!=5) {
    printf("KeySift a.pgm a.xml [thresh edgethresh]\n");
    printf("  Ex: KeySift a.pgm a.xml 0.02 10\n");
    erro("Erro: Numero de argumentos invalido");
  }

  //initModule_nonfree(); // to load SURF/SIFT etc.

  Mat_<GRY> a; le(a,argv[1]);
  SiftFeatureDetector siftDet;
  if (argc==5) {
    double thresh; convArg(thresh,argv[3]);
    double edgethresh; convArg(edgethresh,argv[4]);
    siftDet=SiftFeatureDetector(thresh,edgethresh);
  }

  vector<KeyPoint> k;
  siftDet.detect(a,k);
  printf("KeyPoints k=%u\n",unsigned(k.size()));

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  fs.release();
}

void keysurf(int argc, char** argv)
{ if (argc!=3 && argc!=4) {
    printf("KeySurf a.pgm a.xml [thresh]\n");
    printf("  Ex: KeySurf a.pgm a.xml 8000\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> a; le(a,argv[1]);
  SurfFeatureDetector surfDet;
  if (argc==4) {
    double thresh; convArg(thresh,argv[3]);
    surfDet=SurfFeatureDetector(thresh);
  }

  vector<KeyPoint> k;
  surfDet.detect(a,k);
  printf("Keypoints=%u\n",unsigned(k.size()));

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints", k);
  fs.release();
}

void keyfast(int argc, char** argv)
{ if (argc!=3 && argc!=4) {
    printf("KeyFast a.pgm a.xml [thresh]\n");
    printf("  Ex: KeyFast a.pgm a.xml 1\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> a; le(a,argv[1]);
  FastFeatureDetector fastDet;
  if (argc==4) {
    int thresh; convArg(thresh,argv[3]);
    fastDet=FastFeatureDetector(thresh);
  }

  vector<KeyPoint> k;
  fastDet.detect(a,k);
  printf("Keypoints=%u\n",unsigned(k.size()));

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints", k);
  fs.release();
}

void keystar(int argc, char** argv)
{ if (argc!=3) {
    printf("KeyStar a.pgm a.xml\n");
    printf("  Ex: KeySift a.pgm a.xml\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> a; le(a,argv[1]);
  StarFeatureDetector starDet(64,10);

  vector<KeyPoint> k;
  starDet.detect(a,k);
  printf("Keypoints=%u\n",unsigned(k.size()));

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints", k);
  fs.release();
}

void keyrand(int argc, char** argv)
{ if (argc<7) {
    printf("KeyRand a.pgm b.pgm mat33.txt n keyA.xml keyB.xml\n");
    printf("  [seed=7 diamMin=30 diamMax=40 pertPt=0 pertDiam=0 pertDeg=0]\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> a; le(a,argv[1]);
  Mat_<GRY> b; le(b,argv[2]);

  Mat_<DBL> mat33; leTransf(mat33,argv[3]); 

  int n; convArg(n,argv[4]);

  int seed=7; int diamMin=30; int diamMax=40; int pertPt=0; int pertDiam=0; int pertDeg=0;
  if (argc>=8) { 
    ISTR arq(argc-7,&argv[7]);
    seekArg(arq,"seed",seed);
    seekArg(arq,"diammin",diamMin);
    seekArg(arq,"diammax",diamMax);
    seekArg(arq,"pertpt",pertPt);
    seekArg(arq,"pertdiam",pertDiam);
    seekArg(arq,"pertdeg",pertDeg);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  mysrand(seed);

  vector<KeyPoint> ka;
  vector<KeyPoint> kb;
  Mat_<DBL> pa(3,1), pb(3,1), qa(3,1), qb(3,1);
  KeyPoint keya,keyb;
  //KeyPoint()  :  pt(0,0),  size(0),  angle(-1),  response(0),  octave(0),

  while (int(ka.size())<n) {
    keya.pt.x=myrand()%a.cols;
    keya.pt.y=myrand()%a.rows; // y para baixo
    keya.size=diamMin+myrand()%(diamMax-diamMin); // diametro
    keya.angle=myrand()%360; // angle comeca 3 horas e anda anti-horario

    double raioa=keya.size/2; // size e' diametro
    if (keya.pt.x-raioa<0 || keya.pt.x+raioa>=a.cols || keya.pt.y-raioa<0 || keya.pt.y+raioa>=a.rows) continue;

    pa(0)=keya.pt.x;
    pa(1)=keya.pt.y; // y para baixo
    pa(2)=1;

    double rad=deg2rad(keya.angle);
    Mat_<DBL> v(3,1);
    v(0)=cos(rad)*raioa;
    v(1)=-sin(rad)*raioa; // "-" para ter y para baixo
    v(2)=0;

    qa=pa+v;

    pb=mat33*pa;
    pb(0)=pb(0)/pb(2); pb(1)=pb(1)/pb(2); pb(2)=pb(2)/pb(2);
    qb=mat33*qa;
    qb(0)=qb(0)/qb(2); qb(1)=qb(1)/qb(2); qb(2)=qb(2)/qb(2);

    keyb.pt.x=pb(0);
    keyb.pt.y=pb(1); // keyb y para baixo
    keyb.size = 2* sqrt( elev2(pb(0)-qb(0)) + elev2(pb(1)-qb(1)) ); // diametro
    keyb.angle = rad2deg( atan2(-qb(1)+pb(1),qb(0)-pb(0)) ); // "-" para ter y para baixo

    if (pertPt>0) {
      keyb.pt.x += myrand() % (2*pertPt+1) - pertPt;
      keyb.pt.y += myrand() % (2*pertPt+1) - pertPt;
    }
    if (pertDiam>0)
      keyb.size += myrand() % (2*pertDiam+1) - pertDiam;
    if (pertDeg>0)
      keyb.angle += myrand() % (2*pertDeg+1) - pertDeg;
   
    double raiob=keyb.size/2; // size e' diametro
    if (keyb.pt.x-raiob<0 || keyb.pt.x+raiob>=a.cols || keyb.pt.y-raiob<0 || keyb.pt.y+raiob>=a.rows) continue;

    ka.push_back(keya);
    kb.push_back(keyb);
  }

  FileStorage fs; 
  if (!fs.open(argv[5],FileStorage::WRITE)) erro("Erro abertura arquivo ",argv[5]);
  write(fs,"keypoints", ka);
  fs.release();

  if (!fs.open(argv[6],FileStorage::WRITE)) erro("Erro abertura arquivo ",argv[6]);
  write(fs,"keypoints", kb);
  fs.release();
}

void keycgsize(int argc, char** argv)
{ if (argc!=4) {
    printf("KeyCgSize ent.xml sai.xml fator\n");
    printf("  Change key size por fator. Pode ter ent.xml=sai.xml\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<KeyPoint> k;
  FileStorage fs; 

  if (!fs.open(argv[1],FileStorage::READ)) erro("Erro abertura arquivo ",argv[1]);
  read(fs["keypoints"],k); 
  fs.release();

  double fator; convArg(fator,argv[3]);

  for (unsigned i=0; i<k.size(); i++)
    k[i].size = fator*k[i].size; // size e' diametro

  if (!fs.open(argv[2],FileStorage::WRITE)) erro("Erro abertura arquivo ",argv[2]);
  write(fs,"keypoints", k);
  fs.release();
}

bool maiorKey(KeyPoint a, KeyPoint b) //>
{ return a.size>b.size; }

void keygrande(int argc, char** argv)
{ if (argc!=5) {
    printf("KeyGrande ent.pgm ent.xml sai.xml n\n");
    printf("  Seleciona n maiores keypoints sem interseccao com canvas\n");
    printf("  Se n e' zero ou muito grande, apenas filtra os keypoints com inter\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> key;
  FileStorage fs; 

  if (!fs.open(argv[2],FileStorage::READ)) erro("Erro abertura arquivo ",argv[1]);
  read(fs["keypoints"],key); 
  fs.release();

  int n; convArg(n,argv[4]); 
  if (!(0<=n)) erro("Erro: n invalido");
  sort(key,maiorKey);

  // Eliminar keys que tem interseccao com canvas
  vector<KeyPoint> key2;
  for (int i=0; i<(int)key.size(); i++) {
    KeyPoint k=key[i];
    double raio=k.size/2; // size e' diametro
    if (!(k.pt.x-raio<0 || k.pt.x+raio>=a.cols || k.pt.y-raio<0 || k.pt.y+raio>=a.rows))
      key2.push_back(k);
  }
  printf("Apaguei keys que intersectam canvas. Keys inicio=%u Keys depois=%u.\n",unsigned(key.size()),unsigned(key2.size()));

  if (key2.size()-n>0 && n>0) {
    printf("n=%d size=%u. Vou apagar %u keypoints\n",n,unsigned(key2.size()),unsigned(key2.size()-n));
    key2.erase(key2.begin()+n, key2.end()); 
  } else {
    printf("n=%d size=%u. Nao vou apagar keypoints\n",n,unsigned(key2.size()));
  }

  if (!fs.open(argv[3],FileStorage::WRITE)) erro("Erro abertura arquivo ",argv[3]);
  write(fs,"keypoints", key2);
  fs.release();
}

void keyshow(int argc, char** argv)
{ if (argc!=4 && argc!=5) {
    printf("keyshow a.pgm a.xml a.ppm [modo]\n");
    printf("  modo=r/n/R/N r=rich n=normal minuscula=vermelho maiuscula=multicor\n");
    printf("  Default=r\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  FileStorage fs;
  if (!fs.open(argv[2],FileStorage::READ)) erro("Erro abertura arquivo ",argv[2]);
  vector<KeyPoint> k;
  read(fs["keypoints"],k); 
  fs.release();

  char modo='r';
  if (argc==5) modo=argv[4][0];
  if (modo!='r' && modo!='n' && modo!='R' && modo!='N') erro("Erro: Modo invalido");
  Scalar cor=Scalar::all(-1); if (modo=='r' || modo=='n') cor=Scalar(0,0,255);
  int flags=0; if (modo=='r' || modo=='R') flags=DrawMatchesFlags::DRAW_RICH_KEYPOINTS;

  Mat_<COR> b;
  drawKeypoints(a,k,b,cor,flags);
  imp(b,argv[3]);
}

void keymyshow(int argc, char** argv)
{ if (argc!=4) {
    printf("keymyshow a.pgm a.xml a.ppm\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  FileStorage fs;
  if (!fs.open(argv[2],FileStorage::READ)) erro("Erro abertura arquivo ",argv[2]);
  vector<KeyPoint> k;
  read(fs["keypoints"],k); 
  fs.release();

  Mat_<COR> b=keyMyShow(a,k,'R',1.0);
  imp(b,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< des <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<,
void dessift(int argc, char** argv)
{ if (argc!=3 && argc!=4) {
    printf("DesSift a.pgm DesA.xml [KeyA.xml]\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  if (argc==4) {
    FileStorage fs;
    if (!fs.open(argv[3],FileStorage::READ)) erro("Erro abertura arquivo ",argv[3]);
    read(fs["keypoints"],k); 
    fs.release();
  } else {
    SiftFeatureDetector siftDet;
    siftDet.detect(a,k);
  }
  printf("KeyPoints k=%u\n",unsigned(k.size()));

  SiftDescriptorExtractor siftDes;  
  Mat_<FLT> d;
  siftDes.compute(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

void dessurf(int argc, char** argv)
{ if (argc!=3 && argc!=4) {
    printf("DesSurf a.pgm DesA.xml [KeyA.xml]\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  if (argc==4) {
    FileStorage fs;
    if (!fs.open(argv[3],FileStorage::READ)) erro("Erro abertura arquivo ",argv[3]);
    read(fs["keypoints"],k); 
    fs.release();
  } else {
    SurfFeatureDetector surfDet;
    surfDet.detect(a,k);
  }
  printf("KeyPoints k=%u\n",unsigned(k.size()));

  SurfDescriptorExtractor surfDes;  
  Mat_<FLT> d;
  surfDes.compute(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class SIMPLE {
 public:
  char modo; // 'c' copy; 'i' in-place.
  int tnl; // tnl=91
  int camadas; // camadas=6
  double tuso; // tuso=double(2*camadas+1)/double(2*camadas+2)=0.928571
  int descriptorSize; // camadas       = 1,  2,  3,  4,  5,   6
                      // descriptorSize= 7, 19, 37, 61, 91, 127
  int dist,dimpar;
  vector< Point > v; // vetor de pontos que serao amostrados
  //Mat_<COR> adeb; // so para debugar

  void computeDescriptorC(Mat_<GRY>& a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptorI(Mat_<GRY>& a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptor(Mat_<GRY>& a, vector<KeyPoint>& k, Mat_<FLT>& d);
  SIMPLE(char _modo='i', int _camadas=6, int _dimpar=1);
};

void SIMPLE::computeDescriptorC(Mat_<GRY>& a, KeyPoint key, Mat_<FLT> dr)
// Note que dr nao foi passado por referencia. Dr tem que vir alocado
{ double fator = tnl / ((1/tuso)*key.size);
  //int draio=arredonda(fator*key.size/2.0);
  //int ddiam=2*draio+1;

  // Aparentemente, esta usando coordenadas em ponto flutuante
  double vt1[]={1,0,-key.pt.x,
                0,1,-key.pt.y,
                0,0,1};
  Mat_<double> t1(3,3, vt1);
  double c=cos(deg2rad(key.angle-90));
  double s=sin(deg2rad(key.angle-90));
  double vr1[]={ c,-s,0,
                 s, c,0,
                 0, 0,1};
  Mat_<double> r(3,3,  vr1);
  double ve[]={ fator,0,0,
                0,fator,0,
                0,0,1 };
  Mat_<double> e(3,3,  ve);
  double vt2[]={1,0,double(tnl/2),
                0,1,double(tnl/2),
                0,0,1};
  Mat_<double> t2(3,3, vt2);
  Mat_<double> m=t2*e*r*t1;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<GRY> t;
  if (fator<1.0) warpAffine(a, t, m, Size(tnl,tnl), INTER_AREA, BORDER_REPLICATE);
  else warpAffine(a, t, m, Size(tnl,tnl), INTER_LINEAR, BORDER_REPLICATE);
  //else warpAffine(a, t, m, Size(tnl,tnl), INTER_LANCZOS4, BORDER_REPLICATE);

  Mat_<GRY> u;
  GaussianBlur(t, u, Size(dimpar,dimpar),0,0,BORDER_REPLICATE);
  //u=t.clone();

  // debugging
  //Mat_<COR> b; converte(a,b);
  //relogioCircular(b,arredonda(key.pt.y),arredonda(key.pt.x),arredonda(key.size/2.0),key.angle-90,COR(0,0,255));
  //mostra(b);
  //mostra(u);

  //IMG_<GRY> it(t); it.lc=tnl/2; it.cc=tnl/2; 
  //for (unsigned i=0; i<v.size(); i++) {
  //  Point p=v[i];
  //  it.atC(p.x,p.y)=0;
  //}
  //relogioCircular(t,tnl/2,tnl/2,draio,-90,255);
  //mostra(t);
  
  IMG_<GRY> iu(u); iu.lc=tnl/2; iu.cc=tnl/2; 
  for (unsigned i=0; i<v.size(); i++) {
    Point p=v[i];
    dr(i)=iu.atC(p.x,p.y);
  }
  double desvio=normal01(dr);
  if (desvio<epsilon) erro("Erro: Desvio<epsilon");
  
  //relogioCircular(u,tnl/2,tnl/2,draio,-90,255);
  //mostra(u);
  //relogioCircular(a,arredonda(key.pt.y),arredonda(key.pt.x),arredonda(key.size/2.0),key.angle-90,255);
  //mostra(a);
}

void SIMPLE::computeDescriptorI(Mat_<GRY>& a, KeyPoint key, Mat_<FLT> dr)
// Note que dr nao foi passado por referencia. Dr tem que vir alocado
{ double fator = tnl / ((1/tuso)*key.size); // size e' diametro. fator=91/(1/0.92857) * (1/size) = 84.49987/size

  double ve[]={ 1/fator,       0, 0,
                      0, 1/fator, 0,
                      0,       0, 1};
  Mat_<double> e(3,3,ve);
  double c=cos(deg2rad(-key.angle-90)); 
  double s=sin(deg2rad(-key.angle-90)); 
  double vr[]={ c,-s,0,
               s,c,0,
               0,0,1};
  Mat_<double> r(3,3,  vr);
  double vt[]={1,0,key.pt.x,
               0,1,key.pt.y,
               0,0,1};
  Mat_<double> t(3,3, vt);
  Mat_<double> m=t*r*e;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<double> p(3,1), q(2,1);
  IMG_<GRY> ap(a,0,0);
  for (unsigned i=0; i<v.size(); i++) { // v e' vetor de pontos que serao amostrados,  
    p(0)=v[i].x; p(1)=v[i].y; p(2)=1; 
    q=m*p;
    dr(i)=ap.atx(arredonda(q(1)),arredonda(q(0)));

    //int l=arredonda(q(1)); int c=arredonda(q(0)); // para debugar
    //if (i==0)       adeb(l,c)=COR(0,0,255); // para debugar
    //else if (i==1)  adeb(l,c)=COR(0,255,0); // para debugar
    //else if (i==68) adeb(l,c)=COR(255,0,0); // para debugar
    //else            adeb(l,c)=COR(0,0,0); // para debugar
  }
  double desvio=normal01(dr);
  if (desvio<epsilon) erro("Erro: Desvio<epsilon");
}

void SIMPLE::computeDescriptor(Mat_<GRY>& a, vector<KeyPoint>& k, Mat_<FLT>& d)
{ d.create(k.size(),descriptorSize);
  if (modo=='c') {
    for (unsigned i=0; i<k.size(); i++) {
      computeDescriptorC(a,k[i],d.row(i));
    }
  } else {
    for (unsigned i=0; i<k.size(); i++) {
      computeDescriptorI(a,k[i],d.row(i));
    }
  }
}

SIMPLE::SIMPLE(char _modo, int _camadas, int _dimpar)
{ modo=_modo;
  if (modo!='c' && modo!='i') erro("Erro: Modo invalido");  // copy ou inplace
  camadas=_camadas; // nao conta o elemento (0,0). Default=6
  tnl=91;
  tuso=double(2*camadas+1)/double(2*camadas+2); // camadas=6 => tuso=0.92857
  descriptorSize=1;
  for (int i=1; i<=camadas; i++) 
    descriptorSize += 6*i;
  dist=arredonda(tuso*tnl/double(2*camadas+1));
  //dimpar=roundimpar(tuso*tnl/double(2*camadas+1));
  dimpar=_dimpar;

  Point p(0,0); v.push_back(p);
  for (int j=1; j<=camadas; j++) {
    for (int i=0; i<6*j; i++) {
      double angrad=i*M_2PI/(6*j)+(double(j-1)/camadas)*M_2PI/(6*j);
      p=Point(arredonda(j*dist*cos(angrad)),arredonda(j*dist*sin(angrad)));
      v.push_back(p);
    }
  }
  //printf("descriptorSize=%d v.size=%d\n",descriptorSize,v.size());
  //cout << v << endl;

//   Mat_<COR> t(tnl,tnl,COR(255,255,255));
//   IMG_<COR> it(t); it.lc=tnl/2; it.cc=tnl/2; 
//   for (unsigned i=0; i<v.size(); i++) {
//     Point p=v[i];
//     if (i==0)      it.atC(p.x,p.y)=COR(0,0,255);
//     else if (i==68) it.atC(p.x,p.y)=COR(0,255,0);
//     else           it.atC(p.x,p.y)=COR(0,0,0);
//   }
//  mostra(t);
}

void dessimp(int argc, char** argv)
{ if (argc<3) {
    printf("DesSimp a.pgm DesA.xml [key=KeyA.xml camadas=6 dgauss=11 modo=i/c]\n");
    printf("  Se keyA.xml nao for fornecido, usa-se SIFT.\n");
    printf("  camadas       = 1,  2,  3,  4,  5,   6\n");
    printf("  descriptorSize= 7, 19, 37, 61, 91, 127\n");
    printf("  dgauss deve ser impar. tnl=91.\n");
    printf("  modo=inplace ou copy (default).\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  int camadas=6; // 127
  int dGauss=11;
  char modo='c';
  if (argc>=4) { 
    ISTR arq(argc-3,&argv[3]);

    string st;
    if (seekFileName(arq,"key",st)) {
      FileStorage fs;
      if (!fs.open(st,FileStorage::READ)) erro("Erro abertura arquivo ",st);
      read(fs["keypoints"],k); 
      fs.release();
    } else {
      SiftFeatureDetector siftDet;
      siftDet.detect(a,k);
    }
    printf("KeyPoints k=%u\n",unsigned(k.size()));

    seekArg(arq,"camadas",camadas);
    seekArg(arq,"dgauss",dGauss);
    if (dGauss%2!=1) erro("Erro: dGauss deve ser impar");

    string st2; seekArg(arq,"modo",st2); 
    if (st2.length()>0) modo=st2[0];
    if (modo!='c' && modo!='i') erro("Erro: modo invalido");

    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<GRY> u;
  if (modo=='i')
    GaussianBlur(a, u, Size(dGauss,dGauss),0,0,BORDER_REPLICATE);
  else 
    a.copyTo(u);

  Mat_<FLT> d;
  SIMPLE simple(modo,camadas,dGauss);
  //converte(u,simple.adeb); // para debugar
  simple.computeDescriptor(u,k,d);
  //mostra(simple.adeb); // para debugar

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< RIFF <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class RIFF {
 public:
  Mat_<Vec3s> kernel; // radial, tangencial, anel (-1 indica fora de circulo ou no centro)
                      // Codigo: 0=-> 2=^ 4=<- 6=V
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 4
  int larguraAnel; // 10
  int raio; // raio = naneis*larguraAnel = 40
  int diametro; // diametro = 2*raio+1 = 81
  int nquant; // 5 impar - o programa deve ser reescrito se mudar este valor
  int descriptorSize; // descriptorSize=nquant*nquant*naneis=100
  int tnl; // tnl=diametro+2=83
  double tuso; // tuso=double(diametro)/double(tnl)=0.9759
  int ngauss;

  void inicializa(int naneisp=4, int larguraAnelp=10);
  RIFF(int naneisp=4, int larguraAnelp=10) { inicializa(naneisp,larguraAnelp); }
  int quantiza(int d);
  void riff(Mat_<GRY> a, M3d_<FLT>& features);

  void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void RIFF::inicializa(int naneisp, int larguraAnelp)
{ naneis=naneisp;
  larguraAnel=larguraAnelp;
  raio=naneis*larguraAnel;
  diametro=2*raio+1;
  nquant=5; // Se nao for 5, tem que reescrever o programa
  ngauss=9;
  descriptorSize=nquant*nquant*naneis;
  tnl=diametro+2*(ngauss/2);
  tuso=double(diametro)/double(tnl);

  kernel.create(tnl,tnl);
  //kernel.setTo(Scalar(0,0,0)); // Pode tirar, depois de debugar
  IMG_<Vec3s> kernelp(kernel);
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec3s& p=kernelp.atC(x,y);

      int q=teto( sqrt(double(x*x+y*y))/larguraAnel )-1; 
      if (q>=naneis || q<0) { 
        p[0]=0; p[1]=0; p[2]=-1; continue; 
      } else { 
        p[2]=q; 
        float deg=fastAtan2(y,x);
        deg += 22.5;
        q=chao(deg/45.0);
        switch (q) {
          case 0: 
          case 8: p[0]=0; p[1]=2; break;
          case 1: p[0]=1; p[1]=3; break;
          case 2: p[0]=2; p[1]=4; break;
          case 3: p[0]=3; p[1]=5; break;
          case 4: p[0]=4; p[1]=6; break;
          case 5: p[0]=5; p[1]=7; break;
          case 6: p[0]=6; p[1]=0; break;
          case 7: p[0]=7; p[1]=1; break;
          default: erro("Erro inesperado 2 ");
        }
      }
    }
  //cout << kernel << endl;
  
  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[2]>=0) 
        conta[kernel(l,c)[2]]++;
    }
}

int RIFF::quantiza(int d)
// Esta funcao so funciona para nquant=5
// Se mudar nquant, tem que reescrever esta funcao
{ if (d<-120) return 0;
  if (d<-40)  return 1;
  if (d<=40)  return 2;
  if (d<=120) return 3;
  else return 4;
}

void RIFF::riff(Mat_<GRY> a, M3d_<FLT>& features)
{ if (a.size()!=kernel.size()) erro("Erro riff: tamanho imagem");
  vector< Mat_<SHT> > g(8); // gradiente 8 direcoes: ->  /  ^  \  <-  /  V  \.
  FLT vk0[]={1,0,FLT(tnl/2),
             0,1,FLT(tnl/2),
             0,0,1};
  Mat_<FLT> k0(1,3, vk0);
  filter2D(a, g[0], CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk1[]={0, 0, 1,  
             0, 0, 0,  
            -1, 0, 0};
  Mat_<FLT> k1(3,3, vk1);
  filter2D(a, g[1], CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk2[]={1, 0, -1};
  Mat_<FLT> k2(3,1, vk2);
  filter2D(a, g[2], CV_16S, k2, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk3[]={1, 0, 0,  
             0, 0, 0,  
             0, 0,-1};
  Mat_<FLT> k3(3,3, vk3);
  filter2D(a, g[3], CV_16S, k3, Point(-1,-1), 0, BORDER_REPLICATE); 
  g[4]=-g[0];
  g[5]=-g[1];
  g[6]=-g[2];
  g[7]=-g[3];

  int vfeatures[]={naneis,nquant,nquant};
  features.create(3,vfeatures); // anel, radial, tangencial
  features.setTo(0.0);
  //if (li<raio || ci<raio || li+raio>=a.rows || ci+raio>=a.cols) {
    //devolve features nulos
  //  erro("Erro riff");
  //}
  int gr, gt;
  //IMG_<Vec3s> kernelp(kernel);
  //printf("kernel: nl=%d nc=%d lc=%d cc=%d raio=%d\n",kernelp.nl,kernelp.nc,kernelp.lc,kernelp.cc,raio);
  //printf("a: rows=%d cols=%d li=%d ci=%d\n",a.rows,a.cols,li,ci);
  //erro("aqui");
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      Vec3s k=kernel(l,c);
      if (k[2]==-1) continue;
      if (k[0]<0 || k[0]>=8) erro("Erro k[0] fora da faixa");
      gr=g[ k[0] ](l,c);
      if (k[1]<0 || k[1]>=8) erro("Erro k[1] fora da faixa");
      gt=g[ k[1] ](l,c);
      //printf("l=%d c=%d gr=%d gt=%d\n",l,c,gr,gt);
      if (k[2]<0 || k[2]>=naneis) erro("Erro k[2] fora da faixa");
      features(k[2],quantiza(gr),quantiza(gt))++;
    }
  for (int s=0; s<features.size[0]; s++)
    for (int l=0; l<features.size[1]; l++)
      for (int c=0; c<features.size[2]; c++)
        features(s,l,c) = features(s,l,c) / conta[s];
  //imp3d(features);
}

void RIFF::computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr)
{ double fator = tnl / ((1/tuso)*key.size);

  double vt1[]={1,0,-key.pt.x,
                0,1,-key.pt.y,
                0,0,1};
  Mat_<double> t1(3,3, vt1);
  //double c=cos(deg2rad(key.angle));
  //double s=sin(deg2rad(key.angle));
  //Mat_<double> r(3,3,  (double[]){ c,s,0,
  //                                -s,c,0,
  //                                 0,0,1
  //                               });
  double ve[]={ fator,0,0,
                0,fator,0,
                0,0,1};
  Mat_<double> e(3,3,  ve);
  double vt2[]={1,0,double(tnl/2),
                0,1,double(tnl/2),
                0,0,1};
  Mat_<double> t2(3,3, vt2);
  //Mat_<double> m=t2*e*r*t1;
  Mat_<double> m=t2*e*t1;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<GRY> t;
  if (fator<1.0) warpAffine(a, t, m, Size(tnl,tnl), INTER_AREA, BORDER_REPLICATE);
  else warpAffine(a, t, m, Size(tnl,tnl), INTER_LINEAR, BORDER_REPLICATE);

  Mat_<GRY> u; t.copyTo(u);
  GaussianBlur(u, t, Size(ngauss,ngauss),0,0,BORDER_REPLICATE);

  M3d_<FLT> features;
  riff(t,features);
  
  int i=0;
  for (int s=0; s<features.size[0]; s++)
    for (int l=0; l<features.size[1]; l++)
      for (int c=0; c<features.size[2]; c++) {
        dr(i)=features(s,l,c);
        i++;
      }
}

void RIFF::computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d)
{ d.create(k.size(),descriptorSize);
  for (unsigned i=0; i<k.size(); i++) {
    computeDescriptorC(a,k[i],d.row(i));
  }
} 

void desriff(int argc, char** argv)
{ if (argc<3) {
    printf("DesRiff a.pgm DesA.xml [key=KeyA.xml naneis=4 larguraAnel=10]\n");
    printf("  Se keyA.xml nao for fornecido, usa-se SIFT.\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  int naneis=4; // 100
  int larguraAnel=10;
  if (argc>=4) { 
    ISTR arq(argc-3,&argv[3]);

    string st;
    if (seekFileName(arq,"key",st)) {
      FileStorage fs;
      if (!fs.open(st,FileStorage::READ)) erro("Erro abertura arquivo ",st);
      read(fs["keypoints"],k); 
      fs.release();
    } else {
      SiftFeatureDetector siftDet;
      siftDet.detect(a,k);
    }
    printf("KeyPoints k=%u\n",unsigned(k.size()));

    seekArg(arq,"naneis",naneis);
    seekArg(arq,"larguraanel",larguraAnel);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<FLT> d;
  RIFF r(naneis,larguraAnel);
  r.computeDescriptor(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< RIHOG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class RIHOG {
 public:
  Mat_<Vec2f> kernel; // angulo, anel (-1 indica fora de circulo ou no centro)
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 5
  int raio; // raio = 40
  double larguraAnel; // larguraAnel = raio/naneis
  int diametro; // diametro = 2*raio+1 = 81
  int nbins; // 20
  int ngauss; // 37
  int descriptorSize; // descriptorSize=nbins*naneis=50
  int tnl; // tnl=diametro+2*(ngauss/2)
  double tuso; // tuso=double(diametro)/double(tnl)=0.9759
  bool bcinv;

  void inicializa(int naneisp=10, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true);
  RIHOG(int naneisp=10, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true) { inicializa(naneisp,raiop,nbinsp,ngaussp,bcinvp); }
  int quantiza(double d);
  void rihog(Mat_<GRY> a, Mat_<FLT>& features);

  void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void RIHOG::inicializa(int naneisp, int raiop, int nbinsp, int ngaussp, bool bcinvp)
// tem erros
{ naneis=naneisp;
  raio=raiop;
  larguraAnel=double(raio)/naneis;
  raio=naneis*larguraAnel;
  diametro=2*raio+1;
  nbins=nbinsp;
  ngauss=ngaussp;
  descriptorSize=nbins*naneis;
  tnl=diametro+2*(ngauss/2);
  tuso=double(diametro)/double(tnl);
  bcinv=bcinvp;

  kernel.create(tnl,tnl);
  kernel.setTo(Scalar(0,-1)); // Pode tirar, depois de debugar
  IMG_<Vec2f> kernelp(kernel); // angulo, anel - o centro (lc,cc) e' setado automaticamente
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec2f& p=kernelp.atC(x,y);
      double r=sqrt(double(x*x+y*y));
      int q=teto(r/larguraAnel)-1;
      if (q>=naneis || q<0) { 
        p[0]=0; p[1]=-1; continue; 
      } else { 
        p[0]=fastAtan2(y,x); // entre 0 e 360
        p[1]=q; 
      }
    }
  
  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[1]>=0) 
        conta[kernel(l,c)[1]]++;
    }
}

int RIHOG::quantiza(double d)
// 0<=d<360
{ const double denominador=360.0/nbins;
  const double denominador2=denominador/2;
  if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-1 fora de 0 a 360");
  d+=denominador2;
  if (d>=360) d-=360; // 0<=d<360
  if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-2 fora de 0 a 360");
  int q=int(d/denominador);      
  if (!(0<=q && q<nbins)) erro("Erro RIHOG::quantiza inesperado"); 
  return q;  
}

void RIHOG::rihog(Mat_<GRY> a, Mat_<FLT>& features)
{ 
  if (a.size()!=kernel.size()) erro("Erro rihog: tamanho imagem");
  vector< Mat_<SHT> > g(2); // gradiente 2 direcoes: ->  ^
  FLT vk0[]={-1, 0, 1};
  Mat_<FLT> k0(1,3, vk0);
  filter2D(a, g[0], CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk1[]={1, 0, -1};
  Mat_<FLT> k1(3,1, vk1);
  filter2D(a, g[1], CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 

  features.create(naneis,nbins);
  features.setTo(0.0);
  //if (li<raio || ci<raio || li+raio>=a.rows || ci+raio>=a.cols) {
    //devolve features nulos
    //erro("Erro RIHOG");
  //}

  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      Vec2f k=kernel(l,c); // angulo, anel
      if (k[1]==-1) continue;
      if (k[0]<0 || k[0]>=360) erro("Erro k[0]=angulo fora da faixa");
      if (k[1]<0 || k[1]>=naneis) erro("Erro k[1]=aneis fora da faixa");
      double deg=fastAtan2(g[1](l,c),g[0](l,c))-k[0];
      double mod=sqrt(double(elev2(g[1](l,c))+elev2(g[0](l,c))));
      if (deg<0) deg+=360; // 0<=deg<360
      int q=quantiza(deg);
      features(arredonda(k[1]),q)+=mod;
    }
  for (int l=0; l<features.rows; l++)
    for (int c=0; c<features.cols; c++)
      features(l,c) = features(l,c) / conta[l]; // l=anel
  //imp3d(features);
}

void RIHOG::computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr)
{ double fator = tnl / ((1/tuso)*key.size);

  double vt1[]={1,0,-key.pt.x,
                0,1,-key.pt.y,
                0,0,1};
  Mat_<double> t1(3,3, vt1);
  //double c=cos(deg2rad(key.angle));
  //double s=sin(deg2rad(key.angle));
  //Mat_<double> r(3,3,  (double[]){ c,s,0,
  //                                -s,c,0,
  //                                 0,0,1
  //                               });
  double ve[]={ fator,0,0,
                 0,fator,0,
                 0,0,1
               };
  Mat_<double> e(3,3,  ve);
  double vt2[]={1,0,double(tnl/2),
                0,1,double(tnl/2),
                0,0,1};
  Mat_<double> t2(3,3, vt2);
  //Mat_<double> m=t2*e*r*t1;
  Mat_<double> m=t2*e*t1;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<GRY> t;
  if (fator<1.0) warpAffine(a, t, m, Size(tnl,tnl), INTER_AREA, BORDER_REPLICATE);
  else warpAffine(a, t, m, Size(tnl,tnl), INTER_LINEAR, BORDER_REPLICATE);

  Mat_<GRY> u; t.copyTo(u);
  GaussianBlur(u, t, Size(ngauss,ngauss),0,0,BORDER_REPLICATE);

  Mat_<FLT> features; // anel, angulo_quantizado
  rihog(t,features);
  
  int i=0;
  for (int l=0; l<features.rows; l++)
    for (int c=0; c<features.cols; c++) {
      dr(i)=features(l,c);
      i++;
    }

  if (bcinv) {
    double soma=0.0;
    for (int i=0; i<dr.cols; i++) soma += dr(i);
    double media=soma/dr.cols;
    for (int i=0; i<dr.cols; i++) dr(i) -= media;
    double somaq=0.0;
    for (int i=0; i<dr.cols; i++) somaq += elev2(dr(i));
    double desvio=somaq/dr.cols;
    if (desvio>epsilon)
      for (int i=0; i<dr.cols; i++) dr(i) /= desvio;
    else
      for (int i=0; i<dr.cols; i++) dr(i) = 0;
  }
}

void RIHOG::computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d)
{ d.create(k.size(),descriptorSize);
  for (unsigned i=0; i<k.size(); i++) {
    computeDescriptorC(a,k[i],d.row(i));
  }
}

void desrihog(int argc, char** argv)
{ if (argc<3) {
    printf("DesRihog a.pgm DesA.xml [key=KeyA.xml naneis=5 raio=40 nbins=20 ngauss=37 bcinv=true]\n");
    printf("  Se keyA.xml nao for fornecido, usa-se SIFT.\n");
    printf("  bcinv=brightness-contrast invariant\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  int naneis=5; 
  int raio=40;
  int nbins=20;
  int ngauss=37;
  bool bcinv=true;
  if (argc>=4) { 
    ISTR arq(argc-3,&argv[3]);

    string st;
    if (seekFileName(arq,"key",st)) {
      FileStorage fs;
      if (!fs.open(st,FileStorage::READ)) erro("Erro abertura arquivo ",st);
      read(fs["keypoints"],k); 
      fs.release();
    } else {
      SiftFeatureDetector siftDet;
      siftDet.detect(a,k);
    }
    printf("KeyPoints k=%u\n",unsigned(k.size()));

    seekArg(arq,"naneis",naneis);
    seekArg(arq,"raio",raio);
    seekArg(arq,"nbins",nbins);
    seekArg(arq,"ngauss",ngauss);
    seekArg(arq,"bcinv",bcinv);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<FLT> d;
  RIHOG r(naneis,raio,nbins,ngauss,bcinv);
  r.computeDescriptor(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< RUHOG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class RUHOG {
 public:
  Mat_<Vec2f> kernel; // angulo, anel (-1 indica fora de circulo ou no centro)
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 5
  int raio; // raio = naneis*larguraAnel = 40
  double raioInic; // raioInic = raio / sqrt(naneis)
  vector<double> raios;
  int diametro; // diametro = 2*raio+1 = 81
  int nbins; // 20
  int ngauss; // 37
  int descriptorSize; // descriptorSize=nbins*naneis=50
  int tnl; // tnl=diametro+2*(ngauss/2)
  double tuso; // tuso=double(diametro)/double(tnl)=0.9759
  bool bcinv;

  void inicializa(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true);
  RUHOG(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true) { inicializa(naneisp,raiop,nbinsp,ngaussp,bcinvp); }
  int quantiza(double d);
  void ruhog(Mat_<GRY> a, Mat_<FLT>& features);

  void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void RUHOG::inicializa(int naneisp, int raiop, int nbinsp, int ngaussp, bool bcinvp)
// tem erros
{ naneis=naneisp;
  raio=raiop;
  raioInic = double(raio) / sqrt(double(naneis));
  raios.resize(naneis);
  for (int i=0; i<naneis; i++)
    raios[i] = sqrt(double(i+1)) * raioInic + epsilon;

//   cout << raios << endl;

  diametro=2*raio+1;
  nbins=nbinsp;
  ngauss=ngaussp;
  descriptorSize=nbins*naneis;
  tnl=diametro+2*(ngauss/2);
  tuso=double(diametro)/double(tnl);
  bcinv=bcinvp;

  kernel.create(tnl,tnl);
  kernel.setTo(Scalar(-1,-1)); // Pode tirar, depois de debugar
  IMG_<Vec2f> kernelp(kernel); // angulo, anel - o centro (lc,cc) e' setado automaticamente
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec2f& p=kernelp.atC(x,y);
      double r=sqrt(double(x*x+y*y));
      int q=naneis; 
      for (int i=0; i<naneis; i++)
        if (r<=raios[i]) { q=i; break; }
      if (q>=naneis || r==0.0) { 
        p[0]=-1; p[1]=-1; continue; 
      } else { 
        p[0]=fastAtan2(y,x); // entre 0 e 360
        p[1]=q; 
      }
    }

//   for (int l=0; l<kernel.rows; l++) {
//     for (int c=0; c<kernel.cols; c++) 
//       printf("%7.0f ",kernel(l,c)[0]);
//     printf("\n");
//   }
//   printf("\n");
//   
//   for (int l=0; l<kernel.rows; l++) {
//     for (int c=0; c<kernel.cols; c++) 
//       printf("%7.0f ",kernel(l,c)[1]);
//     printf("\n");
//   }
//   printf("\n");

  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[1]>=0) 
        conta[kernel(l,c)[1]]++;
    }
}

int RUHOG::quantiza(double d)
// 0<=d<360
{ const double denominador=360.0/nbins;
  const double denominador2=denominador/2;
  if (d<0 || 360<=d) erro("Erro RUHOG:quantiza-1 fora de 0 a 360");
  d+=denominador2;
  if (d>=360) d-=360; // 0<=d<360
  if (d<0 || 360<=d) erro("Erro RUHOG:quantiza-2 fora de 0 a 360");
  int q=int(d/denominador);      
  if (!(0<=q && q<nbins)) erro("Erro RUHOG::quantiza inesperado"); 
  return q;  
}

void RUHOG::ruhog(Mat_<GRY> a, Mat_<FLT>& features)
// Calcula features do centro de a
{ if (a.size()!=kernel.size()) erro("Erro RUHOG: tamanho imagem");
  vector< Mat_<SHT> > g(2); // gradiente 2 direcoes: ->  ^
  FLT vk0[]={-1, 0, 1};
  Mat_<FLT> k0(1,3,vk0);
  filter2D(a, g[0], CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk1[]={1, 0, -1};
  Mat_<FLT> k1(3,1, vk1);
  filter2D(a, g[1], CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 

  features.create(naneis,nbins);
  features.setTo(0.0);
  //if (li<raio || ci<raio || li+raio>=a.rows || ci+raio>=a.cols) {
    //devolve features nulos
    //erro("Erro RUHOG");
  //}

  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      Vec2f k=kernel(l,c); // angulo, anel
      if (k[1]==-1) continue;
      if (k[0]<0 || k[0]>=360) erro("Erro k[0]=angulo fora da faixa");
      if (k[1]<0 || k[1]>=naneis) erro("Erro k[1]=aneis fora da faixa");
      double deg=fastAtan2(g[1](l,c),g[0](l,c))-k[0];
      double mod=sqrt(double(elev2(g[1](l,c))+elev2(g[0](l,c))));
      if (deg<0) deg+=360; // 0<=deg<360
      int q=quantiza(deg);
      features(arredonda(k[1]),q)+=mod;
    }
  for (int l=0; l<features.rows; l++)
    for (int c=0; c<features.cols; c++)
      features(l,c) = features(l,c) / conta[l]; // l=anel
  //imp3d(features);
}

void RUHOG::computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr)
{ double fator = tnl / ((1/tuso)*key.size);

  double vt1[]={1,0,-key.pt.x,
                0,1,-key.pt.y,
                0,0,1
               };
  Mat_<double> t1(3,3, vt1);
  //double c=cos(deg2rad(key.angle));
  //double s=sin(deg2rad(key.angle));
  //Mat_<double> r(3,3,  (double[]){ c,s,0,
  //                                -s,c,0,
  //                                 0,0,1
  //                               });
  double ve[]={ fator,0,0,
                0,fator,0,
                0,0,1};
  Mat_<double> e(3,3, ve);
  double vt2[]={1,0,double(tnl/2),
                0,1,double(tnl/2),
                0,0,1};
  Mat_<double> t2(3,3, vt2);
  //Mat_<double> m=t2*e*r*t1;
  Mat_<double> m=t2*e*t1;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<GRY> t;
  if (fator<1.0) warpAffine(a, t, m, Size(tnl,tnl), INTER_AREA, BORDER_REPLICATE);
  else warpAffine(a, t, m, Size(tnl,tnl), INTER_LINEAR, BORDER_REPLICATE);

  Mat_<GRY> u; t.copyTo(u);
  GaussianBlur(u, t, Size(ngauss,ngauss),0,0,BORDER_REPLICATE);

  Mat_<FLT> features; // anel, angulo_quantizado
  ruhog(t,features);
  
  int i=0;
  for (int l=0; l<features.rows; l++)
    for (int c=0; c<features.cols; c++) {
      dr(i)=features(l,c);
      i++;
    }

  if (bcinv) {
    double soma=0.0;
    for (int i=0; i<dr.cols; i++) soma += dr(i);
    double media=soma/dr.cols;
    for (int i=0; i<dr.cols; i++) dr(i) -= media;
    double somaq=0.0;
    for (int i=0; i<dr.cols; i++) somaq += elev2(dr(i));
    double desvio=somaq/dr.cols;
    if (desvio>epsilon)
      for (int i=0; i<dr.cols; i++) dr(i) /= desvio;
    else
      for (int i=0; i<dr.cols; i++) dr(i) = 0;
  }
}

void RUHOG::computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d)
{ d.create(k.size(),descriptorSize);
  for (unsigned i=0; i<k.size(); i++) {
    computeDescriptorC(a,k[i],d.row(i));
  }
}

void desruhog(int argc, char** argv)
{ if (argc<3) {
    printf("DesRUHOG a.pgm DesA.xml [key=KeyA.xml naneis=5 raio=40 nbins=20 ngauss=37 bcinv=true]\n");
    printf("  Se keyA.xml nao for fornecido, usa-se SIFT.\n");
    printf("  bcinv=brightness-contrast invariant\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  int naneis=5; 
  int raio=40;
  int nbins=20;
  int ngauss=37;
  bool bcinv=true;
  if (argc>=4) { 
    ISTR arq(argc-3,&argv[3]);

    string st;
    if (seekFileName(arq,"key",st)) {
      FileStorage fs;
      if (!fs.open(st,FileStorage::READ)) erro("Erro abertura arquivo ",st);
      read(fs["keypoints"],k); 
      fs.release();
    } else {
      SiftFeatureDetector siftDet;
      siftDet.detect(a,k);
    }
    printf("KeyPoints k=%u\n",unsigned(k.size()));

    seekArg(arq,"naneis",naneis);
    seekArg(arq,"raio",raio);
    seekArg(arq,"nbins",nbins);
    seekArg(arq,"ngauss",ngauss);
    seekArg(arq,"bcinv",bcinv);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<FLT> d;
  RUHOG r(naneis,raio,nbins,ngauss,bcinv);
  r.computeDescriptor(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< forapro <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class FORAPRO {
  //const double minampl=0.0025; // ????
  struct FEA { // features
    double rotmap; // - 1 = rotmap: Numeros entre -pi a +pi. orientacao canonica
    vector<double> radang; // angulos - nrad-1 = radang: vetor de angulos entre 0 e +2pi
                   // vector of radial angles
    vector<double> radmag; // comprimentos - nrad - radmag: Vetor de comprimento 1
                   // vector of radial magnitudes
    vector<double> cirvet; // 2*ncir - cirvet: Vetor de comprimento 1
                   // vector of circular features
  };
  struct PESOTOTAL {
    double vra_wt,wm,wa,wc,wt;
    void inicializa(int nrad, int ncir);
  };
  double rpeso(double r);
  void gerarqux(int kth, Mat_<CPX>& _qux);
  void geracqux(int kth, Mat_<CPX>& _qux);
  void geraqux(M3d_<CPX>& qux);
  void convqux(Mat_<GRY>& qu, M3d_<CPX> _qux, M3d_<CPX>& quc);
  void gerafea(M3d_<CPX> quc, FEA& fea);
 public:
  int nrad,ncir;
  int descriptorSize;
  PESOTOTAL pesototal;
  bool radang,radmag,cirvet;
  M3d_<CPX> qux; // nrad+ncir, tnl, tnl
  int raio; 
  int tnl; // tnl=diametro

  void inicializa(int _nrad, int _ncir, int _raio, bool _radang, bool _radmag, bool _cirvet);
  FORAPRO(int _nrad, int _ncir, int _raio, bool _radang, bool _radmag, bool _cirvet) { inicializa(_nrad,_ncir,_raio,_radang,_radmag,_cirvet); }

  void forapro(Mat_<GRY> a, vector<FLT>& features);
  void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void FORAPRO::PESOTOTAL::inicializa(int nrad, int ncir) {
  vra_wt=0.0;
  for (int i=0; i<nrad-1; i++) // kth=i+2
    vra_wt += 1.0/(i+2.0);
  wm=nrad-1;
  wa=nrad-1;
  wc=2*ncir-1;
  wt=wm+wa+wc;
  wm=wm/wt;
  wa=wa/wt;
  wc=wc/wt;
} // equacao (20) do artigo

double FORAPRO::rpeso(double r)
// peso de nucleo radial
// primeiro fator da equacao (7)
{ double fim=raio+0.5;
  double inicio=1.0;
  if (r<=inicio || fim<=r) return 0.0;
  else return sqrt((r-inicio)*(fim-r));
}

void FORAPRO::gerarqux(int kth, Mat_<CPX>& _qux)
// Gera um nucleo radial de ordem kth e diametro tnl.
// Cada elemento de IMGCPX e' complex<float>
// complex<float> == CPX
// complex<double> == CPXD
{ static CPXD jx(0.0, 1.0);
  _qux.create(tnl,tnl);
  IMG_<CPX> qux(_qux);

  double total=0.0;
  for (int l=qux.minl; l<=qux.maxl; l++)
    for (int c=qux.minc; c<=qux.maxc; c++) {
      double r=sqrt(double(l*l+c*c));
      double peso=rpeso(r);
      total += peso;
      if (peso>0.0) {
        double angulo=kth*arg(CPXD(c,-l));
        qux.atc(l,c) = CPX(peso*cos(angulo),peso*sin(angulo));
      } else {
        qux.atc(l,c) = CPX(0.0,0.0);
      }
    }
  for (int l=qux.minl; l<=qux.maxl; l++)
    for (int c=qux.minc; c<=qux.maxc; c++)
      if (qux.atc(l,c)!=CPX(0.0,0.0))
        qux.atc(l,c) /= total;
}

void FORAPRO::geracqux(int kth, Mat_<CPX>& _qux)
// Gera um nucleo circular de ordem kth e diametro tnl.
// Cada elemento de IMGCPX e' complex<float>
{ static CPXD jx(0.0, 1.0);
  _qux.create(tnl,tnl);
  IMG_<CPX> qux(_qux);
  int raio=tnl/2;

  // continuo
  double total=0.0;
  for (int l=qux.minl; l<=qux.maxl; l++)
    for (int c=qux.minc; c<=qux.maxc; c++) {
      double r=sqrt(double(l*l+c*c));
      if (int(r)>raio) qux.atc(l,c)=CPX(0.0,0.0);
      else {
        double peso;
        if (r>0.0) peso=1.0/(M_2PI*r);
        else peso=0.73;
        total += peso;
        //double angulo = -kth * (r/(raio+1)) * M_2PI;
        double angulo = kth * (r/(raio+1)) * M_2PI;
        qux.atc(l,c) = CPX(peso*cos(angulo),peso*sin(angulo));
      }
    }
  for (int l=qux.minl; l<=qux.maxl; l++)
    for (int c=qux.minc; c<=qux.maxc; c++)
      if (qux.atc(l,c)!=CPX(0.0,0.0))
        qux.atc(l,c) /= total;
}

void FORAPRO::geraqux(M3d_<CPX>& qux)
// I3DCPX e' como se fosse vetor de IMGCPX.
// I3DCPX z: z(s,l,c) z(s)
// Chamar gerarqux nrad vezes. Depois, chama geracqux ncir vezes
// tnl = diametro do template
// Exemplo: nrad=3 e ncir=2
// qux tera 5 fatias: 0..2 referem-se a nucleos radiais
//                    3..4 referem-se a nucleos circulares
{ int vqux[]={nrad+ncir,tnl,tnl};
  qux.create(3,vqux);
  for (int i=0; i<nrad; i++) {
    int ip=i+1;
    Mat_<CPX> t=fatia(qux,i);
    gerarqux(ip,t);
  }
  for (int i=nrad; i<nrad+ncir; i++) {
    int ip=i-nrad+1;
    Mat_<CPX> t=fatia(qux,i);
    geracqux(ip,t);
  }
}

void FORAPRO::convqux(Mat_<GRY>& qu, M3d_<CPX> _qux, M3d_<CPX>& quc)
// Convolucao de qu com cada nucleo de qux, gerando quc
// Se qux tiver 5 fatias, quc tambem tera 5 fatias
// Cuidado: Numero de linhas e colunas de quc e' menor que os de qu
{ if (qu.rows!=qu.cols || qu.rows!=_qux.size[1] || qu.cols!=_qux.size[2]) 
    erro("Erro convqux: Dimensoes invalidas");
  int tnl=_qux.size[1];
  int nl=qu.rows-tnl+1; // resulta 1
  int nc=qu.cols-tnl+1; // resulta 1
  if (nl!=1 || nc!=1) erro("Erro inesperado");
  int vquc[]={_qux.size[0],nl,nc};
  quc.create(3,vquc); quc.setTo(Scalar(0,0,0,0));
  int vqux[]={_qux.size[0],_qux.size[1],_qux.size[2]};
  M3d_<CPX> qux(3,vqux);

  Mat_<CPX> qx; converte(qu,qx);

  // Reflexao de _qux
  for (int i=0; i<qux.size[0]; i++) {
    Mat_<CPX> s=fatia(_qux,i);
    Mat_<CPX> t=fatia(qux,i);
    flip(s,t,-1);
  }

  for (int s=0; s<qux.size[0]; s++) {
    quc(s,0,0)=0.0;
    for (int l=0; l<qux.size[1]; l++)
      for (int c=0; c<qux.size[2]; c++)
        quc(s,0,0)+=qx(l,c)*qux(s,l,c);
  }
}

void FORAPRO::gerafea(M3d_<CPX> quc, FEA& fea)
// Gera mfea (matriz de features) a partir do quc
{ if (quc.size[1]!=1 || quc.size[2]!=1) erro("Erro inesperado");
  //rotmap.resize(quc.nl(),quc.nc()); // dim 1 (orientacao canonica)
  //radang.resize(quc.nl(),quc.nc()); // dim nrad-1 (vector of radial angles)
  //radmag.resize(quc.nl(),quc.nc()); // dim nrad (vector of radial magnitudes)
  //cirvet.resize(quc.nl(),quc.nc()); // dim 2*ncir (vector of circular features)

  { // rotmap: Numeros entre -pi a +pi
    CPX x=quc(0,0,0);
    fea.rotmap=arg(x); // arg = fase do numero complexo
  }

  { // radang: vetor de angulos entre 0 e +2pi
    vector<double>& v=fea.radang;
    v.resize(nrad-1);
    double a1=fea.rotmap;
    for (int s=1; s<nrad; s++) {
      int kth=s+1;
      CPX x=quc(s,0,0);
      //double dk=abs(x);
      double ak=arg(x);
      v[s-1]=modulo((ak-kth*a1),M_2PI); 
    }
  }

  { // radmag: Vetor de comprimento 1
    vector<double>& v=fea.radmag;
    v.resize(nrad);
    for (int s=0; s<int(v.size()); s++) v[s]=abs(quc(s,0,0));
    v=versorl1(v);
  }

  if (ncir>0) {
    { // cirvet: Vetor de comprimento 1
      vector<double>& v=fea.cirvet;
      v.resize(2*ncir);
      for (int s=0; s<ncir; s++) {
        v[2*s]  =real(quc(s+nrad,0,0));
        v[2*s+1]=imag(quc(s+nrad,0,0));
      }
      v=versorl1(v);
    }
  }
}

void FORAPRO::inicializa(int _nrad, int _ncir, int _raio, bool _radang, bool _radmag, bool _cirvet)
{ nrad=_nrad;
  ncir=_ncir;
  raio=_raio;
  tnl=2*raio+1;
  radang=_radang;
  radmag=_radmag;
  cirvet=_cirvet;
  descriptorSize=0;
  if (radang) descriptorSize += nrad-1;
  if (radmag) descriptorSize += nrad;
  if (cirvet) descriptorSize += 2*ncir;
  pesototal.inicializa(nrad,ncir);

  geraqux(qux);
}

void FORAPRO::forapro(Mat_<GRY> a, vector<FLT>& features)
{ M3d_<CPX> quc;
  //printf("a.rows=%d a.cols=%d qux.size[0]=%d qux.size[1]=%d qux.size[2]=%d\n",a.rows,a.cols,qux.size[0],qux.size[1],qux.size[2]);
  convqux(a, qux, quc);

  FEA fea; // 2*nrad+2*ncir, quv.nl(), quv.nc()
  gerafea(quc,fea);

  if (radang) 
    for (int i=0; i<int(fea.radang.size()); i++) 
      features.push_back(fea.radang[i]);
  if (radmag) 
    for (int i=0; i<int(fea.radmag.size()); i++) 
      features.push_back(fea.radmag[i]);
  if (cirvet) 
    for (int i=0; i<int(fea.cirvet.size()); i++) 
      features.push_back(fea.cirvet[i]);
}

void FORAPRO::computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr)
{ double fator = double(tnl)/key.size;

  double vt1[]={1,0,-key.pt.x,
                0,1,-key.pt.y,
                0,0,1};
  Mat_<double> t1(3,3, vt1);
  //double c=cos(deg2rad(key.angle));
  //double s=sin(deg2rad(key.angle));
  //Mat_<double> r(3,3,  (double[]){ c,s,0,
  //                                -s,c,0,
  //                                 0,0,1
  //                               });
  double ve[]={ fator,0,0,
               0,fator,0,
               0,0,1};
  Mat_<double> e(3,3,  ve);
  double vt2[]={1,0,double(tnl/2),
                0,1,double(tnl/2),
                0,0,1};
  Mat_<double> t2(3,3, vt2);
  //Mat_<double> m=t2*e*r*t1;
  Mat_<double> m=t2*e*t1;
  m.pop_back(); // elimina a ultima linha da matriz 3x3 para virar matriz 2x3

  Mat_<GRY> t;
  if (fator<1.0) warpAffine(a, t, m, Size(tnl,tnl), INTER_AREA, BORDER_REPLICATE);
  else warpAffine(a, t, m, Size(tnl,tnl), INTER_LINEAR, BORDER_REPLICATE);

  //Mat_<GRY> u; t.copyTo(u);
  //GaussianBlur(u, t, Size(ngauss,ngauss),0,0,BORDER_REPLICATE);

  vector<FLT> features;
  forapro(t,features);
  
  for (int i=0; i<int(features.size()); i++) 
    dr(i)=features[i];
}

void FORAPRO::computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d)
{ d.create(k.size(),descriptorSize);
  for (unsigned i=0; i<k.size(); i++) {
    computeDescriptorC(a,k[i],d.row(i));
  }
}

void desforapro(int argc, char** argv)
{ if (argc<3) {
    printf("DesForapro a.pgm DesA.xml [key=KeyA.xml nrad=4 ncir=4 raio=40 radang=1 radmag=1 cirvet=1]\n");
    printf("  Se keyA.xml nao for fornecido, usa-se SIFT.\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  vector<KeyPoint> k;
  int nrad=4; 
  int ncir=4;
  int raio=40;
  bool radang=true;
  bool radmag=true;
  bool cirvet=true;
  if (argc>=4) { 
    ISTR arq(argc-3,&argv[3]);

    string st;
    if (seekFileName(arq,"key",st)) {
      FileStorage fs;
      if (!fs.open(st,FileStorage::READ)) erro("Erro abertura arquivo ",st);
      read(fs["keypoints"],k); 
      fs.release();
    } else {
      SiftFeatureDetector siftDet;
      siftDet.detect(a,k);
    }
    printf("KeyPoints k=%u\n",unsigned(k.size()));

    seekArg(arq,"nrad",nrad);
    seekArg(arq,"ncir",ncir);
    seekArg(arq,"raio",raio);
    seekArg(arq,"radang",radang);
    seekArg(arq,"radmag",radmag);
    seekArg(arq,"cirvet",cirvet);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<FLT> d;
  FORAPRO r(nrad,ncir,raio,radang,radmag,cirvet);
  r.computeDescriptor(a,k,d);

  FileStorage fs; 
  fs.open(argv[2],FileStorage::WRITE);
  write(fs,"keypoints",k);
  write(fs,"descriptors",d);
  fs.release();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mat <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<,
void impMatch(vector<DMatch>& matches, vector<KeyPoint>& k1, vector<KeyPoint>& k2, string nome, vector<bool> vet={}) {
// vet[i]==true para imprimir, false para ignorar 
  if (vet.size()>0 && vet.size()!=matches.size()) erro("Erro: impMatch: vet.size()!=matches.size()");
  FILE* arq=fopen(nome.c_str(),"wt");

  unsigned conta=0;
  if (vet.size()>0) {
    for (unsigned i=0; i<vet.size(); i++)
      if (vet[i]) conta++;
  } else conta=matches.size();

  fprintf(arq,"%u 15\n",conta);
  fprintf(arq,"//dist-ratio k1.pt.x k1.pt.y k1.size k1.angle k1.response k1.octave k1.class_id " 
              "k2.pt.x k2.pt.y k2.size k2.angle k2.response k2.octave k2.class_id\n");

  for (unsigned i=0; i<matches.size(); i++) {
    if (vet.size()==0 || vet[i]) {
      KeyPoint key1=k1.at(matches.at(i).queryIdx);
      KeyPoint key2=k2.at(matches.at(i).trainIdx);
      fprintf(arq,"%e  %e %e %e %e %e %d %d  %e %e %e %e %e %d %d\n",matches.at(i).distance, 
                   key1.pt.x,key1.pt.y,key1.size,key1.angle,key1.response,key1.octave,key1.class_id,
                   key2.pt.x,key2.pt.y,key2.size,key2.angle,key2.response,key2.octave,key2.class_id);
    }
  }
  fclose(arq);
}

void leMatch(vector<DMatch>& matches, vector<KeyPoint>& k1, vector<KeyPoint>& k2, string nome)
{ FILE* arq=fopen(nome.c_str(),"rt");
  if (arq==NULL) erro("Erro arquivo inexistente ",nome);
  int tamanho; leNum(arq,tamanho); 
  int nc; leNum(arq,nc); if (nc!=15) erro("Erro: Arquivo nao tem 15 colunas");
  matches.resize(tamanho); k1.resize(tamanho); k2.resize(tamanho);
  for (unsigned i=0; i<matches.size(); i++) {
    leNum(arq,matches[i].distance);
    matches[i].queryIdx=matches[i].trainIdx=i;
    matches[i].imgIdx=-1;

    leNum(arq,k1[i].pt.x);
    leNum(arq,k1[i].pt.y);
    leNum(arq,k1[i].size);
    leNum(arq,k1[i].angle);
    leNum(arq,k1[i].response);
    leNum(arq,k1[i].octave);
    leNum(arq,k1[i].class_id);

    leNum(arq,k2[i].pt.x);
    leNum(arq,k2[i].pt.y);
    leNum(arq,k2[i].size);
    leNum(arq,k2[i].angle);
    leNum(arq,k2[i].response);
    leNum(arq,k2[i].octave);
    leNum(arq,k2[i].class_id);
  }
  fclose(arq);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matnn

void matnn(int argc, char **argv)
{ if (argc<4) {
    printf("MatNN: Acha casamentos do des1 para des2 usando 1-NN\n");
    printf("MatNN des1.xml des2.xml match.txt [ brute | flann ]\n");
    printf("  match.txt estara ordenada pela distancia. menor=melhor\n");
    printf("  Default=brute\n");
    erro("Erro: Numero de argumentos invalido");
  }

  FileStorage fs;
  if (!fs.open(argv[1],FileStorage::READ)) erro("Erro abertura arquivo ",argv[1]);
  vector<KeyPoint> k1; read(fs["keypoints"],k1); 
  Mat_<FLT> d1; read(fs["descriptors"],d1);
  fs.release();
  if (int(k1.size())!=d1.rows) erro("Erro: numero de keypoints != descritores");

  if (!fs.open(argv[2],FileStorage::READ)) erro("Erro abertura arquivo ",argv[2]);
  vector<KeyPoint> k2; read(fs["keypoints"],k2); 
  Mat_<FLT> d2; read(fs["descriptors"],d2);
  fs.release();
  if (int(k2.size())!=d2.rows) erro("Erro: numero de keypoints != descritores");

  char modoMatch='b';
  if (argc>=5) {
    ISTR arq(argc-4,&argv[4]);
    string st=leStr(arq);
    while (st!="eof") {
      if (st=="brute") { 
        modoMatch='b';
      } else if (st=="flann") { 
        modoMatch='f';
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=leStr(arq);
    }
  }

  vector<DMatch> matches;
  if (modoMatch=='b') {
    BFMatcher matcher(NORM_L2);
    matcher.match(d1,d2,matches);
  } else if (modoMatch=='f') {
    FlannBasedMatcher matcher;
    matcher.match(d1,d2,matches);
  } else erro("Erro modoMatch desconhecido");
  sort(matches.begin(),matches.end()); // ordenado pela distancia - menor=melhor

  impMatch(matches,k1,k2,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matratio

struct DMatch2
{
  DMatch d;
  float  distance;
  float  ratio;
};

bool menorDistance(DMatch2 a, DMatch2 b) //<
{ return a.distance<b.distance; }

bool menorRatio(DMatch2 a, DMatch2 b) //<
{ return a.ratio<b.ratio; }

void ratioTest(vector< vector<DMatch> >& nnmatches, vector<DMatch>& matches, double ratio, char ordem) 
{ vector<DMatch2> v; DMatch2 m;
  for (vector< vector<DMatch> >::iterator p=nnmatches.begin(); p!=nnmatches.end(); ++p) {
    if (p->size()!=2) erro("Erro acRateTest inesperado");
    if ((*p)[1].distance>epsilon) m.ratio=(*p)[0].distance/(*p)[1].distance;
    else m.ratio=infinito;
    m.distance=(*p)[0].distance;
    m.d=(*p)[0];
    v.push_back(m);
  }
  if (ordem=='d')      sort(v.begin(),v.end(),menorDistance); // ordenado pela distance
  else if (ordem=='r') sort(v.begin(),v.end(),menorRatio); // ordenado pelo ratio
  else erro("Erro: ordem invalida");

  matches.clear();
  for (vector<DMatch2>::iterator p=v.begin(); p!=v.end(); p++) {
    if (ordem=='d')      p->d.distance = p->distance; // ordenado pela distance
    else if (ordem=='r') p->d.distance = p->ratio;  // ordenado pelo ratio
    if (p->ratio<ratio) matches.push_back(p->d);
  }
}

void acRateTest(vector< vector<DMatch> >& nnmatches, vector<DMatch>& matches, double acRate, char ordem) 
{ vector<DMatch2> v; DMatch2 m;
  for (vector< vector<DMatch> >::iterator p=nnmatches.begin(); p!=nnmatches.end(); ++p) {
    if (p->size()!=2) erro("Erro acRateTest inesperado");
    if ((*p)[1].distance>epsilon) m.ratio=(*p)[0].distance/(*p)[1].distance;
    else m.ratio=infinito;
    m.distance=(*p)[0].distance;
    m.d=(*p)[0];
    v.push_back(m);
  }
  if (ordem=='d')      sort(v.begin(),v.end(),menorDistance); // ordenado pela distance
  else if (ordem=='r') sort(v.begin(),v.end(),menorRatio); // ordenado pelo ratio
  else erro("Erro: ordem invalida");

  int aceite=arredonda(acRate*v.size());
  if (aceite<0 || int(v.size())<aceite) erro("Erro acRate invalido");
  printf("Vou apagar %u casamentos\n",unsigned(v.size()-aceite));
  v.erase(v.begin()+aceite, v.end()); 

  matches.clear();
  for (vector<DMatch2>::iterator p=v.begin(); p!=v.end(); p++) {
    if (ordem=='d')      p->d.distance = p->distance; // ordenado pela distance
    else if (ordem=='r') p->d.distance = p->ratio;  // ordenado pelo ratio
    matches.push_back(p->d);
  }
}

void matratio(int argc, char **argv)
{ if (argc<4) {
    printf("MatRatio: Acha casamentos de des1 para des2 onde 1st/2nd>ratio\n");
    printf("MatRatio des1.xml des2.xml match.txt [ratio=val | acrate=val | ordem=r ou d | brute | flann ]\n");
    printf("  ratio: distancia_melhor/distancia_segundo_melhor. Limiariza pelo ratio e ordena por r/d\n");
    printf("  acrate: acceptance rate entre 0 e 1. Ordena por r/d e apaga.\n");
    printf("  ordem=r (ratio) ou d (distance).\n");
    printf("  Default: ratio=0.8 ordem=r brute\n");
    erro("Erro: Numero de argumentos invalido");
  }

  FileStorage fs;
  if (!fs.open(argv[1],FileStorage::READ)) erro("Erro abertura arquivo ",argv[1]);
  vector<KeyPoint> k1; read(fs["keypoints"],k1); 
  Mat_<FLT> d1; read(fs["descriptors"],d1);
  fs.release();
  if (int(k1.size())!=d1.rows) erro("Erro: numero de keypoints != descritores");

  if (!fs.open(argv[2],FileStorage::READ)) erro("Erro abertura arquivo ",argv[2]);
  vector<KeyPoint> k2; read(fs["keypoints"],k2); 
  Mat_<FLT> d2; read(fs["descriptors"],d2);
  fs.release();
  if (int(k2.size())!=d2.rows) erro("Erro: numero de keypoints != descritores");

  char modo='r'; // 'r' (ratio) ou 'a' (acrate)
  char ordem='r'; // 'd' (distance) ou 'r' (ratio)
  double ratio=0.8; 
  double acRate=0.5;
  char modoMatch='b';

  if (argc>=5) {
    ISTR arq(argc-4,&argv[4]);
    string st=leStr(arq);
    while (st!="eof") {
      if (st=="ratio") { 
        modo='r';
        leNum(arq,ratio); 
        if (ratio<0 || 1<ratio) erro("Erro: ratio fora do intervalo");
      } else if (st=="acrate") { 
        modo='a';
        leNum(arq,acRate); 
        if (acRate<0 || 1<acRate) erro("Erro: acRate fora do intervalo");
      } else if (st=="ordem") { 
        string st2=leStr(arq); 
        if (st2=="") erro("Erro: ordem=NULO");
        ordem=st2[0];
        if (ordem!='r' && ordem!='d') erro("Erro: ordem deve ser r ou d");
      } else if (st=="brute") { 
        modoMatch='b';
      } else if (st=="flann") { 
        modoMatch='f';
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=leStr(arq);
    }
  }

  vector< vector<DMatch> > nnmatches;
  if (modoMatch=='b') {
    BFMatcher matcher(NORM_L2);
    matcher.knnMatch(d1,d2,nnmatches,2);
  } else if (modoMatch=='f') {
    FlannBasedMatcher matcher;
    matcher.knnMatch(d1,d2,nnmatches,2);
  } else erro("Erro: modoMatch invalido");

  vector<DMatch> matches;
  if (modo=='r') {  
    ratioTest(nnmatches,matches,ratio,ordem); // ordenado por rati
  } else if (modo=='a') {
    acRateTest(nnmatches,matches,acRate,ordem); // ordenado pela distancia
  } else erro("Erro: modo invalido");
  printf("Matches_iniciais=%u Matches_finals=%u taxa_aceite=%5.2f%%\n",unsigned(nnmatches.size()),unsigned(matches.size()),100.0*matches.size()/nnmatches.size());

  impMatch(matches,k1,k2,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matshow

void matshow(int argc, char **argv)
{ if (argc!=5) {
    printf("MatShow: Mostra os casamentos\n");
    printf("MatShow img1.pgm img2.pgm match.txt sai.png\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> im1; le(im1,argv[1]);
  Mat_<GRY> im2; le(im2,argv[2]);

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[3]);

  Mat_<COR> sai;
  drawMatches(
     im1,k1, // 1st image and its keypoints
     im2,k2, // 2nd image and its keypoints
     matches,            // the matches
     sai,      // the image produced
     Scalar::all(-1),
     Scalar(255,255,255),
     vector<char>(),
//      DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS | DrawMatchesFlags::DRAW_RICH_KEYPOINTS
     0
  ); 
  imp(sai,argv[4]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matmyshow
void matmyshow(int argc, char **argv)
{ if (argc!=5) {
    printf("MatMyShow: Mostra os casamentos sem chamar funcao OpenCV\n");
    printf("MatMyShow img1.pgm img2.pgm match.txt sai.png\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> im1; le(im1,argv[1]);
  Mat_<GRY> im2; le(im2,argv[2]);

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[3]);

  Mat_<COR> sai=matMyShow(im1,im2,matches,k1,k2,'R',4.0);

  imp(sai,argv[4]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matrate
/*
void matrate(int argc, char **argv)
{ if (argc!=3 && argc!=4) {
    printf("MatRate: Calcula taxa de erro/acerto de casamento, tendo matriz 3x3\n");
    printf("MatRate match.txt mat33.txt [dist]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=dist\n");
    printf("  dist default = 15\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[1]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");

  double dist=15;
  if (argc==4) convArg(dist,argv[3]);

  Mat_<DBL> mat33; leTransf(mat33,argv[2]); mat33=mat33.inv();
  int erros=0, acertos=0;
  for (unsigned i=0; i<k1.size(); i++) {
    Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);
    if (norm(t,p1)>dist) erros++;
    else acertos++;
  }
  int total=erros+acertos;
  printf("Total=%d erros=%d(%4.1f%%) acertos=%d(%4.1f%%)\n",total,erros,100.0*erros/total,acertos,100.0*acertos/total);
}
*/

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matroc

void curvaROC(vector<Vec2d>& rate, Mat_<COR>& a, int ampl=8, COR cor=COR(0,0,255))
// rate na ordem TPR, FPR.
// curva que vai de 0.5 a 1
{ a.create(50*ampl+1,50*ampl+1); a.setTo(Scalar(255,255,255));
  for (int i=0; i<=5; i++) {
    reta(a,10*i*ampl,0,10*i*ampl,a.cols-1,COR(128,128,128));
    putTxt(a,10*i*ampl+1,1,format("%3.1f",(10-i)/10.0));
    reta(a,0,10*i*ampl,a.rows-1,10*i*ampl,COR(128,128,128));
    putTxt(a,a.rows-15,10*i*ampl-24,format("%3.1f",i/10.0));
  }
  putTxt(a,16,1,"tpr");
  putTxt(a,a.rows-15-16,a.cols-30,"fpr");
  // linhas
  // 50 -> 50*ampl
  // 100 -> 0
  // l -> 50*ampl-(l-50)*ampl = (100-l)*ampl
  // colunas
  // 0 -> 0
  // 50 -> 50*ampl
  // c -> c*ampl

  bool inicio=true;
  for (int i=0; i<int(rate.size()); i++) {
    double ld=100*rate[i][0];
    if (ld<50 || ld>100) continue;
    double cd=100*rate[i][1];
    if (cd<0 || cd>50) continue;
    int l=arredonda((100-ld)*ampl);
    int c=arredonda(cd*ampl);
    //a(l,c)=cor;
    if (inicio) { moveTo(l,c); inicio=false; }
    else lineTo(a,l,c,cor);
  }
}

/*
void curvaROC(vector<Vec2d>& rate, Mat_<COR>& a, int ampl=4, COR cor=COR(0,0,255))
// rate na ordem TPR, FPR.
// curva que vai de 0 a 1
{ a.create(100*ampl+1,100*ampl+1); a.setTo(Scalar(255,255,255));
  for (int i=0; i<=10; i++) {
    reta(a,10*i*ampl,0,10*i*ampl,a.cols-1,COR(128,128,128));
    putTxt(a,10*i*ampl+1,1,format("%3.1f",(10-i)/10.0));
    reta(a,0,10*i*ampl,a.rows-1,10*i*ampl,COR(128,128,128));
    putTxt(a,a.rows-15,10*i*ampl-24,format("%3.1f",i/10.0));
  }
  putTxt(a,16,1,"tpr");
  putTxt(a,a.rows-15-16,a.cols-30,"fpr");
  // linhas
  // 50 -> 50*ampl
  // 100 -> 0
  // l -> 50*ampl-(l-50)*ampl = (100-l)*ampl
  // colunas
  // 0 -> 0
  // 50 -> 50*ampl
  // c -> c*ampl

  bool inicio=true;
  for (int i=0; i<int(rate.size()); i++) {
    double ld=100*rate[i][0];
    double cd=100*rate[i][1];
    int l=arredonda((100-ld)*ampl);
    int c=arredonda(cd*ampl);
    if (inicio) { moveTo(l,c); inicio=false; }
    else lineTo(a,l,c,cor);
  }
}
*/

void geraTxt(vector<Vec2d>& rate, vector<bool>& positivo, string nomeArq)
// rate na ordem TPR, FPR.
{ FILE* arq=fopen(nomeArq.c_str(),"wt");
  if (!arq) erro("Erro: abertura ",nomeArq);
  fprintf(arq,"%3u %3d\n",unsigned(rate.size()),2);
  fprintf(arq,"//%3s %3s %3s\n","pos","x","y");
  fprintf(arq,"//%3s %3s %3s\n","pos","fpr","tpr");
  for (int i=0; i<int(rate.size()); i++) {
    double yd=100*rate[i][0];
    double xd=100*rate[i][1];
    int y=arredonda(yd);
    int x=arredonda(xd);
    fprintf(arq,"%3d %3d %3d\n",int(positivo[i]),x,y);
  }
  fclose(arq);
}

void matroc(int argc, char **argv)
{ if (argc!=4 && argc!=5 && argc!=6) {
    printf("MatROC: Calcula a curva ROC de casamentos, tendo matriz 3x3\n");
    printf("MatROC match.txt mat33.txt saida.png [dist=val | cor=b,g,r | arquivo=nome.txt]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=dist\n");
    printf("  dist default = 15\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[1]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");
  vector<bool> positivo(k1.size());

  double dist=15;
  GRY blu=0, gre=0, red=255;
  string nomeArq="";
  if (argc>=5) {
    ISTR arq(argc-4,&argv[4]);
    string st=leStr(arq);
    while (st!="eof") {
      if (st=="dist") { 
        leNum(arq,dist); 
        if (dist<=0) erro("Erro: dist fora do intervalo");
      } else if (st=="cor") { 
        leNum(arq,blu); 
        leNum(arq,gre); 
        leNum(arq,red); 
      } else if (st=="arquivo") { 
        nomeArq=leFileName(arq); 
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=leStr(arq);
    }
  }
  COR cor=COR(blu,gre,red);

  Mat_<DBL> mat33; leTransf(mat33,argv[2]); mat33=mat33.inv();
  for (unsigned i=0; i<k1.size(); i++) {
    Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);
    if (norm(t,p1)>dist) positivo[i]=false;
    else positivo[i]=true;
  }

  int pos=0, neg=0;
  for (int i=0; i<int(positivo.size()); i++) {
    if (positivo[i]) pos++;
    else neg++;
  }

  vector<Vec4i> conta(positivo.size()); // tp, fp, fn, tn
  if (positivo[0]) conta[0]=Vec4i(1,0,pos-1,neg);
  else             conta[0]=Vec4i(0,1,pos,neg-1);
  for (int i=1; i<int(conta.size()); i++) {
    conta[i]=conta[i-1];
    if (positivo[i]) { conta[i][0]++; conta[i][2]--; }
    else             { conta[i][1]++; conta[i][3]--; }
  }

  vector<Vec2d> rate(conta.size()); // tpr, fpr
  rate[0][0]=double(conta[0][0])/(conta[0][0]+conta[0][2]);
  rate[0][1]=double(conta[0][1])/(conta[0][1]+conta[0][3]);
  for (int i=1; i<int(rate.size()); i++) {
    rate[i][0]=double(conta[i][0])/(conta[i][0]+conta[i][2]);
    rate[i][1]=double(conta[i][1])/(conta[i][1]+conta[i][3]);
  }

  Mat_<COR> a;
  curvaROC(rate,a,8,cor);
  //curvaROC(rate,a,4,cor);
  imp(a,argv[3]);
  if (nomeArq!="") geraTxt(rate,positivo,nomeArq);
}

//<<<<<<<<<<<<<<<<<<< matacert <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
void curvaAccRej(vector<double>& rate, Mat_<COR>& a, int ampl=4, COR cor=COR(0,0,255), char eixox='r')
{ a.create(100*ampl+1,100*ampl+1); a.setTo(Scalar(255,255,255));

  for (int i=0; i<=10; i++) {
    reta(a,10*i*ampl,0,10*i*ampl,a.cols-1,COR(128,128,128));
    putTxt(a,10*i*ampl+1,1,format("%3.1f",(10-i)/10.0));
    reta(a,0,10*i*ampl,a.rows-1,10*i*ampl,COR(128,128,128));
    putTxt(a,a.rows-15,10*i*ampl-24,format("%3.1f",i/10.0));
  }
  putTxt(a,16,1,"tpr");
  if      (eixox=='r') putTxt(a,a.rows-15-16,a.cols-80,"rej. rate");
  else                 putTxt(a,a.rows-15-16,a.cols-80,"acc. rate");

  bool inicio=true;
  for (int i=0; i<int(rate.size()); i++) {
    double ld=100*rate[i];
    double cd=0.0;
    if      (eixox=='r') cd=100.0*(rate.size()-1-i)/rate.size();
    else if (eixox=='a') cd=100.0*i/rate.size();
    else erro("Erro curvaAccRej: eixox invalido");

    int l=arredonda((100-ld)*ampl);
    int c=arredonda(cd*ampl);
    if (inicio) { moveTo(l,c); inicio=false; }
    else lineTo(a,l,c,cor);
  }
}

void matacert(int argc, char **argv)
{ if (argc<4 || 7<argc) {
    printf("MatAcert: Curva de taxa de acerto, tendo matriz 3x3\n");
    printf("MatAcert match.txt mat33.txt saida.png [dist=val | cor=b,g,r | eixox=a/r]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=dist\n");
    printf("  dist default = 15\n");
    printf("  eixox=a: acceptance. eixox=r: rejection.\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[1]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");
  vector<bool> positivo(k1.size());
  char eixox='r'; // r ou a

  double dist=15;
  GRY blu=0, gre=0, red=255;
  if (argc>=5) {
    ISTR arq(argc-4,&argv[4]);
    string st=leStr(arq);
    while (st!="eof") {
      if (st=="dist") { 
        leNum(arq,dist); 
        if (dist<=0) erro("Erro: dist fora do intervalo");
      } else if (st=="cor") { 
        leNum(arq,blu); 
        leNum(arq,gre); 
        leNum(arq,red); 
      } else if (st=="eixox") { 
        string st2=leStr(arq); 
        eixox=st2[0];
        if (!(eixox=='a' || eixox=='r')) erro("Erro: eixox deve ser a ou r");
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=leStr(arq);
    }
  }
  COR cor=COR(blu,gre,red);

  Mat_<DBL> mat33; leTransf(mat33,argv[2]); mat33=mat33.inv();
  for (unsigned i=0; i<k1.size(); i++) {
    Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);
    if (norm(t,p1)>dist) positivo[i]=false;
    else positivo[i]=true;
  }

  vector<double> rate(positivo.size()); // positivo
  if (positivo[0]) rate[0]=1;
  else             rate[0]=0;
  for (int i=1; i<int(rate.size()); i++) {
    if (positivo[i]) rate[i]=rate[i-1]+1;
    else             rate[i]=rate[i-1];
  }
  for (int i=0; i<int(rate.size()); i++) {
    rate[i]=rate[i]/(i+1);
  }

  Mat_<COR> a;
  curvaAccRej(rate,a,4,cor,eixox);
  imp(a,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matlab

void matlab(int argc, char **argv)
{ if (argc!=4 && argc!=5) {
    printf("Matlab: Exporta arquivo Matlab para tracar ROC, tendo matriz 3x3\n");
    printf("Matlab match.txt mat33.txt saida.mat [dist=val]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=dist\n");
    printf("  dist default = 15\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[1]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");
  vector<bool> positivo(k1.size());

  double dist=15;
  if (argc>=5) {
    ISTR arq(argc-4,&argv[4]);
    string st=leStr(arq);
    while (st!="eof") {
      if (st=="dist") { 
        leNum(arq,dist); 
        if (dist<=0) erro("Erro: dist fora do intervalo");
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=leStr(arq);
    }
  }

  Mat_<DBL> mat33; leTransf(mat33,argv[2]); mat33=mat33.inv();
  for (unsigned i=0; i<k1.size(); i++) {
    Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);
    if (norm(t,p1)>dist) positivo[i]=false;
    else positivo[i]=true;
  }

  FILE* sai=fopen(argv[3],"wt");
  if (sai==0) erro("Erro abertura ",argv[3]);
  for (int i=0; i<int(positivo.size()); i++)
    fprintf(sai,"%d ",int(positivo[i]));
  fprintf(sai,"\n");
  fclose(sai);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< matvis

void matvis(int argc, char **argv)
{ if (argc!=5 && argc!=6) {
    printf("MatVis: Mostra um a um os casamentos, tendo matriz 3x3\n");
    printf("MatVis img1.pgm img2.pgm match.txt mat33.txt [a/e/t]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=15\n");
    printf("  a: so acertos, e: so erros, t: todos(default)\n");
    erro("Erro: Numero de argumentos invalido");
  }
  printf("Aperte ESC para sair\n");

  Mat_<GRY> img1; le(img1,argv[1]);
  Mat_<GRY> img2; le(img2,argv[2]);

  vector<DMatch> matches;
  vector<KeyPoint> k1,k2;
  leMatch(matches,k1,k2,argv[3]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");

  double dist=15;
  Mat_<DBL> mat33; leTransf(mat33,argv[4]); mat33=mat33.inv();

  char modo='t';
  if (argc==6) modo=tolower(argv[5][0]);
  if (modo!='a' && modo!='e' && modo!='t') erro("Erro: Modo invalido");

  namedWindow("janela",CV_WINDOW_NORMAL);
  Mat_<COR> b; converte(grudaV(img1,img2),b);
  for (unsigned i=0; i<k1.size(); i++) {
    Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);

    bool acerto=true;
    if (norm(t,p1)>dist) acerto=false;

    COR cor;
    if (acerto) cor=COR(0,255,0);
    else cor=COR(0,0,255);

    if (modo=='t' || (acerto==true && modo=='a') || (acerto==false && modo=='e')) {
      Mat_<COR> d=b.clone();
      relogioCircular(d,arredonda(k1[i].pt.y),arredonda(k1[i].pt.x),arredonda(k1[i].size/2.0),k1[i].angle-90,cor,2);
      relogioCircular(d,arredonda(k2[i].pt.y+img1.rows),arredonda(k2[i].pt.x),arredonda(k2[i].size/2.0),k2[i].angle-90,cor,2);
      imshow("janela",d);
      if (waitKey(0)==27) break;
    }
  }
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< sobrroc

void sobrroc(int argc, char** argv)
{ if (argc<3) {
    printf("SobrROC sai.ppm r1.png r2.png ...\n");
    erro("Erro: Numero de argumentos invalido");
  }
  vector<string> vs; 
  for (int i=2; i<argc; i++)
    vsWildCard(argv[i],vs);

  Mat_<COR> sai; le(sai,vs[0]);
  for (unsigned i=1; i<vs.size(); i++) {
    Mat_<COR> ent; le(ent,vs[i]);
    if (sai.size()!=ent.size()) 
      erro("Erro: Imagens com tamanhos diferentes");
    for (int l=0; l<ent.rows; l++)
      for (int c=0; c<ent.cols; c++)
        if (ent(l,c)!=COR(255,255,255)) sai(l,c)=ent(l,c);
  }
  imp(sai,argv[1]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<< filtmat

void filtmat(int argc, char **argv)
{ if (argc!=3 && argc!=4 && argc!=5) {
    printf("FiltMat: Calcula taxa de erro/acerto de casamento, tendo matriz 3x3\n");
    printf("         Filtra casamentos errados\n");
    printf("FiltMat match.txt mat33.txt [dist] [filtmat.txt]\n");
    printf("  Acerto: quando distancia(inv(mat33)*k2,k1)<=dist\n");
    printf("  dist default = 15\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<DMatch> matches;
  vector<KeyPoint> k1;
  vector<KeyPoint> k2;
  leMatch(matches,k1,k2,argv[1]);
  if (matches.size()!=k1.size() || matches.size()!=k2.size()) erro("Erro: Dimensoes conflitantes");

  double dist=15;
  if (argc>=4) convArg(dist,argv[3]);

  Mat_<DBL> mat33; leTransf(mat33,argv[2]); mat33=mat33.inv();
  int erros=0, acertos=0;
  Mat_<DBL> p1(3,1), p2(3,1), t(3,1);
  vector<bool> vet(matches.size());
  for (unsigned i=0; i<k1.size(); i++) {
    p1(0)=k1[i].pt.x; p1(1)=k1[i].pt.y; p1(2)=1;
    p2(0)=k2[i].pt.x; p2(1)=k2[i].pt.y; p2(2)=1;
    t=mat33*p2;
    t(0)=t(0)/t(2); t(1)=t(1)/t(2); t(2)=t(2)/t(2);
    if (norm(t,p1)>dist) { erros++; vet[i]=false; }
    else { acertos++; vet[i]=true; }
  }
  int total=erros+acertos;
  printf("Total=%d erros=%d(%4.1f%%) acertos=%d(%4.1f%%)\n",total,erros,100.0*erros/total,acertos,100.0*acertos/total);
  if (argc>=5) impMatch(matches,k1,k2,argv[4],vet);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< DESCRIPT: Teste dos descritores sift, surf, riff e simple de OpenCV, v1.20>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"KeySift   - Gera keypoints usando SIFT\n"
"KeySurf   - Gera keypoints usando SURF\n"
"KeyFast   - Gera keypoints usando FAST\n"
"KeyStar   - Gera keypoints usando STAR\n"
"KeyRand   - Gera keypoints aleatoriamente\n"
"KeyCgSize - Muda o tamanho dos keypoints por fator\n"
"KeyGrande - Seleciona n maiores keypoints\n"
"KeyShow   - Mostra keypoints\n"
"KeyMyShow - Mostra keypoints sem chamar funcao OpenCV\n"
"DesSift   - Calcula descriptors de SIFT (128)\n"
"DesSurf   - Calcula descriptors de SURF (64)\n"
"DesSimp   - Calcula descriptors de SIMPLE (127)\n"
"DesRiff   - Calcula descriptors de RIFF (100)\n"
"DesRihog  - Calcula descriptors de RIHOG (80)\n"
"DesRuhog  - Calcula descriptors de RUHOG (80) com aneis de tamanho variavel\n"
"DesForapro- Calcula descriptors Forapro\n"
"MatNN     - Acha casamentos de des1 para des2 usando 1-NN\n"
"MatRatio  - Acha casamentos de des1 para des2 onde 1st/2st>ratio ou acRate\n"
"MatShow   - Mostra os casamentos\n"
"MatMyShow - Mostra os casamentos sem chamar funcao OpenCV\n"
//"MatRate   - Calcula a taxa de erros/acertos de casamentos, tendo matriz 3x3\n"
"MatROC    - Calcula a curva ROC de casamentos, tendo matriz 3x3\n"
"MatAcert  - Calcula a curva de acertos de casamentos, tendo matriz 3x3\n"
"Matlab    - Exporta arquivo para Matlab tracar ROC, tendo matriz 3x3\n"
"MatVis    - Visualiza um a um os casamentos, tendo matriz 3x3\n"
"SobrROC   - Sobrepoe ROCs\n"
"FiltMat   - Calcula a taxa de acertos de casamentos, tendo matriz 3x3. Filtra.\n"
"...............................................................................\n";

int main(int argc, char** argv)
{
  if (argc<2) {
    printf(about);
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="keysift")   keysift(argc-1,&argv[1]);
    else if (comando=="keysurf")   keysurf(argc-1,&argv[1]);
    else if (comando=="keyfast")   keyfast(argc-1,&argv[1]);
    else if (comando=="keystar")   keystar(argc-1,&argv[1]);
    else if (comando=="keyrand")   keyrand(argc-1,&argv[1]);
    else if (comando=="keycgsize") keycgsize(argc-1,&argv[1]);
    else if (comando=="keygrande") keygrande(argc-1,&argv[1]);
    else if (comando=="keyshow")   keyshow(argc-1,&argv[1]);
    else if (comando=="keymyshow") keymyshow(argc-1,&argv[1]);
    else if (comando=="dessift")   dessift(argc-1,&argv[1]);
    else if (comando=="dessurf")   dessurf(argc-1,&argv[1]);
    else if (comando=="dessimp")   dessimp(argc-1,&argv[1]);
    else if (comando=="desriff")   desriff(argc-1,&argv[1]);
    else if (comando=="desrihog")  desrihog(argc-1,&argv[1]);
    else if (comando=="desruhog")  desruhog(argc-1,&argv[1]);
    else if (comando=="desforapro")desforapro(argc-1,&argv[1]);
    else if (comando=="matnn")     matnn(argc-1,&argv[1]);
    else if (comando=="matratio")  matratio(argc-1,&argv[1]);
    else if (comando=="matshow")   matshow(argc-1,&argv[1]);
    else if (comando=="matmyshow") matmyshow(argc-1,&argv[1]);
    //else if (comando=="matrate")   matrate(argc-1,&argv[1]);
    else if (comando=="matroc")    matroc(argc-1,&argv[1]);
    else if (comando=="matacert")  matacert(argc-1,&argv[1]);
    else if (comando=="matlab")    matlab(argc-1,&argv[1]);
    else if (comando=="matvis")    matvis(argc-1,&argv[1]);
    else if (comando=="sobrroc")   sobrroc(argc-1,&argv[1]);
    else if (comando=="filtmat")   filtmat(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto=%d.%02ds\n",t1/100,t1%100);
  }
}


