#include <cekeikon.h>
#include <map>

/* Nota: Circulo de diametro 3:
kernel:
7 7
0 0 0 0 0 0 0
0 0.0625 0.0625 0.0625 0.0625 0.0625 0
0 0.0625 -0.111111 -0.111111 -0.111111 0.0625 0
0 0.0625 -0.111111 -0.111111 -0.111111 0.0625 0
0 0.0625 -0.111111 -0.111111 -0.111111 0.0625 0
0 0.0625 0.0625 0.0625 0.0625 0.0625 0
0 0 0 0 0 0 0

gray:
5 5
255 255 255 255 255
255   0   0   0 255
255   0   0   0 255
255   0   0   0 255
255 255 255 255 255

borda:
5 5
255 255 255 255 255
255   0   0   0 255
255   0 255   0 255
255   0   0   0 255
255 255 255 255 255

*/

struct FORMA {
  char forma; // c, q, r, ou e
  double d1, d2, ang; // lado1, lado2, angulo em graus
  double esc; // escala da forma (0.5<esc<=1.0). esc=pow2(double(-e)/escOitava)
  double escPir; // escala maxima que deve procurar esta forma na piramide
  int sPir; // andar maximo do piramide. sPir=chao(log2(escPir/esc)+epsilon). Veja ideias_de_granul.docx
  int inic; // 0 no inicio, 1 no meio, 2 no fim da especificacao em .cfg
  Mat_<FLT> f; // kernel em FLT
  Mat_<GRY> b; // borda
  Mat_<GRY> g; // gray 0, 255 ou 128
  //string nome;
};

struct MAXLOCAL {
  FLT r; // correlacao (entre -1 e +1)
  char forma; // c, q, r, ou e. Igual da FORMA
  int li,ci;
  double d1, d2, ang; // lado1, lado2, angulo em graus. lado1 e lado2 ja vem multiplicados pela escala.

  MAXLOCAL() {}
  MAXLOCAL(FLT pr, char pforma, int pli, int pci, double pd1, double pd2, double pang)
    { r=pr; forma=pforma; li=pli; ci=pci; d1=pd1; d2=pd2; ang=pang; }
};

void imp(const vector<MAXLOCAL>& v, string st)
{ FILE* arq=fopen(st.c_str(),"wt");
  if (arq==NULL) erro("Erro abertura arquivo ",st);
  fprintf(arq,"%u\n",unsigned(v.size()));
  fprintf(arq,"%8s %1s %4s %4s %8s %8s %3s\n","//correl","f","l","c","d1","d2","deg");
  for (unsigned i=0; i<v.size(); i++) {
    const MAXLOCAL& m=v[i];
    fprintf(arq,"%8.6f %c %4d %4d %8.5f %8.5f %3.0f\n",m.r,m.forma,m.li,m.ci,m.d1,m.d2,m.ang);
  }
  fclose(arq);
}

void imp(MAXLOCAL m)
{ printf("%f %c %d %d %g %g %g\n",m.r,m.forma,m.li,m.ci,m.d1,m.d2,m.ang); }

void le(vector<MAXLOCAL>& v, string st)
{ FILE* arq=fopen(st.c_str(),"rt");
  if (arq==NULL) erro("Erro abertura arquivo ",st);
  int n; leNum(arq,n); v.reserve(n);
  MAXLOCAL m; string st2;
  for (int i=0; i<n; i++) {
    leNum(arq,m.r);
    st2=leStr(arq); if (st2.size()==0) erro("Erro leitura forma"); m.forma=st2[0];
    leNum(arq,m.li);
    leNum(arq,m.ci);
    leNum(arq,m.d1);
    leNum(arq,m.d2);
    leNum(arq,m.ang);
    v.push_back(m);
  }
  fclose(arq);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
void NegPosUm(Mat_<FLT>& f)
// Faz a somatoria dos negativos dar -1 e dos positivos dar +1.
{ double pos=0.0;
  double neg=0.0;
  for (MatIterator_<FLT> fi=f.begin(); fi!=f.end(); fi++) {
    if (*fi>0.0) pos += *fi;
    else neg -= *fi;
  }
  if (pos<=0.0 || neg<=0.0) erro("Erro: NegPosUm divisao zero");
  for (MatIterator_<FLT> fi=f.begin(); fi!=f.end(); fi++) {
    if (*fi>0.0) *fi /= pos;
    else *fi /= neg;
  }
}

void borda(Mat_<GRY> pmas, Mat_<GRY>& bor)
{ IMG_<GRY> mas(pmas); mas.backg=255;
  bor.create(mas.rows,mas.cols); bor.setTo(255);
  for (int l=0; l<mas.rows; l++)
    for (int c=0; c<mas.cols; c++)
      if ( mas(l,c)==0 && (mas(l-1,c)>0 || mas(l+1,c)>0 || mas(l,c-1)>0 || mas(l,c+1)>0) )
        bor(l,c)=0;
}

void gray1(Mat_<FLT> mas, Mat_<GRY>& gry)
{ gry.create(mas.size());
  for (int l=0; l<mas.rows; l++)
    for (int c=0; c<mas.cols; c++)
      if (mas(l,c)<0) gry(l,c)=0;
      else if (mas(l,c)>0) gry(l,c)=255;
      else if (mas(l,c)==0) gry(l,c)=128;
}

//<<<<<<<<<<<<<<<<<<<<<< bordas
void bordaCirculo(double diametro, Mat_<GRY>& bor)
{ int d=arredonda(diametro);
  Mat_<GRY> mas;
  if (d>=5) {
    int n=ceilImpar(diametro)+2;
    mas.create(n,n); mas.setTo(255);
    int centro=n/2;
    circulo(mas,centro,centro,diametro/2.0,0,-1);
  } else if (d==4) {
    GRY vkernel[]={1, 1, 1, 1, 1, 1, 1,
                   1, 1, 1, 0, 1, 1, 1,
                   1, 1, 0, 0, 0, 1, 1,
                   1, 0, 0, 0, 0, 0, 1,
                   1, 1, 0, 0, 0, 1, 1,
                   1, 1, 1, 0, 1, 1, 1,
                   1, 1, 1, 1, 1, 1, 1};
    Mat_<GRY> kernel(7,7,vkernel);
    mas=255*kernel;
  } else if (d==3) {
    GRY vkernel[]={1, 1, 1, 1, 1,
                   1, 0, 0, 0, 1,
                   1, 0, 0, 0, 1,
                   1, 0, 0, 0, 1,
                   1, 1, 1, 1, 1};
    Mat_<GRY> kernel(5,5,vkernel);
    mas=255*kernel;
  } else if (d==2) {
    GRY vkernel[]={1, 1, 1, 1, 1,
                   1, 1, 0, 1, 1,
                   1, 0, 0, 0, 1,
                   1, 1, 0, 1, 1,
                   1, 1, 1, 1, 1};
    Mat_<GRY> kernel(5,5,vkernel);
    mas=255*kernel;
  } else if (d==1) {
    GRY vkernel[]={ 1, 1, 1,
                    1, 0, 1,
                    1, 1, 1};
    Mat_<GRY> kernel(3,3,vkernel);
    mas=255*kernel;
  }
  borda(mas,bor);
}

void bordaQuadrado(double lado, double ang, Mat_<GRY>& bor)
{ int n=ceilImpar(sqrt(2.0)*lado);
  Mat_<GRY> mas(n,n,255);
  int centro=n/2;
  double eixo=lado/2;
  rotRetang(mas, centro, centro, eixo, eixo, 0, -1, ang);
  borda(mas,bor);
}

void bordaRetangulo(double ladol, double ladoc, double ang, Mat_<GRY>& bor)
{ int n=ceilImpar(sqrt(2.0)*max(ladol,ladoc));
  Mat_<GRY> mas(n,n,255);
  int centro=n/2;
  double eixol=ladol/2;
  double eixoc=ladoc/2;
  rotRetang(mas, centro, centro, eixol, eixoc, 0, -1, ang);
  borda(mas,bor);
}

void bordaElipse(double ladol, double ladoc, double ang, Mat_<GRY>& bor)
{ int n=ceilImpar(max(ladol,ladoc));
  Mat_<GRY> mas(n,n,255);
  int centro=n/2;
  double eixol=ladol/2;
  double eixoc=ladoc/2;
  elipse(mas, centro, centro, eixol, eixoc, 0, -1, ang);
  borda(mas,bor);
}

//<<<<<<<<<<<<<<<<<<<<<< grays
void grayCirculo(double diametro, Mat_<GRY>& gry)
{ int d=arredonda(diametro);
  if (d>=5) {
    int n=ceilImpar(diametro)+2;
    gry.create(n,n); gry.setTo(255);
    int centro=n/2;
    circulo(gry,centro,centro,diametro/2.0,0,-1);
  } else if (d==4) {
    GRY vkernel[]={1, 1, 1, 1, 1, 1, 1,
                   1, 1, 1, 0, 1, 1, 1,
                   1, 1, 0, 0, 0, 1, 1,
                   1, 0, 0, 0, 0, 0, 1,
                   1, 1, 0, 0, 0, 1, 1,
                   1, 1, 1, 0, 1, 1, 1,
                   1, 1, 1, 1, 1, 1, 1};
    Mat_<GRY> kernel(7,7,vkernel);
    gry=255*kernel;
  } else if (d==3) {
    GRY vkernel[]={1, 1, 1, 1, 1,
                   1, 0, 0, 0, 1,
                   1, 0, 0, 0, 1,
                   1, 0, 0, 0, 1,
                   1, 1, 1, 1, 1};
    Mat_<GRY> kernel(5,5,vkernel);
    gry=255*kernel;
  } else if (d==2) {
    GRY vkernel[]={1, 1, 1, 1, 1,
                   1, 1, 0, 1, 1,
                   1, 0, 0, 0, 1,
                   1, 1, 0, 1, 1,
                   1, 1, 1, 1, 1};
    Mat_<GRY> kernel(5,5,vkernel);
    gry=255*kernel;
  } else if (d==1) {
    GRY vkernel[]={ 1, 1, 1,
                    1, 0, 1,
                    1, 1, 1};
    Mat_<GRY> kernel(3,3,vkernel);
    gry=255*kernel;
  }
}

void grayQuadrado(double lado, double ang, Mat_<GRY>& gry)
{ int n=ceilImpar(sqrt(2.0)*lado);
  gry.create(n,n); gry.setTo(255);
  int centro=n/2;
  double eixo=lado/2;
  rotRetang(gry, centro, centro, eixo, eixo, 0, -1, ang);
}

void grayRetangulo(double ladol, double ladoc, double ang, Mat_<GRY>& gry)
{ int n=ceilImpar(sqrt(2.0)*max(ladol,ladoc));
  gry.create(n,n); gry.setTo(255);
  int centro=n/2;
  double eixol=ladol/2;
  double eixoc=ladoc/2;
  rotRetang(gry, centro, centro, eixol, eixoc, 0, -1, ang);
}

void grayElipse(double ladol, double ladoc, double ang, Mat_<GRY>& gry)
{ int n=ceilImpar(max(ladol,ladoc));
  gry.create(n,n); gry.setTo(255);
  int centro=n/2;
  double eixol=ladol/2;
  double eixoc=ladoc/2;
  elipse(gry, centro, centro, eixol, eixoc, 0, -1, ang);
}

void gray2(MAXLOCAL m, Mat_<GRY>& g)
{ if      (m.forma=='c') grayCirculo(m.d1,g);
  else if (m.forma=='q') grayQuadrado(m.d1,m.ang,g);
  else if (m.forma=='r') grayRetangulo(m.d1,m.d2,m.ang,g);
  else if (m.forma=='e') grayElipse(m.d1,m.d2,m.ang,g);
}

//<<<<<<<<<<<<<<<< formas <<<<<<<<<<<<<<<<<<<<<<<<<<<
void kerCirculo(double diametro, Mat_<FLT>& mas)
{
  int d=arredonda(diametro);
  if (d>=5) {
    int n=ceilImpar(sqrt(2.0)*diametro)+2;
    mas.create(n,n); mas.setTo(0.0);
    int centro=n/2;
    circulo(mas,centro,centro,sqrt(2.0)*diametro/2.0,+1.0,-1);
    circulo(mas,centro,centro,diametro/2.0,-1.0,-1);
  } else if (d==4) {
    FLT vkernel[]={ 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 1, 0, 0, 0, 0,
                    0, 0, 0, 1,-1, 1, 0, 0, 0,
                    0, 0, 1,-1,-1,-1, 1, 0, 0,
                    0, 1,-1,-1,-1,-1,-1, 1, 0,
                    0, 0, 1,-1,-1,-1, 1, 0, 0,
                    0, 0, 0, 1,-1, 1, 0, 0, 0,
                    0, 0, 0, 0, 1, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0};
    Mat_<FLT> kernel(9,9,vkernel);
    mas=kernel.clone();
  } else if (d==3) {
    FLT vkernel[]={ 0, 0, 0, 0, 0, 0, 0,
                    0, 1, 1, 1, 1, 1, 0,
                    0, 1,-1,-1,-1, 1, 0,
                    0, 1,-1,-1,-1, 1, 0,
                    0, 1,-1,-1,-1, 1, 0,
                    0, 1, 1, 1, 1, 1, 0,
                    0, 0, 0, 0, 0, 0, 0};
    Mat_<FLT> kernel(7,7,vkernel);
    mas=kernel.clone();
  } else if (d==2) {
    FLT vkernel[]={ 0, 0, 1, 0, 0,
                    0, 1,-1, 1, 0,
                    1,-1,-1,-1, 1,
                    0, 1,-1, 1, 0,
                    0, 0, 1, 0, 0};
    Mat_<FLT> kernel(5,5,vkernel);
    mas=kernel.clone();
  } else if (d==1) {
    FLT vkernel[]={ 1, 1, 1,
                    1,-1, 1,
                    1, 1, 1};
    Mat_<FLT> kernel(3,3,vkernel);
    mas=kernel.clone();
  }
  NegPosUm(mas);
}

void kerQuadrado(double lado, double ang, Mat_<FLT>& mas)
// ang em degrees
{ int n=ceilImpar(2.0*lado);
  mas.create(n,n); mas.setTo(0.0);
  int centro=n/2;

  double eixo2=sqrt(2.0)*lado/2;
  rotRetang(mas, centro, centro, eixo2, eixo2, +1.0, -1, ang);
  double eixo1=lado/2;
  rotRetang(mas, centro, centro, eixo1, eixo1, -1.0, -1, ang);
  NegPosUm(mas);
}

void kerRetangulo(double ladol, double ladoc, double ang, Mat_<FLT>& mas)
// ang em degrees
{ int n=ceilImpar(2.0*max(ladol,ladoc));
  mas.create(n,n); mas.setTo(0.0);
  int centro=n/2;

  double eixo2l=sqrt(2.0)*ladol/2;
  double eixo2c=sqrt(2.0)*ladoc/2;
  rotRetang(mas, centro, centro, eixo2l, eixo2c, +1.0, -1, ang);
  double eixo1l=ladol/2;
  double eixo1c=ladoc/2;
  rotRetang(mas, centro, centro, eixo1l, eixo1c, -1.0, -1, ang);
  NegPosUm(mas);
}

void kerElipse(double ladol, double ladoc, double ang, Mat_<FLT>& mas)
// ang em degrees
{ int n=ceilImpar(sqrt(2.0)*max(ladol,ladoc))+2;
  mas.create(n,n); mas.setTo(0.0);
  int centro=n/2;

  double eixo2l=sqrt(2.0)*ladol/2.0;
  double eixo2c=sqrt(2.0)*ladoc/2.0;
  elipse(mas, centro, centro, eixo2l, eixo2c, +1.0, -1, ang);
  double eixo1l=ladol/2.0;
  double eixo1c=ladoc/2.0;
  elipse(mas, centro, centro, eixo1l, eixo1c, -1.0, -1, ang);
  NegPosUm(mas);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
void copia(FORMA& a, FORMA& b)
{ b=a;
  b.f=a.f.clone();
  b.b=a.b.clone();
  b.g=a.g.clone();
}

void constroiKernel(string nome, vector<FORMA>& ker)
{ FILE* arq=fopen(nome.c_str(),"rt");
  if (arq==NULL) erro("Erro abertura arquivo ",nome);
  string st;
  double passoAng=10; int escOitava=5; double escMin=0.5, escMax=1.0, escPir=4.0;
  FORMA fo,dumb;

  st=minuscula(leStr(arq));
  while (st!="eof") {
    if (st=="passoang") {
      leNum(arq,passoAng);
    } else if (st=="escoitava") {
      leNum(arq,escOitava);
      if (!(1<=escOitava)) erro("Erro: deve ter 1<=escOitava");
    } else if (st=="escala" || st=="escmin") {
      leNum(arq,escMin);
      if (!(0<=escMin)) erro("Erro: deve ter 0<=escMin");
    } else if (st=="a" || st=="escmax") {
      leNum(arq,escMax);
      if (!(escMin<=escMax)) erro("Erro: deve ter escMin<=escMax");
    } else if (st=="piramide" || st=="escpir") {
      leNum(arq,escPir);
      if (!(escMax<=escPir)) erro("Erro: deve ter escMax<=escPir");
    } else if (st=="circulo") {
      double d; leNum(arq,d);
      double nOitavas=log2(escMax/escMin);
      int nFormas=arredonda(nOitavas*escOitava+1);
      for (int e=0; e<nFormas; e++) {
        double esc=escMax*pow2(double(-e)/escOitava);
        // Assim, a menor escala (maior e)  aproximadamente escMin.
        fo.forma='c'; fo.d2=fo.d1=esc*d; fo.ang=0; fo.esc=esc;
        if (e<escOitava) {  fo.escPir=escPir; fo.sPir=chao(log2(escPir/esc)+1e-6); }
        else { fo.escPir=esc; fo.sPir=0; }
        if (e==0) fo.inic=0; else if (e==nFormas-1) fo.inic=2; else fo.inic=1;
        kerCirculo(fo.d1,fo.f); bordaCirculo(fo.d1,fo.b); grayCirculo(fo.d1,fo.g);
        //fo.nome=format("c_%03d",arredonda(fo.d1));
        ker.push_back(dumb); copia(fo,ker.back());
      }
    } else if (st=="quadrado") {
      double d; leNum(arq,d);
      double nOitavas=log2(escMax/escMin);
      int nFormas=arredonda(nOitavas*escOitava+1);
      for (double a=0; a<90; a += passoAng) {
        for (int e=0; e<nFormas; e++) {
          double esc=escMax*pow2(double(-e)/escOitava);
          // Assim, a menor escala (maior e)  aproximadamente escMin.
          fo.forma='q'; fo.d2=fo.d1=esc*d; fo.ang=a; fo.esc=esc;
          if (e<escOitava) { fo.escPir=escPir; fo.sPir=chao(log2(escPir/esc)+1e-6); }
          else { fo.escPir=esc; fo.sPir=0; }
          if (e==0) fo.inic=0; else if (e==nFormas-1) fo.inic=2; else fo.inic=1;
          kerQuadrado(fo.d1,a,fo.f); bordaQuadrado(fo.d1,a,fo.b); grayQuadrado(fo.d1,a,fo.g);
          //fo.nome=format("q_%03d_%03d",arredonda(fo.d1),arredonda(fo.ang));
          ker.push_back(dumb); copia(fo,ker.back());
        }
      }
    } else if (st=="retangulo") {
      double d1,d2; leNum(arq,d1); leNum(arq,d2);
      double nOitavas=log2(escMax/escMin);
      int nFormas=arredonda(nOitavas*escOitava+1);
      for (double a=0; a<180; a += passoAng) {
        for (int e=0; e<nFormas; e++) {
          double esc=escMax*pow2(double(-e)/escOitava);
          // Assim, a menor escala (maior e)  aproximadamente escMin.
          fo.forma='r'; fo.d1=esc*d1; fo.d2=esc*d2; fo.ang=a; fo.esc=esc;
          if (e<escOitava) { fo.escPir=escPir; fo.sPir=chao(log2(escPir/esc)+1e-6); }
          else { fo.escPir=esc; fo.sPir=0; }
          if (e==0) fo.inic=0; else if (e==nFormas-1) fo.inic=2; else fo.inic=1;
          kerRetangulo(fo.d1,fo.d2,a,fo.f); bordaRetangulo(fo.d1,fo.d2,a,fo.b); grayRetangulo(fo.d1,fo.d2,a,fo.g);
          //fo.nome=format("r_%03d_%03d_%03d",arredonda(fo.d1),arredonda(fo.d2),arredonda(fo.ang));
          ker.push_back(dumb); copia(fo,ker.back());
        }
      }
    } else if (st=="elipse") {
      double d1,d2; leNum(arq,d1); leNum(arq,d2);
      double nOitavas=log2(escMax/escMin);
      int nFormas=arredonda(nOitavas*escOitava+1);
      for (double a=0; a<180; a += passoAng) {
        for (int e=0; e<nFormas; e++) {
          double esc=escMax*pow2(double(-e)/escOitava);
          // Assim, a menor escala (maior e)  aproximadamente escMin.
          fo.forma='e'; fo.d1=esc*d1; fo.d2=esc*d2; fo.ang=a; fo.esc=esc;
          if (e<escOitava) { fo.escPir=escPir; fo.sPir=chao(log2(escPir/esc)+1e-6); }
          else { fo.escPir=esc; fo.sPir=0; }
          if (e==0) fo.inic=0; else if (e==nFormas-1) fo.inic=2; else fo.inic=1;
          kerElipse(fo.d1,fo.d2,a,fo.f); bordaElipse(fo.d1,fo.d2,a,fo.b); grayElipse(fo.d1,fo.d2,a,fo.g);
          //fo.nome=format("e_%03d_%03d_%03d",arredonda(fo.d1),arredonda(fo.d2),arredonda(fo.ang));
          ker.push_back(dumb); copia(fo,ker.back());
        }
      }
    } else {
      erro("Erro: Palavra desconhecida ",st);
    }
    st=minuscula(leStr(arq));
  }
  fclose(arq);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
void filtLin(const PIRAMIDE<FLT>& pir, const Mat_<FLT>& ker, PIRAMIDE<FLT>& corr, int maxAndar)
{ cria(pir,corr);
  for (int i=0; i<=maxAndar; i++) {
    filter2D(pir.v[i],corr.v[i],CV_32F,ker,Point(-1,-1),0,BORDER_REPLICATE);
  }
}

void multFiltLin(const PIRAMIDE<FLT>& pir, const vector<FORMA>& ker, int inic, int fim,
                 PIRAMIDE<FLT>& corr, PIRAMIDE<int>& ind)
{ cria(pir,corr,FLT(-1.0)); cria(pir,ind,-1);
  PIRAMIDE<FLT> temp;
  for (int i=inic; i<=fim; i++) {
    int maxAndar=min(ker[i].sPir,int(pir.v.size()-1));
    filtLin(pir,ker[i].f,temp,maxAndar);
    for (int s=0; s<=maxAndar; s++)
      for (int l=0; l<temp.v[s].rows; l++)
        for (int c=0; c<temp.v[s].cols; c++)
          if (temp.v[s](l,c)>corr.v[s](l,c)) { corr.v[s](l,c)=temp.v[s](l,c); ind.v[s](l,c)=i; }
  }
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void findMAXLOCAL(PIRAMIDE<FLT>& corr, PIRAMIDE<int>& ind, vector<FORMA>& ker,
                  vector<MAXLOCAL>& v, double minCorr)
{
  for (unsigned s=0; s<corr.v.size(); s++) {
    for (int l=1; l<corr.v[s].rows-1; l++)
      for (int c=1; c<corr.v[s].cols-1; c++) {
        FLT valor=corr.v[s](l,c);
        if (valor<minCorr) goto sai1;
        for (int l2=-1; l2<=1; l2++)
          for (int c2=-1; c2<=1; c2++)
            if (valor<corr.v[s](l+l2,c+c2)) goto sai1;
        // E' maximo local (extrema)
        {
          FORMA& f=ker[ind.v[s](l,c)];
          int li=0,ci=0; bool dentro=ind.p2i(l,c,s,li,ci);
          if (!dentro) erro("Erro findMaxLocal: Nao dentro");
          double esc=pow2(int(s)); double d1=esc*f.d1; double d2=esc*f.d2;
          v.push_back(MAXLOCAL(valor,f.forma,li,ci,d1,d2,f.ang));
        }
        sai1:;
      }
  }
}

void mergeMAXLOCAL(const vector<MAXLOCAL>& u, vector<MAXLOCAL>& v)
// Gruda u no final de v, sem verificar nada
{ for (unsigned i=0; i<u.size(); i++) {
    v.push_back(u[i]);
  }
}

typedef multimap<int,MAXLOCAL> MMAPA;
typedef MMAPA::iterator IMMAPA;

/*
void imp(MAXLOCAL m)
{ printf("(%c,%d,%d)\n",m.forma,m.li,m.ci);
}

void imp(vector<MAXLOCAL>& v)
{ for (unsigned i=0; i<v.size(); i++) {
    printf("(%c,%d,%d) ",v[i].forma,v[i].li,v[i].ci);
  }
  printf("\n");
}

void imp(MMAPA& m)
{ for (IMMAPA p=m.begin(); p!=m.end(); p++) {
    printf("(%c,%d,%d) ",p->second.forma,p->second.li,p->second.ci);
  }
  printf("\n");
}

void imp(MMAPA& m, IMMAPA l, IMMAPA r)
{ for (IMMAPA p=l; p!=r; p++) {
    printf("(%c,%d,%d) ",p->second.forma,p->second.li,p->second.ci);
  }
  printf("\n");
}
*/

void ordena(vector<MAXLOCAL>& a, int l, int r)
{ assert(l<=r);
  int i=l; int j=r;
  MAXLOCAL x=a[(l+r)/2];
  do {
    while (a[i].r>x.r) i++;
    while (x.r>a[j].r) j--;
    if (i<=j) { swap(a[i],a[j]); i++; j--; }
  } while (i<=j);
  if (l<j) ordena(a,l,j);
  if (i<r) ordena(a,i,r);
}

void eliminaProximo(vector<MAXLOCAL>& v, vector<MAXLOCAL>& u, int maxDist)
// Elimina MAXLOCAL muito proximos de v
// Na entrada, v esta ordenado decrescente pela correlacao
// Na saida, u esta ordenado decrescente pela correlacao
{ MMAPA m;
  for (unsigned i=0; i<v.size(); i++) {
    IMMAPA low=m.lower_bound(v[i].ci-maxDist);
    IMMAPA up=m.upper_bound(v[i].ci+maxDist);
    for (IMMAPA p=low; p!=up; p++) {
      if (v[i].forma==p->second.forma && abs(v[i].li - p->second.li)<=maxDist) {
        assert(abs(v[i].ci-p->second.ci)<=maxDist);
        goto saida;
      }
    }
    m.insert(make_pair(v[i].ci,v[i]));
    saida:;
  }
  u.clear();
  for (IMMAPA p=m.begin(); p!=m.end(); p++)
    u.push_back(p->second);
  ordena(u,0,u.size()-1);
}
// eliminaProximo que nao usava multimap e demorava
//{ if (u.size()>0) u.clear();
//  for (unsigned i=0; i<v.size(); i++) {
//    for (unsigned j=0; j<u.size(); j++)
//      if (v[i].forma==u[j].forma && abs(v[i].li-u[j].li)<=maxDist && abs(v[i].ci-u[j].ci)<=maxDist)
//        goto saida;
//    u.push_back(v[i]);
//    saida:;
//  }
//}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
void kernel(int argc, char** argv)
{ if (argc!=3) {
    printf("Kernel kernel.cfg kernel.pgm\n");
    erro("Erro: Numero de argumentos invalido");
  }

  vector<FORMA> ker;
  constroiKernel(argv[1],ker);
  printf("Numero de kernels=%u\n",unsigned(ker.size()));
  for (unsigned i=0; i<ker.size(); i++) {
    printf("forma=%c d1=%g d2=%g ang=%g",
            ker[i].forma,ker[i].d1,ker[i].d2,ker[i].ang);
    printf(" esc=%g escPir=%g sPir=%d inic=%d",
            ker[i].esc,ker[i].escPir,ker[i].sPir,ker[i].inic);
    printf("\n");
  }
  //printf("\n");
  Mat_<GRY> ver,hor,g;
  for (unsigned i=0; i<ker.size(); i++) {
    gray1(ker[i].f,g);
    if (ker[i].inic==2) {
      ver=grudaV(ver,g,1,0);
      hor=grudaH(hor,ver,1,0);
      ver.create(0,0);
    } else ver=grudaV(ver,g,1,0);
  }
  /*
  for (unsigned i=0; i<ker.size(); i++) {
    if (ker[i].inic==2) {
      ver=grudaV(ver,ker[i].g,1,0);
      hor=grudaH(hor,ver,1,0);
      ver.create(0,0);
    } else ver=grudaV(ver,ker[i].g,1,0);
  }
  for (unsigned i=0; i<ker.size(); i++) {
    if (ker[i].inic==2) {
      ver=grudaV(ver,ker[i].b,1,0);
      hor=grudaH(hor,ver,1,0);
      ver.create(0,0);
    } else ver=grudaV(ver,ker[i].b,1,0);
  }
  */
  imp(hor,argv[2]);
}

void correla(int argc, char** argv)
{ if (argc<4 || 5<argc) {
    printf("Correla imagem.pgm kernel.cfg imagem.ho1 [minCorr] [maxDist]\n");
    printf("  maxLocal com correlacao<minCorr e' eliminado\n");
    printf("  um maxLocal com distancia<=maxDist de outro maxLocal e' eliminado\n");
    printf("  Default: minCorr=0.001 maxDist=2\n");
    erro("Erro: Numero de argumentos invalido");
  }

  double minCorr=0.001;
  if (argc>=5) convArg(minCorr,argv[4]);

  int maxDist=2;
  if (argc>=6) convArg(maxDist,argv[5]);

  printf("Lendo %s e construindo piramide...\n",argv[1]);
  Mat_<GRY> ent; le(ent,argv[1]);
  Mat_<FLT> entf; converte(ent,entf);
  PIRAMIDE<FLT> pir(entf,4);

  printf("Lendo %s e construindo kernels...\n",argv[2]);
  vector<FORMA> ker;
  constroiKernel(argv[2],ker);
  printf("Numero de kernels=%u\n",unsigned(ker.size()));

  printf("Calculando correlacoes em piramide e achando maximos locais...\n");
  PIRAMIDE<FLT> corr; PIRAMIDE<int> ind;
  vector<MAXLOCAL> u,v;
  int inic=0;
  for (unsigned i=0; i<ker.size(); i++) {
    if (ker[i].inic==2) {
      printf("Aplicando filtro #%u-#%u\r",inic,i);
      multFiltLin(pir,ker,inic,i,corr,ind); inic=i+1;
      findMAXLOCAL(corr,ind,ker,u,minCorr);
      mergeMAXLOCAL(u,v);
      u.clear();
    }
  }
  printf("\n");

  printf("Ordenando os %u maximos locais em ordem decrescente de correlacao...\n",unsigned(v.size()));
  ordena(v,0,v.size()-1);

  printf("Eliminando os maximos locais muito proximos...\n");
  eliminaProximo(v,u,maxDist);

  printf("Imprimindo os %u maximos locais\n",unsigned(u.size()));
  imp(u,argv[3]);
}

void oldcorrela(int argc, char** argv)
{ if (argc<4 || 5<argc) {
    printf("OldCorrela imagem.pgm kernel.cfg imagem.ho1 [minCorr]\n");
    printf("  minCorr: default=0.001\n");
    erro("Erro: Numero de argumentos invalido");
  }

  double minCorr=0.001;
  if (argc>=5) convArg(minCorr,argv[4]);

  printf("Lendo imagem.pgm e construindo piramide...\n");
  Mat_<GRY> ent; le(ent,argv[1]);
  Mat_<FLT> entf; converte(ent,entf);
  PIRAMIDE<FLT> pir(entf);

  printf("Lendo kernel.cfg e construindo kernels...\n");
  vector<FORMA> ker;
  constroiKernel(argv[2],ker);
  printf("Numero de kernels=%u\n",unsigned(ker.size()));

  printf("Calculando correlacoes em piramide...\n");
  PIRAMIDE<FLT> corr; PIRAMIDE<int> ind;
  multFiltLin(pir,ker,0,ker.size()-1,corr,ind);
  imp(corr.v[0],"oldcorrela.img");

  printf("Procurando maximos locais em cada andar da piramide...\n");
  vector<MAXLOCAL> v;
  findMAXLOCAL(corr,ind,ker,v,minCorr);
  printf("Numero de maximos locais original=%u\n",unsigned(v.size()));

  printf("Ordenando os maximos locais em ordem decrescente de correlacao...\n");
  ordena(v,0,v.size()-1);

  printf("Imprimindo os maximos locais\n");
  imp(v,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

double interseccao(MAXLOCAL& m, Mat_<GRY>& g, MATRIZ< vector<int> >& inter)
// retorna numero entre 0 e 1
{ gray2(m,g);
  int contaArea=0; int contaInter=0;
  int lc=g.rows/2; int cc=g.cols/2;
  for (int l2=0; l2<g.rows; l2++)
    for (int c2=0; c2<g.cols; c2++)
      if (g(l2,c2)==0) {
        contaArea++;
        if (inter.atn(m.li+l2-lc,m.ci+c2-cc).size()>0) contaInter++;
      }
  return double(contaInter)/double(contaArea);
}

double interseccao2(MAXLOCAL& m, int ps, Mat_<GRY>& g, MATRIZ< vector<int> >& inter)
// retorna numero entre 0 e 1
{ gray2(m,g);
  int contaArea=0; int contaInter=0;
  int lc=g.rows/2; int cc=g.cols/2;
  for (int l2=0; l2<g.rows; l2++)
    for (int c2=0; c2<g.cols; c2++)
      if (g(l2,c2)==0) {
        contaArea++;
        vector<int>& t=inter.atn(m.li+l2-lc,m.ci+c2-cc);
        if (t.size()>1 || (t.size()==1 && t.at(0)!=ps)) contaInter++;
      }
  return double(contaInter)/double(contaArea);
}

void filtInter(vector<MAXLOCAL>& v, vector<MAXLOCAL>& w, Size tam,
               double minCorrCir, double maxInterCir, double minCorrRet, double maxInterRet,
               double minCorrQua, double maxInterQua, double minCorrEli, double maxInterEli)
// Na entrada, v deve estar ordenado em ordem decrescente por r
// Na saida, w deve estar ordenado decrescente por r
{ MATRIZ< vector<int> > inter(tam.height,tam.width); // tem indice de w
  for (unsigned l=0; l<inter.nl(); l++)
    for (unsigned c=0; c<inter.nc(); c++)
      inter(l,c).reserve(3);
  vector<int> t; t.push_back(-1); inter.backgv=t; // deixa backg com 1 elemento - ocupado

  typedef map<int,int> MAPA;
  typedef MAPA::iterator IMAPA;
  MAPA ma;
  for (unsigned i=0; i<v.size(); i++) {
    // verifica se correlacao e' suficientemente grande
    MAXLOCAL m=v[i];
    if ((m.forma=='c' && m.r<minCorrCir) ||
        (m.forma=='q' && m.r<minCorrQua) ||
        (m.forma=='r' && m.r<minCorrRet) ||
        (m.forma=='e' && m.r<minCorrEli)) continue;

    // calcula se mais de maxInter de m intersecta as formas com correlacoes maiores
    Mat_<GRY> g;
    double inters=interseccao(m,g,inter);
    if ((m.forma=='c' && inters>maxInterCir) ||
        (m.forma=='q' && inters>maxInterQua) ||
        (m.forma=='r' && inters>maxInterRet) ||
        (m.forma=='e' && inters>maxInterEli)) continue;

    // Insere m e atualiza a e inter
    w.push_back(m);
    int wi=w.size()-1;
    for (int l2=0; l2<g.rows; l2++)
      for (int c2=0; c2<g.cols; c2++)
        if (g(l2,c2)==0) {
          int l=m.li+l2-g.rows/2; int c=m.ci+c2-g.cols/2;
          if (0<=l && l<(int)inter.nl() && 0<=c && c<(int)inter.nc())
            inter(l,c).push_back(wi);
        }

    // Detecta todas as formas que m intersecta
    ma.clear();
    for (int l2=0; l2<g.rows; l2++)
      for (int c2=0; c2<g.cols; c2++)
        if (g(l2,c2)==0) {
          int l=m.li+l2-g.rows/2; int c=m.ci+c2-g.cols/2;
          if (0<=l && l<(int)inter.nl() && 0<=c && c<(int)inter.nc()) {
            vector<int>& t=inter(l,c);
            for (unsigned i=0; i<t.size(); i++) {
              assert(t[i]>=0);
              if (t[i]!=wi) ma[t[i]]=t[i]; // Multiplas copias de t[i] nao podem ser colocadas em ma.
            }
          }
        }
    // Abaixo, *p lista todas as formas que m intersecta, sem repeticoes
    bool esconde=false;
    for (IMAPA p=ma.begin(); p!=ma.end(); p++) {
      MAXLOCAL m2=w.at(p->second);
      Mat_<GRY> g2;
      double inters=interseccao2(m2,p->second,g2,inter);
      if ((m2.forma=='c' && inters>maxInterCir) ||
          (m2.forma=='q' && inters>maxInterQua) ||
          (m2.forma=='r' && inters>maxInterRet) ||
          (m2.forma=='e' && inters>maxInterEli)) { esconde=true; break; }
    }
    if (esconde) { // Tem que eliminar m
      //assert(w.back()==m);
      w.pop_back();
      for (int l2=0; l2<g.rows; l2++)
        for (int c2=0; c2<g.cols; c2++)
          if (g(l2,c2)==0) {
            int l=m.li+l2-g.rows/2; int c=m.ci+c2-g.cols/2;
            if (!(0<=l && l<g.rows && 0<=c && c<=g.cols)) continue;
            vector<int>& t=inter(l,c);
            if (t.size()>0) {
              assert(t.back()==int(wi));
              t.pop_back();
            }
          }
    }
  }
}

void tracaContorno2(Mat_<COR>& sai, MAXLOCAL m, bool bold, COR cor)
{
//   COR cor;
//   if (m.forma=='c') cor=COR(0,0,255);
//   else cor=COR(0,255,0);

  Mat_<GRY> bor;
  if      (m.forma=='c') bordaCirculo(m.d1,bor);
  else if (m.forma=='q') bordaQuadrado(m.d1,m.ang,bor);
  else if (m.forma=='r') bordaRetangulo(m.d1,m.d2,m.ang,bor);
  else if (m.forma=='e') bordaElipse(m.d1,m.d2,m.ang,bor);

  if (bold) {
    IMG_<GRY> borda(bor); borda.backg=255;
    for (int l2=borda.minl; l2<=borda.maxl; l2++)
      for (int c2=borda.minc; c2<borda.maxc; c2++)
        if (borda.atc(l2,c2)==0 || borda.atc(l2-1,c2)==0 || borda.atc(l2+1,c2)==0 || borda.atc(l2,c2-1)==0 || borda.atc(l2,c2+1)==0)
          dot(sai,m.li+l2,m.ci+c2,cor);
  } else {
    int lc=bor.rows/2;
    int cc=bor.cols/2;
    for (int l2=0; l2<bor.rows; l2++)
      for (int c2=0; c2<bor.cols; c2++)
        if (bor(l2,c2)==0)
          dot(sai,m.li+l2-lc,m.ci+c2-cc,cor);
  }
}

Mat_<COR> tracaContorno(Mat_<COR> ent, vector<MAXLOCAL>& v, bool bold, COR cor) {
  Mat_<COR> sai; ent.copyTo(sai);
  for (unsigned i=0; i<v.size(); i++) {
    tracaContorno2(sai,v[i],bold,cor);
  }
  return sai;
}

Mat_<COR> tracaIndice(Mat_<COR> ent, vector<MAXLOCAL>& v)
{ Mat_<COR> sai; ent.copyTo(sai);
  for (unsigned i=0; i<v.size(); i++) {
    string st=format("%d",i);
    putTxt(sai,v[i].li,v[i].ci,st,COR(0,0,255));
  }
  return sai;
}

Mat_<COR> tracaPonto(Mat_<COR> ent, vector<MAXLOCAL>& v, bool bold=false)
{ Mat_<COR> sai; ent.copyTo(sai);
  for (unsigned i=0; i<v.size(); i++) {
    if (bold) ponto(sai,v[i].li,v[i].ci,COR(0,255,0),3);
    else      ponto(sai,v[i].li,v[i].ci,COR(0,255,0),1);
  }
  return sai;
}

void traca(int argc, char** argv)
{ if (argc<4) {
    printf("Traca imagem.ppm imagem.ho1 saida.ppm [c|C] [i] [p|P]\n");
    printf("  Mostra os maximos locais (ho1 ou ho2) sem filtrar\n");
    printf("  c=contorno C=contorno bold\n");
    printf("  i=indice da forma\n");
    printf("  p=ponto central P=ponto bold\n");
    erro("Erro: Numero de argumentos invalido");
  }

  bool tContorno=true; bool tContornoBold=false;
  bool tIndice=false;
  bool tPonto=false; bool tPontoBold=false;
  for (int i=4; i<argc; i++) {
    if      (argv[i][0]=='c') { tContorno=true; tContornoBold=false; }
    else if (argv[i][0]=='C') { tContorno=true; tContornoBold=true;  }
    else if (argv[i][0]=='i') { tIndice=true; }
    else if (argv[i][0]=='p') { tPonto=true; tPontoBold=false;  }
    else if (argv[i][0]=='P') { tPonto=true; tPontoBold=true;  }
  }

  //printf("Lendo %s...\n",argv[1]);
  Mat_<COR> ent; le(ent,argv[1]);

  //printf("Lendo %s...\n",argv[2]);
  vector<MAXLOCAL> v; le(v,argv[2]);
  //printf("Numero de maximos locais originais=%u\n",unsigned(v.size()));

  //printf("Tracando contornos...\n");
  Mat_<COR> sai; ent.copyTo(sai);
  if (tContorno) sai=tracaContorno(sai,v,tContornoBold,COR(0,0,255));
  if (tIndice) sai=tracaIndice(sai,v);
  if (tPonto) sai=tracaPonto(sai,v,tPontoBold);
  imp(sai,argv[3]);
}

Mat_<COR> extraiAm(MAXLOCAL m, Mat_<COR> ent, double fatorZoom=1.0)
{ m.d1=m.d1*fatorZoom; m.d2=m.d2*fatorZoom;
  Mat_<GRY> g; gray2(m,g);
  assert(g.rows%2==1 && g.cols%2==1);
  int lc=g.rows/2; int cc=g.cols/2;
  Mat_<COR> sai(g.size());
  for (int l=0; l<sai.rows; l++)
    for (int c=0; c<sai.cols; c++) {
      int lent=l+m.li-lc;
      int cent=c+m.ci-cc;
      if ( lent<0 || ent.rows<=lent || cent<0 || ent.cols<=cent )
        continue;
      if ( g(l,c)==0 ) sai(l,c)=ent(lent,cent);
      else sai(l,c)=COR(0,0,0);
    }
  return sai;
}

void extraiam(int argc, char** argv)
{ if (argc<3) {
    printf("extraiAm: Extrai amostras dentro das formas localizadas\n");
    printf("extraiAm imagem.jpg fatorZoom [f0 f1 ...]\n");
    printf("  Le imagem.jpg + imagem.ho2 e imp imagem????.png\n");
    erro("Erro: Numero de argumentos invalido");
  }

  double fatorZoom;
  convArg(fatorZoom,argv[2]);

  string nome=string(argv[1]);
  printf("Lendo %s\n",nome.c_str());
  Mat_<COR> ent; le(ent,nome);

  string nomeSemSuf=semSufixo(nome);
  string nomeHo2=nomeSemSuf+".ho2";
  cout << "Lendo " << nomeHo2 << endl;
  vector<MAXLOCAL> v; le(v,nomeHo2);
  printf("Numero de maximos locais originais=%u\n",unsigned(v.size()));

  if (argc==3) {
    for (unsigned i=0; i<v.size(); i++) {
      Mat_<COR> am=extraiAm(v[i],ent,fatorZoom);
      string st=nomeSemSuf+format("%04d.png",i);
      imp(am,st);
    }
  } else {
    for (int i=3; i<argc; i++) {
      int j; convArg(j,argv[i]);
      Mat_<COR> am=extraiAm(v[j],ent,fatorZoom);
      string st=nomeSemSuf+format("%04d.png",j);
      imp(am,st);
    }
  }
}

void extraicla(int argc, char** argv)
{ if (argc<3) {
    printf("extraiCla: Extrai amostras classificadas dentro das formas localizadas\n");
    printf("extraiCla imagem.jpg fatorZoom [f0 f1 ...]\n");
    printf("  Le imagem.jpg+imagem.ho2+imagem.mat e imp CC-FFFF.png\n");
    printf("  CC: Classificacao\n");
    printf("  FFFF: Identificao da forma\n");
    erro("Erro: Numero de argumentos invalido");
  }

  double fatorZoom;
  convArg(fatorZoom,argv[2]);

  string nome=string(argv[1]);
  cout << "Lendo " << nome << endl;
  Mat_<COR> ent; le(ent,nome);

  string nomeSemSuf=semSufixo(nome);
  vector<MAXLOCAL> v; le(v,nomeSemSuf+".ho2");

  Mat_<GRY> cla; le(cla,nomeSemSuf+".mat");
  if (argc==3) {
    for (unsigned i=0; i<v.size(); i++) {
      Mat_<COR> am=extraiAm(v[i],ent,fatorZoom);
      string st=format("%02d-%04d.png",cla(i),i);
      imp(am,st);
    }
  } else {
    for (int i=3; i<argc; i++) {
      int j; convArg(j,argv[i]);
      Mat_<COR> am=extraiAm(v[j],ent,fatorZoom);
      string st=format("%02d-%04d.png",cla(j),j);
      imp(am,st);
    }
  }
}

void filtra(int argc, char** argv)
{ if (argc<4) {
    printf("Filtra imagem.pgm imagem.ho1 saida.ppm|saida.ho2 parametros\n");
    printf("  parametros: default   minCorrDef maxInterDef (0.1 e 0.3)\n");
    printf("              circulo   minCorrCir maxInterCir\n");
    printf("              retangulo minCorrRet maxInterRet\n");
    printf("              quadrado  minCorrQua maxInterQua\n");
    printf("              elipse    minCorrEli maxInterEli\n");
    erro("Erro: Numero de argumentos invalido");
  }

  double minCorrDef=0.1;
  double minCorrCir=-1;
  double minCorrRet=-1;
  double minCorrQua=-1;
  double minCorrEli=-1;
  double maxInterDef=0.3;
  double maxInterCir=-1;
  double maxInterRet=-1;
  double maxInterQua=-1;
  double maxInterEli=-1;
  if (argc>=5) {
    ISTR arq; arq.arg2str(argc-4,&argv[4]);
    string st=minuscula(leStr(arq));
    while (st!="eof") {
      if (st=="default") {
        leNum(arq,minCorrDef); minCorrCir=minCorrRet=minCorrQua=minCorrEli=minCorrDef;
        leNum(arq,maxInterDef); maxInterCir=maxInterRet=maxInterQua=maxInterEli=maxInterDef;
      } else if (st=="circulo") {
        leNum(arq,minCorrCir);
        leNum(arq,maxInterCir);
      } else if (st=="retangulo") {
        leNum(arq,minCorrRet);
        leNum(arq,maxInterRet);
      } else if (st=="quadrado") {
        leNum(arq,minCorrQua);
        leNum(arq,maxInterQua);
      } else if (st=="elipse") {
        leNum(arq,minCorrEli);
        leNum(arq,maxInterEli);
      } else {
        erro("Erro: Palavra desconhecida ",st);
      }
      st=minuscula(leStr(arq));
    }
  }
  if (minCorrCir<0) minCorrCir=minCorrDef;
  if (minCorrRet<0) minCorrRet=minCorrDef;
  if (minCorrQua<0) minCorrQua=minCorrDef;
  if (minCorrEli<0) minCorrEli=minCorrDef;
  if (maxInterCir<0) maxInterCir=maxInterDef;
  if (maxInterRet<0) maxInterRet=maxInterDef;
  if (maxInterQua<0) maxInterQua=maxInterDef;
  if (maxInterEli<0) maxInterEli=maxInterDef;

  printf("Lendo %s...\n",argv[1]);
  Mat_<GRY> ent; le(ent,argv[1]);

  printf("Lendo %s...\n",argv[2]);
  vector<MAXLOCAL> v;
  le(v,argv[2]);
  printf("Numero de maximos locais originais=%u\n",unsigned(v.size()));

  printf("Filtrando os maximos locais com interseccao>maxInter...\n");
  vector<MAXLOCAL> w;
  filtInter(v,w,ent.size(),minCorrCir,maxInterCir,minCorrRet,maxInterRet,minCorrQua,maxInterQua,minCorrEli,maxInterEli);
  printf("Numero de maximos locais apos filtro interseccao=%u\n",unsigned(w.size()));

  if (sufixo(argv[3])=="ho2") {
    imp(w,argv[3]);
  } else {
    printf("Tracando contornos...\n");
    Mat_<COR> ent2; converte(ent,ent2);
    Mat_<COR> sai=tracaContorno(ent2,w,false,COR(0,0,255));
    imp(sai,argv[3]);
  }
}

void relat(int argc, char** argv)
{ if (argc!=3) {
    printf("Relat ent.ho2 sai.rel\n");
    erro("Erro: Numero de argumentos invalido");
  }

  printf("Lendo %s...\n",argv[1]);
  vector<MAXLOCAL> v;
  le(v,argv[1]);
  printf("Numero de maximos locais originais=%u\n",unsigned(v.size()));

  map<double,int> circulos;
  map<double,int> retangulos; // retangulos ou quadrados
  for (int i=0; i<int(v.size()); i++) {
    MAXLOCAL m=v[i];
    if (m.forma=='c') {
      double area=cvRound(M_PI*elev2(m.d1/2));
      circulos[area]++; // se nao existir, assume zero?
    } else if (m.forma=='q' || (m.forma=='r')) {
      double area=cvRound(m.d1*m.d2);
      retangulos[area]++; // se nao existir, assume zero?
    }
  }

  FILE* arq=fopen(argv[2],"wt");
  if (arq==0) erro("Erro abertura ",argv[2]);
  fprintf(arq,"Circulos:\n");
  fprintf(arq,"%6s %6s\n","area","quant");
  for (map<double,int>::iterator it=circulos.begin(); it!=circulos.end(); it++)
    fprintf(arq,"%6.0f %6d\n",(*it).first,(*it).second);

  fprintf(arq,"Retangulos e quadrados:\n");
  fprintf(arq,"%6s %6s\n","area","quant");
  for (map<double,int>::iterator it=retangulos.begin(); it!=retangulos.end(); it++)
    fprintf(arq,"%6.0f %6d\n",(*it).first,(*it).second);
  fclose(arq);
}

void contaform(int argc, char** argv)
{ vector<string> nomearq; vsWildCard("*.ho2",nomearq);
  for (unsigned i=0; i<nomearq.size(); i++) {
    string nome=nomearq[i];
    FILE* arq=fopen(nome.c_str(),"rt");
    char linha[256];
    fscanf(arq,"%s",linha);
    cout << nome << " " << linha << endl;
    fclose(arq);
  }
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< MGranul.exe: Programas para granulometria multi-formas v1.23>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"  Kernel    - Gera imagem das mascaras a partir de kernel.cfg\n"
"  Correla   - Maximos locais (.ho1) das correlacoes com mascaras multiformas\n"
"  Traca     - Traca formas de maximos locais (.ho1 ou .ho2) sem filtrar\n"
"  ExtraiAm  - Extrai amostras dentro das formas localizadas\n"
"  ExtraiCla - Extrai amostras classificadas dentro das formas localizadas\n"
"  Filtra    - Filtra maximos locais (.ho1)\n"
"  OldCorrela- Maximos locais (.ho1) das correlacoes - versao antiga\n"
"  Relat     - Le .ho2 e gera .rel\n"
"  ContaForm - Le *.ho2 e imprime numero total de formas\n"
"...............................................................................\n";

int main(int argc, char** argv)
{
  printf(about);
  if (argc<2) {
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="kernel")     kernel(argc-1,&argv[1]);
    else if (comando=="correla")    correla(argc-1,&argv[1]);
    else if (comando=="traca")      traca(argc-1,&argv[1]);
    else if (comando=="extraiam")   extraiam(argc-1,&argv[1]);
    else if (comando=="extraicla")  extraicla(argc-1,&argv[1]);
    else if (comando=="filtra")     filtra(argc-1,&argv[1]);
    else if (comando=="oldcorrela") oldcorrela(argc-1,&argv[1]);
    else if (comando=="relat")      relat(argc-1,&argv[1]);
    else if (comando=="contaform")  contaform(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto: %d.%02d segundos.\n",t1/100,t1%100);
  }
}


