#include <cekeikon.h>
#include <stack>

int media(vector<GRY> r)
{ int soma=0;
  for (unsigned i=0; i<r.size(); i++)
    soma += r[i];
  return arredonda(soma,r.size());
}

void segmenc(int argc, char** argv)
{ if (!(5<=argc && argc<=9)) {
    printf("SegmenC: Segmenta Mat_<COR> pela cor, usando flann\n");
    printf("  Pixels pretos no ax nao sao levados em conta\n");
    printf("SegmenC ax.ppm ay.pgm qx.ppm qp.pgm [ntrees] [checks] [knn] [m|a]\n");
    printf("  ntrees=numero de arvores. Def=4\n");
    printf("  checks=numero de buscas recursivas. Def=4\n");
    printf("  knn=numero de vizinhos mais proximos. Def=3\n");
    printf("  decisao=criterio de desempate entre vizinhos. Median|Average. Def=m.\n");
    erro("Erro: Numero de argumentos invalido");
  }

  printf("Lendo arquivos...\n");
  Mat_<COR> ax; le(ax,argv[1]);
  Mat_<GRY> ay; le(ay,argv[2]);
  if (ax.size!=ay.size) erro("Erro: Tamanhos diferentes");

  Mat_<COR> qx; le(qx,argv[3]);
  Mat_<GRY> qp(qx.rows,qx.cols); 

  int ntrees=4;
  if (6<=argc) convArg(ntrees,argv[5]);
 
  int checks=16;
  if (7<=argc) convArg(checks,argv[6]);
  
  int knn=3;
  if (8<=argc) convArg(knn,argv[7]);
      
  char decisao='m';
  if (9<=argc) decisao=tolower(argv[8][0]);
  if (!(decisao=='m' || decisao=='a')) 
    erro("Erro: Decisao invalido"); 

  printf("Criando arvores...\n");
  Mat_<FLT> features(ax.rows*ax.cols,3);
  Mat_<FLT> saidas(ax.rows*ax.cols,1);
  int i=0;
  for (int l=0; l<ax.rows; l++)
    for (int c=0; c<ax.cols; c++) {
      COR cor=ax(l,c);
      if (cor!=COR(0,0,0) && cor!=COR(0,0,255)) {
        features(i,0)=cor[0]/255.0;
        features(i,1)=cor[1]/255.0;
        features(i,2)=cor[2]/255.0;
        saidas(i)=ay(l,c)/255.0;
        i=i+1;
      }
    }
  flann::Index ind(features,flann::KDTreeIndexParams(ntrees)); // usa 4 arvores

  printf("Fazendo busca...\n");
  vector<FLT> query(3); 
  vector<int> indices(knn);
  vector<FLT> dists(knn);
  vector<GRY> r(knn);
  for (int l=0; l<qp.rows; l++) {
    int p=arredonda(1000*l,qp.rows);
    printf("Completou %3d.%d%%\r",p/10,p%10);
    for (int c=0; c<qp.cols; c++) {
      COR cor=qx(l,c);
      query[0]=cor[0]/255.0;
      query[1]=cor[1]/255.0;
      query[2]=cor[2]/255.0;
      ind.knnSearch(query,indices,dists,knn,flann::SearchParams(checks));
      for (int i=0; i<knn; i++)
        r[i]=GRY(255*saidas(indices[i]));
      if (decisao=='m') {
        nth_element(r.begin(), r.begin()+r.size()/2, r.end());
        qp(l,c) = r[r.size()/2];
      } else {
        qp(l,c) = media(r);
      }
    }
  }
  printf("\n");
  imp(qp,argv[4]);
}

void segmencm(int argc, char** argv)
{ if (argc<4) {
    printf("SegmenCM: Segmenta Mat_<COR> pela cor, com flann e multi-exemplos\n");
    printf("  Pixels pretos no ax nao sao levados em conta\n");
    printf("SegmenCM qx.ppm qp.pgm [ntrees=4 checks=4 knn=3]\n"); 
    printf("  samples ax1.ppm ay1.pgm [ax2.ppm ay2.ppm...]\n");
    printf("ntrees=numero de arvores. Def=4\n");
    printf("checks=numero de buscas recursivas. Def=4\n");
    printf("knn=numero de vizinhos mais proximos. Def=3\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<COR> qx; le(qx,argv[1]);
  Mat_<GRY> qp(qx.rows,qx.cols); 

  int ntrees=4; 
  int checks=4;
  int knn=3;

  ISTR arq(argc-3,&argv[3]);
  string st="";
  while (true) {
    st=leStr(arq);
    if (st=="ntrees") leNum(arq,ntrees);
    else if (st=="checks") leNum(arq,checks);
    else if (st=="knn") leNum(arq,knn);
    else if (st=="samples") break;
    else erro("Erro: palavra desconhecida ",st);
  }

  printf("Lendo arquivos...\n");
  vector< Mat_<COR> > ax;
  vector< Mat_<GRY> > ay;
  Mat_<COR> axt;
  Mat_<GRY> ayt;
  while (st!="eof") {
    st=leFileName(arq); if (st=="eof") break;
    //cout << st << endl;
    le(axt,st);
    st=leFileName(arq); if (st=="eof") erro("Erro eof inesperado");
    le(ayt,st);
    //cout << st << endl;
    if (axt.size!=ayt.size) erro("Erro: Tamanhos diferentes");
    ax.push_back(axt);
    ay.push_back(ayt);
  }

  if (ax.size()==0) erro("Erro: Sem samples");
  int tam=0;
  for (int j=0; j<int(ax.size()); j++)
    tam=tam+ax[j].total();  

  printf("Criando arvores...\n");
  Mat_<FLT> features(tam,3);
  Mat_<FLT> saidas(tam,1);
  int i=0;
  for (int j=0; j<int(ax.size()); j++)
    for (int l=0; l<ax[j].rows; l++)
      for (int c=0; c<ax[j].cols; c++) {
        COR cor=ax[j](l,c);
        if (cor!=COR(0,0,0) && cor!=COR(0,0,255)) {
          features(i,0)=cor[0]/255.0;
          features(i,1)=cor[1]/255.0;
          features(i,2)=cor[2]/255.0;
          saidas(i)=ay[j](l,c)/255.0;
          i=i+1;
        }
      }
  flann::Index ind(features,flann::KDTreeIndexParams(ntrees)); // usa 4 arvores

  printf("Fazendo busca...\n");
  vector<FLT> query(3); 
  vector<int> indices(knn);
  vector<FLT> dists(knn);
  vector<GRY> r(knn);
  for (int l=0; l<qp.rows; l++) {
    int p=arredonda(1000*l,qp.rows);
    printf("Completou %3d.%d%%\r",p/10,p%10);
    for (int c=0; c<qp.cols; c++) {
      COR cor=qx(l,c);
      query[0]=cor[0]/255.0;
      query[1]=cor[1]/255.0;
      query[2]=cor[2]/255.0;
      ind.knnSearch(query,indices,dists,knn,flann::SearchParams(checks));
      for (int i=0; i<knn; i++)
        r[i]=GRY(255*saidas(indices[i]));
      nth_element(r.begin(), r.begin()+r.size()/2, r.end());
      qp(l,c) = r[r.size()/2];
    }
  }
  printf("\n");
  imp(qp,argv[2]);
}

void multedit(int argc, char** argv)
{ if (!(4<=argc && argc<=6)) {
    printf("MultEdit: Limpa amostras ruidosas\n");
    printf("  Pinta de preto em ax pixels duvidosos\n");
    printf("MultEdit ax.ppm ay.pgm cx.ppm [knn] [disc]\n");
    printf("  knn=numero de vizinhos. Def=5\n");
    printf("  disc=nr min de discordantes p/ elim amostra. Def=chao(knn/2)\n");
    printf("  ay.pgm deve ser binaria (somente 0 e 1)\n");
    erro("Erro: Numero de argumentos invalido");
  }

  printf("Lendo arquivos...\n");
  Mat_<COR> ax; le(ax,argv[1]);
  Mat_<GRY> ay; le(ay,argv[2]);
  if (ax.size!=ay.size) erro("Erro: Tamanhos diferentes");

  int knn=5;
  if (5<=argc) convArg(knn,argv[4]);

  int disc=knn/2;
  if (6<=argc) convArg(disc,argv[5]);

  int ntrees=4;
  int checks=16;

  printf("Criando arvores...\n");
  Mat_<FLT> features(ax.rows*ax.cols,3);
  Mat_<FLT> saidas(ax.rows*ax.cols,1);
  int i=0;
  for (int l=0; l<ax.rows; l++)
    for (int c=0; c<ax.cols; c++) {
      COR cor=ax(l,c);
      if (cor!=COR(0,0,0) && cor!=COR(0,0,255)) {
        features(i,0)=cor[0]/255.0;
        features(i,1)=cor[1]/255.0;
        features(i,2)=cor[2]/255.0;
        saidas(i)=ay(l,c)/255.0;
        i=i+1;
      }
    }
  flann::Index ind(features,flann::KDTreeIndexParams(ntrees)); // usa 4 arvores

  printf("Fazendo busca...\n");
  vector<FLT> query(3); 
  vector<int> indices(knn);
  vector<FLT> dists(knn);
  i=0;
  for (int l=0; l<ax.rows; l++) {
    int p=arredonda(1000*l,ax.rows);
    printf("Completou %3d.%d%%\r",p/10,p%10);
    for (int c=0; c<ax.cols; c++) {
      COR cor=ax(l,c);
      if (cor!=COR(0,0,0) && cor!=COR(0,0,255)) {
        query[0]=features(i,0);
        query[1]=features(i,1);
        query[2]=features(i,2);
        ind.knnSearch(query,indices,dists,knn,flann::SearchParams(checks));
        int zeros=0; int uns=0;
        for (int j=0; j<knn; j++) {
          FLT f=saidas(indices[j]);
          if (f==0.0) zeros++;
          else if (f==1.0) uns++;
          else erro("Erro: ay.pgm deve ser binaria");
        }
        if ( (saidas(i)==0 && uns>=disc) || (saidas(i,0)==1 && zeros>=disc) )
          ax(l,c)=COR(0,0,0);
        i=i+1;
      }
    }
  }
  printf("\n");
  imp(ax,argv[3]);
}

struct NO {
  int l,c;
  FLT r,g,b,s;
  bool condensado;
};

void condense(int argc, char** argv)
{ if (argc!=4) {
    printf("Condense: Joga fora amostras desnecessarias\n");
    printf("  Pinta de vermelho em ax os pixels desnecessarios\n");
    printf("Condense cx.ppm ay.pgm dx.ppm\n");
    printf("  cx.ppm e' imagem que passou pelo multedit\n");
    printf("  ay.pgm deve ser binaria (somente 0 e 1) onde ax nao for preto\n");
    printf("  dx.ppm e' cx.ppm com pixels desnecessarios pintados em vermelho\n");
    erro("Erro: Numero de argumentos invalido");
  }

  printf("Lendo arquivos...\n");
  Mat_<COR> cx; le(cx,argv[1]);
  Mat_<GRY> ay; le(ay,argv[2]);
  if (cx.size!=ay.size) erro("Erro: Tamanhos diferentes");
  Mat_<COR> dx=cx.clone();

  vector<NO> v(cx.total());

  int i=0;
  for (int l=0; l<cx.rows; l++)
    for (int c=0; c<cx.cols; c++) {
      COR cor=cx(l,c);
      if (cor!=COR(0,0,0) && cor!=COR(0,0,255)) {
        NO& no=v[i];
        no.l=l;
        no.c=c;
        no.r=cor[0]/255.0;
        no.g=cor[1]/255.0;
        no.b=cor[2]/255.0;
        no.s=ay(l,c)/255.0;
        no.condensado=true;
        i=i+1;
      }
    }
  random_shuffle(v.begin(),v.end());

  Mat_<FLT> features(2,3);
  Mat_<FLT> saidas(2,1);

  features(0,0)=v[0].r; features(0,1)=v[0].g; features(0,1)=v[0].b; saidas(0,0)=v[0].s; 
  features(1,0)=v[1].r; features(1,1)=v[1].g; features(1,1)=v[1].b; saidas(1,0)=v[1].s;

  vector<FLT> query(3); 
  vector<int> indices(1);
  vector<FLT> dists(1);
  Mat_<FLT> linhaf(1,3);
  Mat_<FLT> linhas(1,1);
  for (int j=2; j<int(v.size()); j++) {
    NO& no=v[j];
    flann::Index ind(features,flann::LinearIndexParams()); // forca bruta
    query[0]=no.r; query[1]=no.g; query[2]=no.b;
    ind.knnSearch(query,indices,dists,1);
    if (saidas(indices[0])!=no.s) { // busca deu erro. Inserir no em features
      linhaf(0,0)=no.r; linhaf(0,1)=no.g; linhaf(0,1)=no.b; 
      features.push_back(linhaf);
      linhas(0,0)=no.s; 
      saidas.push_back(linhas);
    } else no.condensado=false;
  }

  for (int j=0; j<int(v.size()); j++) {
    NO& no=v[j];
    if (no.condensado==false) 
      dx(no.l,no.c)=COR(0,0,255);
  }
  imp(dx,argv[3]);
}

namespace ContaCCB {

inline void expande(stack<int>& pilha, IMG_<GRY>& entp, int l, int c)
{ if (entp(l,c)==0) {
    pilha.push(l); pilha.push(c);
  }
}

void limpa(Mat_<GRY>& ent, int l, int c, Mat_<COR>& sai, Point& centro, int& area)
{ int somal=0;
  int somac=0;
  int conta=0;
  IMG_<GRY> entp(ent,0,0); entp.backg=255;
  stack<int> pilha;
  pilha.push(l); pilha.push(c);
  while (!pilha.empty()) {
    int c=pilha.top(); pilha.pop();
    int l=pilha.top(); pilha.pop();
    entp(l,c)=1;
    sai(l,c)=COR(0,0,0);
    somal+=l; somac+=c; conta++;

    expande(pilha,entp,l-1,c);
    expande(pilha,entp,l+1,c);
    expande(pilha,entp,l,c-1);
    expande(pilha,entp,l,c+1);
  }
  centro=Point(arredonda(somac,conta),arredonda(somal,conta));
  area=conta;
}

void conta(Mat_<GRY>& ent, Mat_<COR>& sai, int& cc)
{ cc=0;
  sai.create(ent.size()); 
  sai.setTo(COR(255,255,255));
  for (int l=0; l<ent.rows; l++)
    for (int c=0; c<ent.cols; c++)
      if (ent(l,c)==0) {
        Point centro; int area;
        limpa(ent,l,c,sai,centro,area);
        string st=format("%d",cc);
        putTxt(sai,centro.y-6,centro.x-4,st,COR(0,0,255));
        st=format("%d",area);
        putTxt(sai,centro.y+6,centro.x-4,st,COR(0,255,0));
        cc++;
      }
}

}

void contaccb(int argc, char** argv)
{ using namespace ContaCCB;
  if (argc!=3) {
    printf("ContaCCB ent.pbm sai.png\n");
    printf("Conta componentes 4-conexos, calculando areas.\n");
    printf("  Rotulo=vermelho, Area=verde\n");
    printf("  ent.pbm e' binarizada (se nao for).\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<GRY> ent; le(ent,argv[1]);
  for (int l=0; l<ent.rows; l++)
    for (int c=0; c<ent.cols; c++)
      if (ent(l,c)<128) ent(l,c)=0;
      else ent(l,c)=255;

  Mat_<COR> sai;
  int n; conta(ent,sai,n);
  imp(sai,argv[2]);
  printf("Numero de componentes conexos: %d\n",n);
}

inline void expande(stack<int>& pilha, IMG_<GRY>& entp, int l, int c)
{ if (entp(l,c)==0) {
    pilha.push(l); pilha.push(c);
  }
}

void limpa(Mat_<GRY>& ent, int l, int c, Point& centro, int& area)
{ int somal=0;
  int somac=0;
  int conta=0;
  IMG_<GRY> entp(ent,0,0);
  stack<int> pilha;
  pilha.push(l); pilha.push(c);
  while (!pilha.empty()) {
    int c=pilha.top(); pilha.pop();
    int l=pilha.top(); pilha.pop();
    entp(l,c)=1;
    somal+=l; somac+=c; conta++;

    expande(pilha,entp,l-1,c);
    expande(pilha,entp,l+1,c);
    expande(pilha,entp,l,c-1);
    expande(pilha,entp,l,c+1);
  }
  centro=Point(arredonda(somac,conta),arredonda(somal,conta));
  area=conta;
}

void conta(Mat_<COR>& ori, Mat_<GRY>& ent, Mat_<COR>& sai, int& cc)
{ using namespace Morphology;
  Mat_<GRY> borda=ent ^ (ent+Mat_<GRY>(3,3,255));
  cc=0;
  sai=ori.clone();
  for (int l=0; l<ent.rows; l++)
    for (int c=0; c<ent.cols; c++)
      if (borda(l,c)) sai(l,c)=COR(0,0,255);

  for (int l=0; l<ent.rows; l++)
    for (int c=0; c<ent.cols; c++)
      if (ent(l,c)==0) {
        Point centro; int area;
        limpa(ent,l,c,centro,area);
        string st=format("%d",cc);
        putTxt(sai,centro.y-6,centro.x-4,st,COR(0,0,255));
        st=format("%d",area);
        putTxt(sai,centro.y+6,centro.x-4,st,COR(0,128,0));
        cc++;
      }
}

void showarea(int argc, char** argv)
{ if (argc!=4) {
    printf("ShowArea ori.jpg ent.pbm sai.png\n");
    printf("Conta componentes 4-conexos, calculando areas.\n");
    printf("  Rotulo=vermelho, Area=verde\n");
    printf("  ent.pbm e' binarizada (se nao for).\n");
    erro("Erro: Numero de argumentos invalido");
  }

  Mat_<COR> ori; le(ori,argv[1]);
  Mat_<GRY> ent; le(ent,argv[2]);
  if (ori.size()!=ent.size()) erro("Erro: tamanhos diferentes ori.jpg e ent.pbm");
  for (int l=0; l<ent.rows; l++)
    for (int c=0; c<ent.cols; c++)
      if (ent(l,c)<128) ent(l,c)=0;
      else ent(l,c)=255;

  Mat_<COR> sai;
  int n; conta(ori,ent,sai,n);
  imp(sai,argv[3]);
  printf("Numero de componentes conexos: %d\n",n);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< Paviment.exe: Programas para analisar pavimento v1.00>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"  SegmenC  - Segmenta imagem usando k-NN\n"
"  SegmenCM - Segmenta imagem usando k-NN e multi-samples\n"
"  MultEdit - Limpa amostras ruidosas\n"
"  Condense - Elimina amostras desnecessarias\n"
"  ContaCCB - Conta e numera componentes conexos\n"
"  ShowArea - Conta e numera componentes conexos sobreposto a imagem original\n"
"...............................................................................\n";

int main(int argc, char** argv)
{ 
  printf(about);
  if (argc<2) {
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="segmenc")  segmenc(argc-1,&argv[1]);
    else if (comando=="segmencm") segmencm(argc-1,&argv[1]);
    else if (comando=="multedit") multedit(argc-1,&argv[1]);
    else if (comando=="condense") condense(argc-1,&argv[1]);
    else if (comando=="contaccb") contaccb(argc-1,&argv[1]);
    else if (comando=="showarea") showarea(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto: %d.%02d segundos.\n",t1/100,t1%100);
  }
}
  
  
