#include <cekeikon.h>
#include <queue>

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void ecg2img(vector< vector<int> >& ecg, vector<COR> cor, Mat_<COR>& sai)
{ if (ecg.size()>cor.size()) erro("Erro ecg2img: ecg e cor devem ser do mesmo tamanho");

  unsigned maxx=0;
  for (unsigned l=0; l<ecg.size(); l++) 
    if (ecg[l].size()>maxx) maxx=ecg[l].size();

  int miny=maxint; int maxy=minint;
  for (unsigned l=0; l<ecg.size(); l++) 
    for (unsigned c=0; c<ecg[l].size(); c++) {
      int valor=ecg[l][c];
      if (valor>maxy) maxy=valor;
      if (valor<miny) miny=valor;
    }

  int nl=-miny+maxy+1; int nc=maxx;
  sai.create(nl,nc); sai.setTo(Scalar(255,255,255));

  int passo=10; 
  for (int i=passo; i<maxy; i+=passo) 
    reta(sai,sai.rows-1+miny-i,0,sai.rows-1+miny-i,sai.cols-1,COR(182,182,182));
  for (int i=-passo; i>miny; i-=passo) 
    reta(sai,sai.rows-1+miny-i,0,sai.rows-1+miny-i,sai.cols-1,COR(182,182,182));
  passo=100; 
  for (int i=passo; i<maxy; i+=passo) 
    reta(sai,sai.rows-1+miny-i,0,sai.rows-1+miny-i,sai.cols-1,COR(96,96,96));
  for (int i=-passo; i>miny; i-=passo) 
    reta(sai,sai.rows-1+miny-i,0,sai.rows-1+miny-i,sai.cols-1,COR(96,96,96));
  reta(sai,sai.rows-1+miny,0,sai.rows-1+miny,sai.cols-1,COR(0,0,0));

  passo=10; 
  for (unsigned i=passo; i<maxx; i+=passo) 
    reta(sai,0,i,sai.rows-1,i,COR(182,182,182));
  passo=100; 
  for (unsigned i=passo; i<maxx; i+=passo) 
    reta(sai,0,i,sai.rows-1,i,COR(96,96,96));
  reta(sai,0,0,sai.rows-1,0,COR(0,0,0));

  for (unsigned l=0; l<ecg.size(); l++) {
    moveTo(sai.rows-1+miny-ecg[l][0],0);
    for (unsigned c=1; c<ecg[l].size(); c++) 
      lineTo(sai,sai.rows-1+miny-ecg[l][c],c,cor[l]);
  }
}

void ecg2img(vector<int>& ecg, COR cor, Mat_<COR>& sai)
{ vector< vector<int> > ecg2(1); ecg2[0]=ecg;
  vector<COR> cor2(1); cor2[0]=cor;
  ecg2img(ecg2,cor2,sai);
}

void ecg2img(int argc, char** argv)
{ if (argc<3) {
    printf("ECG2IMG: Converte ECG em imagem\n");
    printf("ECG2IMG imagem.png ECG1.ecg ECG2.ecg ...\n");
    printf("  Os ECGs serao de cores R G B C M Y aleatorio\n");
    erro("Erro: Numero invalido de argumentos");
  }
  int inicio=2;
  vector< vector<int> > ecg(argc-inicio);
  vector<COR> cor(argc-inicio);
  srand(7);

  for (int l=0; l<argc-inicio; l++) {
    le(ecg[l],argv[l+inicio]); 
    switch (l) {
      case 0: cor[l]=COR(0,0,255); break;   // red
      case 1: cor[l]=COR(0,255,0); break;   // green
      case 2: cor[l]=COR(255,0,0); break;   // blue
      case 3: cor[l]=COR(128,128,0); break; // cyan
      case 4: cor[l]=COR(128,0,128); break; // magenta
      case 5: cor[l]=COR(0,128,128); break; // yellow
      default: cor[l]=COR(rand()%150,rand()%150,rand()%150);
    }
  }

  Mat_<COR> a;
  ecg2img(ecg,cor,a);
  imp(a,argv[1]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

int dcReject(vector<int>& v) // Elimina nivel DC e retorna media
{ double soma=0.0;
  for (unsigned i=0; i<v.size(); i++) soma += v[i];
  int media = arredonda(soma / v.size());
  for (unsigned i=0; i<v.size(); i++) v[i] -= media;
  return media;
}

void dcreject(int argc, char** argv)
{ if (argc!=3) {
    printf("DCReject: Elimina nivel DC de ECG\n");
    printf("DCReject ent.ecg sai.ecg\n");
    erro("Erro: Numero invalido de argumentos");
  }

  vector<int> ecg; le(ecg,argv[1]);
  dcReject(ecg);
  imp(ecg,argv[2]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
struct FI1 { int v; unsigned c; };
struct FI1menor {
  bool operator() (const FI1& x, const FI1& y) const
    { return (x.v<y.v); }
};
typedef priority_queue<FI1,vector<FI1>,FI1menor> FI1DEC; // Ordena em ordem decrescente
struct FI1maior {
  bool operator() (const FI1& x, const FI1& y) const
    { return (x.v>y.v); }
};
typedef priority_queue<FI1,vector<FI1>,FI1maior> FI1INC; // Ordena em ordem decrescente

vector<unsigned> kmax(vector<int>& a, int k, int d)
// Acha k maiores elementos separados por distancia>=d
// Se d==0, acha k maiores elementos.
// Nota: pixels==minint nao sao levados em conta
// Nota: Na saida, pode ter menos de k elementos
{ FI1DEC pq;
  FI1 f;
  for (unsigned c=0; c<a.size(); c++) {
    if (a[c]>minint) {
      f.v=a[c]; f.c=c;
      pq.push(f);
    }
  }

  vector<unsigned> psai;
  if (d>0) {
    for (int i=0; i<k; i++) {
      bool coloca;
      do {
        if (pq.empty()) goto saida;
        FI1 f=pq.top();
        pq.pop();
        coloca=true;
        for (unsigned j=0; j<psai.size(); j++) {
          if (abs(int(psai[j])-int(f.c))<d) {
            coloca=false; break;
          }
        }
        if (coloca) { // distancia(f.c,psai[j])>=d, qualquer j
          psai.push_back(f.c);
        }
      } while (!coloca);
    }
  } else {
    for (int i=0; i<k; i++) {
      if (pq.empty()) goto saida;
      FI1 f=pq.top();
      pq.pop();
      psai.push_back(f.c);
    }
  }

  saida:
  return psai;
}

vector<unsigned> lmin(vector<int>& a, int l, int d)
// Acha l menores elementos separados por distancia>=d
// Se d==0, acha l menores elementos.
// Nota: pixels==minint nao sao levados em conta
// Nota: Na saida, pode ter menos de l elementos
{ FI1INC pq;
  FI1 f;
  for (unsigned c=0; c<a.size(); c++) {
    if (a[c]>minint) {
      f.v=a[c]; f.c=c;
      pq.push(f);
    }
  }

  vector<unsigned> psai;
  if (d>0) {
    for (int i=0; i<l; i++) {
      bool coloca;
      do {
        if (pq.empty()) goto saida;
        FI1 f=pq.top();
        pq.pop();
        coloca=true;
        for (unsigned j=0; j<psai.size(); j++) {
          if (abs(int(psai[j])-int(f.c))<d) {
            coloca=false; break;
          }
        }
        if (coloca) { // distancia(f.c,psai[j])>=d, qualquer j
          psai.push_back(f.c);
        }
      } while (!coloca);
    }
  } else {
    for (int i=0; i<l; i++) {
      if (pq.empty()) goto saida;
      FI1 f=pq.top();
      pq.pop();
      psai.push_back(f.c);
    }
  }

  saida:
  return psai;
}


//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

struct PICO {
  int c, dx, dy; double zx,zy;
};

void impPicos(vector<PICO>& picos, string st)
{ FILE* arq=fopen(st.c_str(),"wt");
  if (arq==0) erro("Erro abertura arquivo ",st);
  fprintf(arq,"%u\n",unsigned(picos.size()));
  fprintf(arq,"%8s %5s %5s %5s %5s\n","//local","deslX","zoomX","deslY","zoomY");
  for (unsigned i=0; i<picos.size(); i++) {
    fprintf(arq,"%8u %5d %5.2f %5d %5.2f\n",picos[i].c,picos[i].dx,picos[i].zx,picos[i].dy,picos[i].zy);
  }
  fclose(arq);
}

void lePicos(vector<PICO>& picos, string st)
{ FILE* arq=fopen(st.c_str(),"rt");
  if (arq==0) erro("Erro arquivo inexistente ",st);
  int n; leNum(arq,n);
  picos.resize(n);
  for (int i=0; i<n; i++) {
    PICO& p=picos[i];
    leNum(arq,p.c);
    leNum(arq,p.dx);
    leNum(arq,p.zx);
    leNum(arq,p.dy);
    leNum(arq,p.zy);
  }
  fclose(arq);
}

void detpico(int argc, char** argv)
{ if (argc!=5 && argc!=6) {
    printf("detpico: Detecta pico R\n");
    printf("detpico ECG.ecg sai.ecg|sai.pic nPicos sepMin [+|-]\n");
    printf("  default: Detecta pico positivo +\n");
    erro("Erro: Numero invalido de argumentos");
  }

  vector<int> ecg; le(ecg,argv[1]); 
  vector<int> sai(ecg.size(),0); 
  int nPicos; convArg(nPicos,argv[3]);
  int sepMin; convArg(sepMin,argv[4]);

  char modo='+'; if (argc>5) modo=argv[5][0];
  int sinal;
  if (modo=='+') sinal=1;
  else if (modo=='-') sinal=-1;
  else erro("Erro: Modo invalido");

  vector<unsigned> picos;
  if (modo=='+') picos=kmax(ecg, nPicos, sepMin);
  else picos=lmin(ecg, nPicos, sepMin);
  sort(picos.begin(), picos.end());

  string suf=sufixo(argv[2]);
  if (suf=="ecg") {
    for (unsigned i=0; i<picos.size(); i++) {
      sai.at(picos.at(i))=sinal*100;
    }
    imp(sai,argv[2]);
  } else if (suf=="pic") {
    vector<PICO> picos2(picos.size());
    for (unsigned i=0; i<picos.size(); i++) {
      picos2[i].c=picos[i];
      picos2[i].dx=0;
      picos2[i].zx=1.0;
      picos2[i].dy=0;
      picos2[i].zy=1.0;
    }

    impPicos(picos2,argv[2]);
  } else erro("Erro: Sufixo desconhecido ",argv[2]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void minimosQuadrados(vector<int>& y, vector<int>& x, double& a, double& b)
// Calcular Y=X*beta+epsilon, beta=[a,b]
{ if (y.size()!=x.size()) erro("Erro minimosQuadrados: Dimensoes diferentes");
  Mat_<double> Y(y.size(),1);
  for (int i=0; i<Y.rows; i++) Y(i)=y[i];
  //cout << "Y=" << Y << endl;

  Mat_<double> X(x.size(),2, 1.0);
  for (int i=0; i<X.rows; i++) X(i,0)=x[i];
  //cout << "X=" << X << endl;

  Mat_<double> beta = (X.t() * X).inv() * X.t() * Y;
  if (beta.size()!=Size(1,2)) erro("Erro inesperado: dimensao beta");
  a=beta(0);
  b=beta(1);
  //cout << "beta=" << beta << endl;
}

void ajustemp(int argc, char** argv)
{ if (argc!=4 && argc!=6) {
    printf("AjusTemp: Ajusta template nos picos pelo minimos quadraticos\n");
    printf("AjusTemp ent.ecg ent.pic temp.ecg temp.pic ajust.pic\n");
    printf("AjusTemp ent temp ajust\n");
    erro("Erro: Numero invalido de argumentos");
  }

  string argv1,argv2,argv3,argv4,argv5;
  if (argc==6) {
    argv1=argv[1]; argv2=argv[2]; argv3=argv[3]; argv4=argv[4]; argv5=argv[5];
  } else {
    argv1=string(argv[1])+".ecg"; argv2=string(argv[1])+".pic"; 
    argv3=string(argv[2])+".ecg"; argv4=string(argv[2])+".pic"; 
    argv5=string(argv[3])+".pic";
  }

  vector<int> ka; le(ka,argv1); VECTOR<int> a(ka); a.backg=0; 
  vector<PICO> pa; lePicos(pa,argv2); //impPicos(pa,"con");

  vector<int> kt; le(kt,argv3); VECTOR<int> t(kt); t.backg=0;
  vector<PICO> kpt; lePicos(kpt,argv4); //impPicos(kpt,"con");
  if (kpt.size()>1) erro("Erro: template com mais de um pico");
  PICO pt=kpt[0];

  vector<PICO> ps=pa;
 
  // vetor temporario que guarda trecho de a em torno do pico
  vector<int> kb(kt.size()); VECTOR<int> b(kb); b.backg=0; 
  for (int i=0; i<int(pa.size()); i++) {
    for (int j=0; j<int(t.size); j++) 
      b(j)=a(j+pa[i].c-pt.c); 
    double a,b;
    minimosQuadrados(kb,kt,a,b);
    ps[i].dy=b;
    ps[i].zy=a;
  }
  impPicos(ps,argv5);
}

void picotemp(int argc, char** argv)
{ if (argc!=4 && argc!=6) {
    printf("PicoTemp: Convolui pico com template, fazendo ajuste deslY e zoomY\n");
    printf("PicoTemp ent.ecg ent.pic temp.ecg temp.pic picotemp.ecg\n");
    printf("PicoTemp ent temp picotemp\n");
    erro("Erro: Numero invalido de argumentos");
  }

  string argv1,argv2,argv3,argv4,argv5;
  if (argc==6) {
    argv1=argv[1]; argv2=argv[2]; argv3=argv[3]; argv4=argv[4]; argv5=argv[5];
  } else {
    argv1=string(argv[1])+".ecg"; argv2=string(argv[1])+".pic"; 
    argv3=string(argv[2])+".ecg"; argv4=string(argv[2])+".pic"; 
    argv5=string(argv[3])+".ecg";
  }

  vector<int> ka; le(ka,argv1); VECTOR<int> a(ka); a.backg=0; 
  vector<PICO> pa; lePicos(pa,argv2); //impPicos(pa,"con");

  vector<int> kt; le(kt,argv3); VECTOR<int> t(kt); t.backg=0;
  vector<PICO> kpt; lePicos(kpt,argv4); //impPicos(kpt,"con");
  if (kpt.size()>1) erro("Erro: template com mais de um pico");
  PICO pt=kpt[0];

  vector<int> ks(ka.size(),0); VECTOR<int> s(ks); s.backg=0;
 
  for (unsigned i=0; i<pa.size(); i++) {
    int delta=pa[i].c-pt.c;
    for (unsigned j=0; j<t.size; j++) {
      s(j+delta)=pa[i].zy*t(j)+pa[i].dy; // todos somas e subs com sinal
    }    
  }
  imp(ks,argv5);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void subecg(int argc, char** argv)
{ if (argc!=4) {
    printf("SubECG: Subtrai um ECG de outro\n");
    printf("SubECG ent1.ecg ent2.ecg sai.ecg\n");
    erro("Erro: Numero invalido de argumentos");
  }

  vector<int> ent1; le(ent1,argv[1]); 
  vector<int> ent2; le(ent2,argv[2]); 
  if (ent1.size()!=ent2.size()) erro("Erro: ECGs com tamanhos diferentes");
  vector<int> sai(ent1.size());
  for (unsigned i=0; i<sai.size(); i++) 
    sai[i]=ent1[i]-ent2[i];
  imp(sai,argv[3]);
}

void negativo(int argc, char** argv)
{ if (argc!=3) {
    printf("Negativo: Gera ECG negativo\n");
    printf("Negativo ent.ecg sai.ecg\n");
    erro("Erro: Numero invalido de argumentos");
  }

  vector<int> ent; le(ent,argv[1]); 
  for (unsigned i=0; i<ent.size(); i++) 
    ent[i]=-ent[i];
  imp(ent,argv[2]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< ECG: Programas para processar ECG v1.04>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"ECG2IMG   - Converte ECG em imagem\n"
"DCReject  - Elimina nivel DC de ECG\n"
"DetPico   - Detecta pico R\n"
"AjusTemp  - Ajusta template no pico pelo metodo minimos quadraticos\n"
"PicoTemp  - Convolui pico com template, fazendo ajuste deslY e zoomY\n"
"SubECG    - Subtrai um ECG de outro\n"
"Negativo  - Gera ECG negativo\n"
"...............................................................................\n";

int main(int argc, char** argv)
{ if (argc<2) {
    printf(about);
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="ecg2img")   ecg2img(argc-1,&argv[1]);
    else if (comando=="dcreject")  dcreject(argc-1,&argv[1]);
    else if (comando=="detpico")   detpico(argc-1,&argv[1]);
    else if (comando=="ajustemp")  ajustemp(argc-1,&argv[1]);
    else if (comando=="picotemp")  picotemp(argc-1,&argv[1]);
    else if (comando=="subecg")    subecg(argc-1,&argv[1]);
    else if (comando=="negativo")  negativo(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto=%d.%02ds\n",t1/100,t1%100);
  }
}

/*
Falta fazer:
1) Calcular automaticamente ZoomY
2) Permitir alterar manualmente ZoomX
*/
 