#include <cekeikon.h>

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< RIHOG <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
class RIHOG {
 public:
  Mat_<Vec2f> kernel; // angulo, anel (-1 indica fora de circulo ou no centro)
  vector<int> conta; // quantos pixeis tem cada anel
  int naneis; // 5
  int raio; // raio = naneis*larguraAnel = 40
  double raioInic; // raioInic = raio / sqrt(naneis)
  vector<double> raios; // Onde termina cada anel
  int diametro; // diametro = 2*raio+1 = 81
  int nbins; // 20
  int ngauss; // 37
  int descriptorSize; // descriptorSize=nbins*naneis=100
  int tnl; // tnl=diametro+2*(ngauss/2)
  double tuso; // tuso=double(diametro)/double(tnl)
  bool bcinv; // brightness-contrast invariant
  double denominador;
  double denominador2;

  void inicializa(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true);
  RIHOG(int naneisp=5, int raiop=40, int nbinsp=20, int ngaussp=37, bool bcinvp=true) { inicializa(naneisp,raiop,nbinsp,ngaussp,bcinvp); }
  int quantiza(double d);
  void rihog(Mat_<GRY> a, M3d_<FLT>& features);

  //void computeDescriptorC(Mat_<GRY> a, KeyPoint key, Mat_<FLT> dr);
  //void computeDescriptor(Mat_<GRY> a, vector<KeyPoint>& k, Mat_<FLT>& d);
};

void RIHOG::inicializa(int naneisp, int raiop, int nbinsp, int ngaussp, bool bcinvp)
{ naneis=naneisp;
  raio=raiop;
  raioInic = double(raio) / sqrt(double(naneis));
  raios.resize(naneis);
  for (int i=0; i<naneis; i++)
    raios[i] = sqrt(double(i)) * raioInic;
  diametro=2*raio+1;
  nbins=nbinsp;
  ngauss=ngaussp;
  descriptorSize=nbins*naneis;
  tnl=diametro+2*(ngauss/2);
  tuso=double(diametro)/double(tnl);
  bcinv=bcinvp;
  denominador=360.0/nbins;
  denominador2=denominador/2;

  kernel.create(tnl,tnl);
  IMG_<Vec2f> kernelp(kernel);
  for (int x=kernelp.minx; x<=kernelp.maxx; x++)
    for (int y=kernelp.miny; y<=kernelp.maxy; y++) {
      Vec2f& p=kernelp.atC(x,y);

      double r=sqrt(double(x*x+y*y));
      int q=0; 
      for (int i=0; i<naneis; i++)
        if (r<raios[i]) { q=i; break; }
      if (q>=naneis || r==0.0) { 
        p[0]=0; p[1]=-1; continue; 
      } else { 
        p[0]=fastAtan2(y,x); // entre 0 e 360
        p[1]=q; 
      }
    }
  
  conta=vector<int>(naneis,0);
  for (int l=0; l<kernel.rows; l++)
    for (int c=0; c<kernel.cols; c++) {
      if (kernel(l,c)[1]>=0) 
        conta[kernel(l,c)[1]]++;
    }
}

int RIHOG::quantiza(double d)
// 0<=d<360
{ 
  //if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-1 fora de 0 a 360");
  d+=denominador2;
  if (d>=360) d-=360; // 0<=d<360
  //if (d<0 || 360<=d) erro("Erro RIHOG:quantiza-2 fora de 0 a 360");
  int q=int(d/denominador);      
  //if (!(0<=q && q<nbins)) erro("Erro RIHOG::quantiza inesperado"); 
  return q;  
}

void RIHOG::rihog(Mat_<GRY> _a, M3d_<FLT>& features)
{ if (_a.rows<kernel.rows || _a.cols<kernel.cols) erro("Erro RIHOG::rihog tamanho imagem");

  Mat_<GRY> a;
  GaussianBlur(_a, a, Size(ngauss,ngauss), 0, 0, BORDER_REPLICATE);

  vector< Mat_<SHT> > g(2); // gradiente 2 direcoes: ->  ^
  FLT vk0[]={-1, 0, 1};
  Mat_<FLT> k0(1,3, vk0);
  filter2D(a, g[0], CV_16S, k0, Point(-1,-1), 0, BORDER_REPLICATE); 
  FLT vk1[]={1, 0, -1};
  Mat_<FLT> k1(3,1,vk1);
  filter2D(a, g[1], CV_16S, k1, Point(-1,-1), 0, BORDER_REPLICATE); 

  Mat_<FLT> mag(a.size());
  Mat_<FLT> deg(a.size());
  for (int l=0; l<a.rows; l++)
    for (int c=0; c<a.cols; c++) {
      float g0=g[0](l,c);
      float g1=g[1](l,c);
      mag(l,c)=sqrt(g1*g1+g0*g0);
      deg(l,c)=fastAtan2(g1,g0); // fastAtan2(y,x)
    }

  int vint[]={a.rows,a.cols,descriptorSize};
  features.create(3,vint);
  features.setTo(0.0);

  Mat_<FLT> f(naneis,nbins,0.0);
  for (int l=raio; l<a.rows-raio; l++) {
    printf("Linha %d\r",l);
    for (int c=raio; c<a.cols-raio; c++) {
      for (int l2=-raio; l2<=raio; l2++)
        for (int c2=-raio; c2<=raio; c2++) {
          Vec2f k=kernel(l2+raio,c2+raio); // angulo, anel
          if (k[1]==-1) continue; // centro ou fora do maior raio
          //if (k[0]<0 || k[0]>=360) erro("Erro k[0]=angulo fora da faixa"); // Pode tirar, apos debugar
          //if (k[1]<0 || k[1]>=naneis) erro("Erro k[1]=aneis fora da faixa"); // Pode tirar, apos debugar
          float gmag=mag(l+l2,c+c2);
          float gdeg=deg(l+l2,c+c2)-k[0];
          if (gdeg<0) gdeg+=360; // 0<=gdeg<360
          int q=quantiza(gdeg);
          f(arredonda(k[1]),q)+=gmag;
        }
      if (bcinv) {
        double soma=0.0;
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            soma += f(l2,c2);
        double media=soma/f.total();
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            f(l2,c2) -= media;
        double somaq=0.0;
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            somaq += elev2(f(l2,c2));
        double desvio=somaq/f.total();
        if (desvio>epsilon) {
          for (int l2=0; l2<f.rows; l2++)
            for (int c2=0; c2<f.cols; c2++) 
              f(l2,c2) /= desvio;
        } else {
          for (int l2=0; l2<f.rows; l2++)
            for (int c2=0; c2<f.cols; c2++) 
              f(l2,c2) = 0.0;
        }
      } else {
        for (int l2=0; l2<f.rows; l2++)
          for (int c2=0; c2<f.cols; c2++) 
            f(l2,c2) /= conta[l2]; // l=anel
      }
      int i=0;
      for (int l2=0; l2<f.rows; l2++)
        for (int c2=0; c2<f.cols; c2++) {
          features(l,c,i)=f(l2,c2);
          i++;
        }
    }
  }
  printf("\n");
}

void features(int argc, char** argv)
{ if (argc<3) {
    printf("Features: Gera RIHOG para todos os pixels validos\n");
    printf("Features ent.pgm sai.img \"naneis=5 raio=40 nbins=20 ngauss=37 bcinv=true\"\n");
    printf("  bcinv=brightness-contrast invariant\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Mat_<GRY> a; le(a,argv[1]);

  int naneis=5; 
  int raio=40;
  int nbins=20;
  int ngauss=37;
  bool bcinv=true;
  if (argc>3) { 
    ISTR arq(argc-3,&argv[3]);
    string st;
    seekArg(arq,"naneis",naneis);
    seekArg(arq,"raio",raio);
    seekArg(arq,"nbins",nbins);
    seekArg(arq,"ngauss",ngauss);
    seekArg(arq,"bcinv",bcinv);
    if (arq.st.size()!=0) erro("Erro: Argumento invalido ",arq.st);
  }

  Mat_<FLT> d;
  RIHOG r(naneis,raio,nbins,ngauss,bcinv);
  Mat_<FLT> features;
  r.rihog(a,features);
  imp(features,argv[2]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< RIHOG.exe: Template matching baseado em rotation-invariant HOG v1.01>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"  Features - Gera RIHOG para todos os pixels validos\n"
"...............................................................................\n";

int main(int argc, char** argv)
{ 
  printf(about);
  if (argc<2) {
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    int t1=centseg();
    string comando=minuscula(argv[1]);
    if      (comando=="features") features(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
    t1=centseg()-t1;
    printf("Tempo gasto: %d.%02d segundos.\n",t1/100,t1%100);
  }
}
  
  
