#include <cekeikon.h>
#include <opencv2/nonfree/nonfree.hpp>

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<,
void surf(int argc, char** argv)
{ if (argc!=3) {
    printf("SURF: Chama SURF\n");
    printf("SURF ent.pgm sai.key\n");
    erro("Erro: Numero de argumentos invalido");
  }

  initModule_nonfree(); // to load SURF/SIFT etc.

  Mat a=imread(argv[1],0); 
  Mat mask(a.size(),CV_8UC1,255);
  vector<KeyPoint> keypoints;
  vector<float> descriptors;

  SURF s;
  s(a,mask,keypoints,descriptors);
  if (s.descriptorSize()!=128) erro("Erro: descriptorSize!=128");

  FILE* arq=fopen(argv[2],"wt");
  fprintf(arq,"%u %d\n",unsigned(keypoints.size()),s.descriptorSize());

  unsigned j=0;
  for (unsigned i=0; i<keypoints.size(); i++) {
    KeyPoint& k=keypoints[i];
    fprintf(arq,"%g %g %g %g\n",k.pt.y,k.pt.x,k.size/6,deg2rad(k.angle));
    fprintf(arq,"  ");
    for (int l=0; l<s.descriptorSize(); l++) {
      fprintf(arq,"%g ",descriptors[j++]);
    }
    fprintf(arq,"\n");
  }

  fclose(arq);

}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

/*
void showkey(int argc, char** argv)
{ 
  if (argc!=4) {
    printf("ShowKey img.pgm img.key sai.ppm\n");
    erro("Erro: Numero de argumentos invalido");
  }
  IMGCOR img; pvLe(img,argv[1]);

  FILE* arq=fopen(argv[2],"rt");
  int n,m;
  if (fscanf(arq,"%d %d",&n,&m)!=2) erro("Erro leitura");
  if (m!=128) erro("Erro numero de caractericas");

  for (int i=0; i<n; i++) {
    float l,c,s,a;
    if (fscanf(arq,"%f %f %f %f",&l,&c,&s,&a)!=4) erro("Erro formato");
    relogiocircular(img,arredonda(l),arredonda(c),arredonda(1.5*s),
                        rad2deg(a),COR(255,0,0),1);
    for (int j=0; j<128; j++) 
      if (fscanf(arq,"%d",&m)!=1) erro("Erro caracteristica");
  }
  fclose(arq);
  pvImp(img,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
typedef struct {
  double l1, c1;     // Subpixel location of keypoint. 
  double s1, a1;     // Scale and orientation (range [-PI,PI])
  double l2, c2;     // Subpixel location of keypoint. 
  double s2, a2;     // Scale and orientation (range [-PI,PI])
  bool casado;
  double d;

  double s() {
    return s2/s1;
  }
  double a() {
    return modulo(a2-a1,M_2PI);
  }
} CASAMENTO;

typedef struct {
  double l, c, s, a, d;
} CASANORM;

vector<CASAMENTO> LeCasamento(string nome)
{ FILE *arq=fopen(nome.c_str(), "rt");
  if (!arq) erro("Could not open file ",nome);

  int n;
  if (fscanf(arq,"%d",&n) != 1)
    erro("Invalid file beginning.");
  vector<CASAMENTO> v(n);
 
  for (int i=0; i<n; i++) {
    int ch=0;
    CASAMENTO& m=v[i];
    if (fscanf(arq,"%lf %lf %lf %lf %lf %lf %lf %lf %c %lf",&m.l1,&m.c1,&m.s1,&m.a1,&m.l2,&m.c2,&m.s2,&m.a2,&ch,&m.d) != 10)
      erro("Invalid match file format.");
    //printf("%lf %lf %lf %lf %lf %lf %lf %lf %c %lf",m.l1,m.c1,m.s1,m.a1,m.l2,m.c2,m.s2,m.a2,ch,m.d);
    m.casado=(ch=='*');
  }
 
  fclose(arq);
  return v;
}

void ImpCasamento(vector<CASAMENTO> v, string nome)
{ FILE *arq=fopen(nome.c_str(), "wt");
  if (!arq) erro("Could not open file ",nome);

  fprintf(arq,"%d\n",v.size());
  for (unsigned i=0; i<v.size(); i++) {
    CASAMENTO& m=v[i];
    char ch;
    if (m.casado) ch='*'; else ch='_';
    fprintf(arq,"%10.2f %10.2f %10.3f %10.2f   %10.2f %10.2f %10.3f %10.2f %c %10.2f\n",m.l1,m.c1,m.s1,m.a1,m.l2,m.c2,m.s2,m.a2,ch,m.d);
  }
  fclose(arq);
}

void ImpCasanorm(vector<CASANORM> casanorm, string nome)
{ FILE* sai=fopen(nome.c_str(),"wt");
  fprintf(sai,"%d\n",casanorm.size());
  for (unsigned i=0; i<casanorm.size(); i++)
    fprintf(sai,"%12g %12g %12g %12g %12g\n",casanorm[i].l,casanorm[i].c,casanorm[i].s,rad2deg(casanorm[i].a),casanorm[i].d);
  fclose(sai);
}

vector<CASANORM> LeCasanorm(string nome)
{ FILE *arq=fopen(nome.c_str(), "rt");
  if (!arq) erro("Could not open file ",nome);

  int n;
  if (fscanf(arq,"%d",&n) != 1)
    erro("Invalid file beginning.");
  vector<CASANORM> v(n);
 
  for (int i=0; i<n; i++) {
    CASANORM& m=v[i];
    if (fscanf(arq,"%lf %lf %lf %lf %lf",&m.l,&m.c,&m.s,&m.a,&m.d) != 5)
      erro("Invalid match file format.");
    m.a=deg2rad(m.a);
  }
 
  fclose(arq);
  return v;
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<,
namespace Match {

// Data structure for a keypoint.  Lists of keypoints are linked
//   by the "next" field.
typedef struct KeypointSt {
  float row, col;             // Subpixel location of keypoint. 
  float scale, ori;           // Scale and orientation (range [-PI,PI]) 
  unsigned char *descrip;     // Vector of descriptor values 
  struct KeypointSt *next;    // Pointer to next keypoint in list. 
} *Keypoint;

Keypoint ReadKeys(FILE *fp)
{
  int i, j, num, len, val;
  Keypoint k, keys = NULL;

  if (fscanf(fp, "%d %d", &num, &len) != 2)
    erro("Invalid keypoint file beginning.");

  if (len != 128)
    erro("Keypoint descriptor length invalid (should be 128).");

  for (i = 0; i < num; i++) {
    // Allocate memory for the keypoint. 
    k = (Keypoint) malloc(sizeof(struct KeypointSt));
    k->next = keys;
    keys = k;
    k->descrip = (unsigned char*)malloc(len);

    if (fscanf(fp, "%f %f %f %f", &(k->row), &(k->col), &(k->scale),
        &(k->ori)) != 4)
      erro("Invalid keypoint file format.");

    for (j = 0; j < len; j++) {
      if (fscanf(fp, "%d", &val) != 1 || val < 0 || val > 255)
        erro("Invalid keypoint file value.");
      k->descrip[j] = (unsigned char) val;
    }
  }
  return keys;
}

Keypoint ReadKeyFile(char *filename)
{
  FILE *file;

  file = fopen (filename, "r");
  if (! file)
    erro("Could not open file", filename);

  return ReadKeys(file);
}

// Return squared distance between two keypoint descriptors.
int DistSquared(Keypoint k1, Keypoint k2)
{
    int i, dif, distsq = 0;
    unsigned char *pk1, *pk2;

    pk1 = k1->descrip;
    pk2 = k2->descrip;

    for (i = 0; i < 128; i++) {
      dif = (int) *pk1++ - (int) *pk2++;
      distsq += dif * dif;
    }
    return distsq;
}


// This searches through the keypoints in klist for the two closest
//   matches to key.  If the closest is less than 0.6 times distance to
//   second closest, then return the closest match.  Otherwise, return
//   NULL.
Keypoint CheckForMatch(Keypoint key, Keypoint klist, double limiar, bool& casado, int& dist)
{
  int dsq, distsq1 = 100000000, distsq2 = 100000000;
  Keypoint k, minkey = NULL;

  // Find the two closest matches, and put their squared distances in
  // distsq1 and distsq2.
  for (k = klist; k != NULL; k = k->next) {
    dsq = DistSquared(key, k);

    if (dsq < distsq1) {
      distsq2 = distsq1;
      distsq1 = dsq;
      minkey = k;
    } else if (dsq < distsq2) {
      distsq2 = dsq;
    }
  }
  if (sqrt(double(distsq1)/double(distsq2)) < limiar) casado=true;
  else casado=false;
  dist=distsq1;

  return minkey;
}

// Given a pair of images and their keypoints, pick the first keypoint
//   from one image and find its closest match in the second set of
//   keypoints.  Then write the result to a file.
vector<CASAMENTO> FindMatches(Keypoint keys1, Keypoint keys2, double limiar)
{ vector<CASAMENTO> v;
  Keypoint k, match;
  int count = 0;

  // Match the keys in list keys1 to their best matches in keys2.
  for (k= keys1; k != NULL; k = k->next) {
    bool casado; int dist;
    match = CheckForMatch(k, keys2, limiar, casado, dist);

    count++;
    CASAMENTO m;
    m.l1=k->row; m.c1=k->col; m.s1=k->scale; m.a1=k->ori;
    m.l2=match->row; m.c2=match->col; m.s2=match->scale; m.a2=match->ori;
    m.casado=casado; m.d=sqrt(double(dist));
    v.push_back(m);
  }
  return v;  
}

}

void match(int argc, char **argv)
{ using namespace Match;

  if (argc<4) {
    printf("Match: Acha casamentos de dois conjuntos de keypoints de SIFT\n");
    printf("Match key1.key key2.key result.res [limiar]\n");
    printf("  limiar: entre 0 e 1.\n");
    printf("  Parametros originais do Lowe: limiar=0.6 mododet=l modolink=r\n");
    erro("Erro: Numero de argumentos invalido");
  }
  Keypoint k1 = ReadKeyFile(argv[1]);
  Keypoint k2 = ReadKeyFile(argv[2]);

  double limiar=0.6;
  if (argc>=5) 
    if (sscanf(argv[4],"%lf",&limiar)!=1) erro("Erro: Leitura limiar");

  vector<CASAMENTO> v=FindMatches(k1, k2, limiar);
  ImpCasamento(v,argv[3]);
}



//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

IMGCOR FindMatches(IMGCOR im1, IMGCOR im2, vector<CASAMENTO>& match, char modolink, COR cor1, COR cor2)
{
  IMGCOR result=grudav(im1,im2);
  for (unsigned k=0; k<match.size(); k++) {
    CASAMENTO m=match[k];
    if (m.casado) {
      if (modolink=='r') {
        reta(result, arredonda(m.l1), arredonda(m.c1), arredonda(m.l2)+im1.nl(), arredonda(m.c2), cor1);
      } else if (modolink=='q') {
        ponto(result, arredonda(m.l1), arredonda(m.c1), cor1, 7);
        result(arredonda(m.l1), arredonda(m.c1))= cor2;
        ponto(result, arredonda(m.l2+im1.nl()), arredonda(m.c2), cor1, 7);
        result(arredonda(m.l2)+im1.nl(), arredonda(m.c2)) = cor2;
      } else if (modolink=='c') {
        int raio=arredonda(1.0*m.s1);
        double deg=rad2deg(m.a1);
        relogiocircular(result,arredonda(m.l1),arredonda(m.c1),raio,deg,cor1);
        raio=arredonda(1.0*m.s2);
        deg=rad2deg(m.a2);
        relogiocircular(result,arredonda(m.l2)+im1.nl(),arredonda(m.c2),raio,deg,cor1);
      } else erro("Erro: Modolink invalido");
    }
  }
  return result;
}

void showmat(int argc, char **argv)
{ if (argc!=5 && argc!=6 && argc!=7) {
    printf("ShowMat: mostra os casamentos de keypoints gerados pelo ColSIFT/ColMatch.\n");
    printf("ShowMat im1.ppm im2.ppm match.res result.ppm [modolink] [cor]\n");
    printf("  modolink: r=reta_interliga q=pto_quadrado c=circulo\n");
    printf("  cor = r g b c m y\n");
    erro("Erro: Numero de argumentos invalido");
  }

  IMGCOR im1,im2; IMGGRY g1,g2;
  pvLe(g1,argv[1]); im1=g1;
  pvLe(g2,argv[2]); im2=g2;

  vector<CASAMENTO> match=LeCasamento(argv[3]);
  char modolink='r';
  if (argc>=6) modolink=tolower(argv[5][0]);

  COR cor1=COR(255,0,0); COR cor2=COR(0,255,255); 
  if (argc>=7) { 
    char ch=tolower(argv[9][0]);
    if      (ch=='r') { cor1=COR(255,0,0); cor2=COR(0,255,255); }
    else if (ch=='g') { cor1=COR(0,255,0); cor2=COR(255,0,255); }
    else if (ch=='b') { cor1=COR(0,0,255); cor2=COR(255,0,255); }
    else if (ch=='c') { cor1=COR(0,255,255); cor2=COR(255,0,0); }
    else if (ch=='m') { cor1=COR(255,0,255); cor2=COR(0,255,0); }
    else if (ch=='y') { cor1=COR(255,255,0); cor2=COR(0,0,255); }
    else erro("Erro: Cor invalido");
  }
  IMGCOR sai=FindMatches(im1, im2, match, modolink, cor1, cor2);
  pvImp(sai,argv[4]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

vector<CASANORM> constroicasanorm(vector<CASAMENTO> &match, IMGCOR& q)
{ vector<CASANORM> casanorm;

  for (unsigned i=0; i<match.size(); i++) {
    if (match[i].casado) {
      double dl=q.lc()-match[i].l1;
      double dc=q.cc()-match[i].c1;

      MATRIZ<double> s(3,3, 0.0);
      s(0,0)=s(1,1)=match[i].s();
      s(2,2)=1.0;

      double theta=match[i].a();
      MATRIZ<double> r(3,3, 0.0);
      r(0,0)=cos(theta);  r(0,1)=sin(theta);
      r(1,0)=-sin(theta); r(1,1)=cos(theta);
                                              r(2,2)=1.0;

      MATRIZ<double> d2(1,3, dl, dc, 1.0);
      d2=d2*s*r;
      MATRIZ<double> p2(1,3, d2(0)+match[i].l2, d2(1)+match[i].c2, 1.0);
  
      CASANORM h; h.l=p2(0); h.c=p2(1); h.s=match[i].s(); h.a=match[i].a(); h.d=match[i].d;
      casanorm.push_back(h);
    }
  }
  return casanorm;
}

void normcasa(int argc, char **argv)
{ if (argc!=4) {
    printf("NormCasa: Normaliza casamentos Q->A de colsift/colmatch.\n");
    printf("NormCasa q.ppm match.res casanorm.txt\n");
    erro("Erro: Numero de argumentos invalido");
  }

  IMGCOR q;
  pvLe(q,argv[1]);

  vector<CASAMENTO> match=LeCasamento(argv[2]);
  vector<CASANORM> casanorm=constroicasanorm(match,q);
  ImpCasanorm(casanorm,argv[3]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

bool mesmogrupo(CASANORM q, double maxdifsf, double maxdifang, double maxdifpix, PONTO2 mediapix, double mediasf, CPXD mediaa)
{ if (difangrad(arg(mediaa),q.a)>maxdifang) return false;
  PONTO2 p(q.l,q.c);
  if (mediasf*distancia(mediapix,p)>maxdifpix) return false;
  if (mediasf*fabs(mediasf-q.s)>maxdifsf) return false;
  return true;
}

CASANORM vhough(vector<CASANORM> casanorm, double maxdifsf, double maxdifang, double maxdifpix)
{ 
  if (casanorm.size()==0) {
    CASANORM q; q.l=0; q.c=0; q.s=1; q.a=0; q.d=0;
    return q;
  }

  // converte distancia em peso. Menor distancia => peso 1. Maior distancia => peso menorpeso.
  double menorpeso=0.6; 
  double maiordist=-infinito;
  double menordist=infinito;
  for (unsigned i=0; i<casanorm.size(); i++) {
    if (casanorm[i].d>maiordist) maiordist=casanorm[i].d;
    if (casanorm[i].d<menordist) menordist=casanorm[i].d;
  }
  if (maiordist-menordist>epsilon) {
    for (unsigned i=0; i<casanorm.size(); i++) 
      casanorm[i].d=(maiordist-casanorm[i].d)*(1.0-menorpeso)/(maiordist-menordist)+menorpeso;
  } else {
    for (unsigned i=0; i<casanorm.size(); i++) 
      casanorm[i].d=1.0;
  }
  //for (unsigned i=0; i<casanorm.size(); i++) {
  //  CASANORM& m=casanorm[i];
  //  printf("l=%12g c=%12g a=%12g s=%12g d=%12g\n",m.l,m.c,rad2deg(m.a),m.s,m.d);
  //}

  // comeca vhough
  VETOR<int> classe(casanorm.size(),-1);
  int nclasse=0;

  CASANORM& m=casanorm[0];
  double maxsomapeso=m.d;
  PONTO2 maxpix(m.l,m.c);
  double maxsf=m.s;
  CPXD maxa=polar(m.d,m.a);

  for (unsigned i=0; i<casanorm.size(); i++) { 
    if (classe(i)==-1) {
      nclasse++; classe(i)=nclasse-1;

      m=casanorm[i];
      double somapeso=m.d; 
      PONTO2 somapix=m.d*PONTO2(m.l,m.c); PONTO2 mediapix(m.l,m.c);
      double somasf=m.d*m.s; double mediasf=m.s;
      CPXD somaa=polar(m.d,m.a); CPXD mediaa=polar(1.0,m.a);

      for (unsigned j=i+1; j<casanorm.size(); j++) {
        if (classe(j)==-1 && 
            mesmogrupo(casanorm[j],maxdifsf,maxdifang,maxdifpix,mediapix,mediasf,mediaa)) {
          classe(j)=nclasse-1;
          CASANORM& p=casanorm[j];
          somapeso+=p.d; 
          somapix=somapix+p.d*PONTO2(p.l,p.c); mediapix=(1.0/somapeso)*somapix;
          somasf+=p.d*p.s; mediasf=(1.0/somapeso)*somasf;
          somaa+=polar(p.d,p.a); 

          if (somapeso>maxsomapeso) {
            maxsomapeso=somapeso;
            maxpix=mediapix;
            maxa=somaa;
            maxsf=mediasf;
          }
        }
      }
    }
  }
  CASANORM q; q.l=maxpix(0); q.c=maxpix(1); q.s=maxsf; q.a=arg(maxa); q.d=maxsomapeso;
  return q;
}

void vhough(int argc, char **argv)
{ if (argc<5) {
    printf("VHough: Reconhece 1 instancia de Q colsift/colmatch/normcasa.\n");
    printf("VHough q.ppm a.ppm casanorm.txt p.ppm [maxdifsf] [maxdifang] [maxdifpix]\n");
    printf("  maxdifsf. Default=0.3 [*escalamedia]\n");
    printf("  maxdifang em graus. Default=15 graus\n");
    printf("  maxdifpix default=7 [*escalamedia]\n");
    erro("Erro: Numero de argumentos invalido");
  }

  IMGGRY qg,ag; IMGCOR q,a;
  pvLe(qg,argv[1]); q=qg;
  pvLe(ag,argv[2]); a=ag;

  vector<CASANORM> casanorm=LeCasanorm(argv[3]);

  double maxdifsf=0.3;
  if (argc>=6) if (sscanf(argv[5],"%lf",&maxdifsf)!=1) erro("Erro: Leitura maxdifsf");

  double maxdifang=15;
  if (argc>=7) if (sscanf(argv[6],"%lf",&maxdifang)!=1) erro("Erro: Leitura maxdifang"); 
  maxdifang=deg2rad(maxdifang);

  double maxdifpix=7;
  if (argc>=8) if (sscanf(argv[7],"%lf",&maxdifpix)!=1) erro("Erro: Leitura maxdifpix");

  CASANORM m=vhough(casanorm,maxdifsf,maxdifang,maxdifpix);
  //printf("l=%12g c=%12g a=%12g s=%12g d=%12g\n",m.l,m.c,rad2deg(m.a),m.s,m.d);
  relogioretangular(a,cvRound(m.l),cvRound(m.c),cvRound(m.s*q.nl()),cvRound(m.s*q.nc()),rad2deg(m.a),COR(255,0,0),2);
  puttxt(a,cvRound(m.l)+1,cvRound(m.c)+1,semsufixo(argv[1]),COR(0,0,255));
  pvImp(a,argv[4]);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void find_obj(int argc, char **argv)
{ if (argc<4) {
    printf("Find_obj: Executa sift, match, normcasa e vhough e reconhece 1 obj.\n");
    printf("Find_obj q.ppm a.ppm p.ppm [maxdifsf] [maxdifang] [maxdifpix]\n");
    printf("  maxdifsf. Default=0.3 [*escalamedia]\n");
    printf("  maxdifang em graus. Default=15 graus\n");
    printf("  maxdifpix default=7 [*escalamedia]\n");
    erro("Erro: Numero de argumentos invalido");
  }

  char st[1024];
  sprintf(st,"sift sift %s sifttemp_q.key",argv[1]); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");

  sprintf(st,"sift sift %s sifttemp_a.key",argv[2]); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");

  sprintf(st,"sift match sifttemp_q.key sifttemp_a.key sifttemp.res"); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");

  sprintf(st,"sift normcasa %s sifttemp.res sifttemp.txt",argv[1]); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");

  sprintf(st,"sift vhough %s %s sifttemp.txt %s",argv[1],argv[2],argv[3]); 
  for (int i=4; i<argc; i++) {
    strcat(st," "); strcat(st,argv[i]);
  }
  printf("%s\n",st); if (system(st)==-1) erro("Erro");

  remove("sifttemp_q.key");
  remove("sifttemp_a.key");
  remove("sifttemp.res");
  remove("sifttemp.txt");
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void vhough(IMGCOR& a, string nomeq, string nomecasanorm)
{ IMGCOR q;
  pvLe(q,nomeq);

  vector<CASANORM> casanorm=LeCasanorm(nomecasanorm);

  double maxdifsf=0.3;
  double maxdifang=15;
  double maxdifpix=7;

  CASANORM m=vhough(casanorm,maxdifsf,maxdifang,maxdifpix);
  //printf("l=%12g c=%12g a=%12g s=%12g d=%12g\n",m.l,m.c,rad2deg(m.a),m.s,m.d);
  relogioretangular(a,cvRound(m.l),cvRound(m.c),cvRound(m.s*q.nl()),cvRound(m.s*q.nc()),rad2deg(m.a),COR(255,0,0),2);
  puttxt(a,cvRound(m.l)+1,cvRound(m.c)+1,semsufixo(nomeq),COR(0,0,255));
}

void findnobj(int argc, char **argv)
{ if (argc<4) {
    printf("FindNobj: Executa sift, match, normcasa e vhough e reconhece n objs.\n");
    printf("FindNobj p.ppm a.pgm q1.pgm q2.pgm ...\n");
    erro("Erro: Numero de argumentos invalido");
  }

  char st[1024];
  sprintf(st,"sift sift %s sifttemp_a.key",argv[2]); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");
  IMGGRY ag; IMGCOR a; pvLe(ag,argv[2]); a=ag;

  for (int i=3; i<argc; i++) {
    sprintf(st,"sift sift %s sifttemp_q.key",argv[i]); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    sprintf(st,"sift match sifttemp_q.key sifttemp_a.key sifttemp.res"); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    sprintf(st,"sift normcasa %s sifttemp.res sifttemp.txt",argv[i]); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    vhough(a, argv[i], "sifttemp.txt");
  }
  pvImp(a,argv[1]);

  remove("sifttemp_q.key");
  remove("sifttemp_a.key");
  remove("sifttemp.res");
  remove("sifttemp.txt");
}

void findnth(int argc, char **argv)
{ if (argc<5) {
    printf("FindNth: fimdnobj com threshold.\n");
    printf("FindNth threshold p.ppm a.pgm q1.pgm q2.pgm ...\n");
    erro("Erro: Numero de argumentos invalido");
  }

  char st[1024];
  sprintf(st,"sift sift %s sifttemp_a.key",argv[3]); 
  printf("%s\n",st); if (system(st)==-1) erro("Erro");
  IMGGRY ag; IMGCOR a; pvLe(ag,argv[3]); a=ag;

  for (int i=4; i<argc; i++) {
    sprintf(st,"sift sift %s sifttemp_q.key",argv[i]); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    sprintf(st,"sift match sifttemp_q.key sifttemp_a.key sifttemp.res %s",argv[1]); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    sprintf(st,"sift normcasa %s sifttemp.res sifttemp.txt",argv[i]); 
    printf("%s\n",st); if (system(st)==-1) erro("Erro");

    vhough(a, argv[i], "sifttemp.txt");
  }
  pvImp(a,argv[2]);

  remove("sifttemp_q.key");
  remove("sifttemp_a.key");
  remove("sifttemp.res");
  remove("sifttemp.txt");
}
*/


//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< SURF: Programas para rodar SURF de OpenCV21, v1.00>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"SURF     - chama SURF para gerar keypoints\n"
//"ShowKey  - Mostra keypoints\n"
//"Match    - acha os casamentos entre dois conjuntos de keypoints\n"
//"ShowMat  - mostra os casamentos de colsift/colmatch\n"
//"NormCasa - normaliza os casamentos de colsift/colmatch QA\n"
//"VHough   - Vector Hough QA\n"
//"Find_obj - Executa sift, match, normcasa e vhough e reconhece 1 obj\n"
//"FindNobj - Executa sift, match, normcasa e vhough e reconhece n objetos\n"
//"FindNth  - FindNobj com threshold\n"
"...............................................................................\n";

int main(int argc, char** argv)
{
  if (argc<2) {
    printf(about);
    printf("%s",help);
    erro("Erro: Numero de argumentos invalido");
  } else {
    string comando=minuscula(argv[1]);
    if      (comando=="surf")  surf(argc-1,&argv[1]);
    //else if (comando=="showkey")  showkey(argc-1,&argv[1]);
    //else if (comando=="match") match(argc-1,&argv[1]);
    //else if (comando=="showmat")  showmat(argc-1,&argv[1]);
    //else if (comando=="normcasa") normcasa(argc-1,&argv[1]);
    //else if (comando=="vhough")   vhough(argc-1,&argv[1]);
    //else if (comando=="find_obj") find_obj(argc-1,&argv[1]);
    //else if (comando=="findnobj") findnobj(argc-1,&argv[1]);
    //else if (comando=="findnth") findnth(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
  }
}


