// avx-th.cpp
// compila avx-th.cpp -cek -avx

// <mmintrin.h>  MMX
// <xmmintrin.h> SSE
// <emmintrin.h> SSE2
// <pmmintrin.h> SSE3
// <tmmintrin.h> SSSE3
// <smmintrin.h> SSE4.1
// <nmmintrin.h> SSE4.2
// <ammintrin.h> SSE4A
// <wmmintrin.h> AES
// <immintrin.h> AVX

#include <cekeikon.h>
#include <immintrin.h>
#include <thread>             

void inicializa(Mat_<float>& A, Mat_<float>& B, int n, int seed)
{ A.create(n,n);
  B.create(n,n);
  srand(seed);
  for (int l=0; l<n; l++)
    for (int c=0; c<n; c++)
      A(l,c)=rand()%n;
  for (int l=0; l<n; l++)
    for (int c=0; c<n; c++)
      B(l,c)=rand()%n;
}

void multMatConvencional(const Mat_<float>& A, const Mat_<float>& B, Mat_<float>& D)
{ D.create(A.rows,B.cols);
  for (int l=0; l<A.rows; l++)
    for (int c=0; c<B.cols; c++) {
      float d=0.0;
      for (int i=0; i<A.cols; i++)
        d=d+A(l,i)*B(i,c);
      D(l,c)=d;
    }
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

namespace AVXTH {

float* AN; float* BN; Mat_<float> DN;
int rows, cols, nextN, nLoop;
int nthreads;

void mult(int t)
{
  __m256 e;
  float* pe=(float*)&e;
  __m256 m;
  __m256* pa;
  __m256* pb;

  int d=teto(rows,nthreads);
  int inic=t*d;
  int fim=min( (t+1)*d, rows );

  for (int l=inic; l<fim; l++) {
    for (int c=0; c<cols; c++) { 
      e=_mm256_setzero_ps(); // e[0, 1, 2, 3] = 0.0
      pa = (__m256*) (&AN[l*nextN]);
      pb = (__m256*) (&BN[c*nextN]);
      for (int i=0; i<nLoop; i++) {
        //e=e+AN[l][i]*BN[c][i];
        m = _mm256_mul_ps(*pa, *pb);  // m = *pa * *pb
        e = _mm256_add_ps(e, m);      // e = e + m
        pa++;
        pb++;
      }
      DN(l,c)=pe[0]+pe[1]+pe[2]+pe[3]+pe[4]+pe[5]+pe[6]+pe[7];
    }
  }
}

} // namespace

void multMatAvxTh(const Mat_<float>& A, const Mat_<float>& B, Mat_<float>& D)
{ using namespace AVXTH;
  cols=B.cols;
  rows=A.rows;
  nextN=teto(cols,8)*8;
  nLoop=nextN/8;
  //printf("cols=%d nextN=%d nLoop=%d\cols",cols,nextN,nLoop);

  AN = (float*)_aligned_malloc(cols*nextN * sizeof(float), 32);
  if (AN==NULL) erro("Erro AN");
  BN = (float*)_aligned_malloc(cols*nextN * sizeof(float), 32);
  if (BN==NULL) erro("Erro BN");

  for (int l=0; l<cols; l++) {
    for (int c=0; c<cols; c++) AN[l*nextN+c]=A(l,c);
    for (int c=cols; c<nextN; c++) AN[l*nextN+c]=0.0f;
  }
  for (int l=0; l<cols; l++) {
    for (int c=0; c<cols; c++) BN[l*nextN+c]=B(c,l);
    for (int c=cols; c<nextN; c++) BN[l*nextN+c]=0.0f;
  }
  D.create(cols,cols); DN=D;

  nthreads=4;
  vector<thread> threads(nthreads);
  for (int t=0; t<nthreads; t++) threads[t]=thread(mult,t);
  for (int t=0; t<nthreads; t++) threads[t].join();

  _aligned_free(AN);
  _aligned_free(BN);
}

//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

int diferenca(const Mat_<float>& A, const Mat_<float>& B)
{ if (A.size()!=B.size()) erro("Erro diferenca");
  int conta=0;
  for (int l=0; l<A.rows; l++)
    for (int c=0; c<A.cols; c++)
      if (abs(A(l,c)-B(l,c))>abs(A(l,c)/100000.0)) conta++;
  return conta;
}

int main(int argc, char** argv) {
  if (argc!=3) {
    printf("AVX: Multiplica duas matrizes nxn\n");
    printf("  Utiliza instrucoes AVX 256 bits\n");
    printf("AVX n seed\n");
    printf("  As matrizes sao preenchidas com numeros de 0 a n-1\n");
    erro("Erro: Numero de argumentos invalido");
  }

  int n;
  convArg(n,argv[1]);
  int seed;
  convArg(seed,argv[2]);
  Mat_<float> A,B;
  inicializa(A,B,n,seed);

  int t1=centseg();
  Mat_<float> D(n,n,0.0f);
  multMatConvencional(A,B,D);
  int t2=centseg()-t1;
  printf("Tempo gasto convencional: %d.%02ds\n",t2/100,t2%100);

  int t3=centseg();
  Mat_<float> E;
  multMatAvxTh(A,B,E);
  int t4=centseg()-t3;
  printf("Tempo gasto AVX: %d.%02ds\n",t4/100,t4%100);

  int conta=diferenca(D,E);
  printf("Numero de elementos diferentes: %d\n",conta);
}
