// filtro3.cpp
// Filtro linear usando AVX e multi-thread
// compila filtro3.cpp -cek -avx
#include <cekeikon.h>
#include <immintrin.h>
#include <thread>             

void inicializa(Mat_<float>& A, int n, int seed)
{ A.create(n,n);
  mysrand(seed);
  for (int l=0; l<n; l++)
    for (int c=0; c<n; c++)
      A(l,c)=myuniform();
}

Mat_<float> filtroConvencional(const Mat_<float>& A, const Mat_<float>& F)
{ Mat_<float> AE(A.rows+F.rows-1, A.cols+F.cols-1, 0.0f);
  int lc=F.rows/2;
  int cc=F.cols/2;
  Mat_<float> roi(AE, Rect(lc,cc,A.rows,A.cols) );
  A.copyTo(roi);

  Mat_<float> D(A.rows,A.cols);
  float s;
  for (int l=0; l<A.rows; l++)
    for (int c=0; c<A.cols; c++) {
      s=0.0f;
      for (int l2=0; l2<F.rows; l2++)
        for (int c2=0; c2<F.cols; c2++) 
          s += AE(l+l2,c+c2)*F(l2,c2);
      D(l,c)=s;
    }
  return D;
}

namespace AVXTH {

Mat_<float> AE,D; float* FA; 
int Arows, Acols, Frows, nextCols, nthreads;

void filt(int t)
{
  int d=teto(Arows,nthreads);
  int inic=t*d;
  int fim=min( (t+1)*d, Arows );

  __m256 a,f,s,m;
  float* ps=(float*)(&s);
  for (int l=inic; l<fim; l++) {
    for (int c=0; c<Acols; c++) { 
      s=_mm256_setzero_ps(); // s[0,1,2,3,4,5,6,7] = 0.0
      for (int l2=0; l2<Frows; l2++)
        for (int c2=0; c2<nextCols; c2+=8) { 
          //s=s+AE[l+l2][c+c2]*FA[l2][c2];
          a=_mm256_loadu_ps(&AE(l+l2,c+c2)); // carga nao-alinhada
          f=_mm256_load_ps(&FA[nextCols*l2+c2]); // carga alinhada
          m=_mm256_mul_ps(a, f);  // m = a * f
          s=_mm256_add_ps(s, m);  // s = s + m
        }
      D(l,c)=ps[0]+ps[1]+ps[2]+ps[3]+ps[4]+ps[5]+ps[6]+ps[7];
    }
  }
}

} // namespace

Mat_<float> filtroAvx2(const Mat_<float>& A, const Mat_<float>& F)
{ using namespace AVXTH;
  Arows=A.rows; Acols=A.cols;
  Frows=F.rows; 
  nextCols=teto(F.cols,8)*8;

  AE.create(A.rows+F.rows-1,A.cols+F.cols-1); AE.setTo(0.0f);
  int lc=F.rows/2;
  int cc=F.cols/2;
  Mat_<float> roi(AE, Rect(lc,cc,A.rows,A.cols) );
  A.copyTo(roi);

  int nextRows=teto(F.rows,8)*8;
  int nextCols=teto(F.cols,8)*8;
  FA = (float*)_aligned_malloc( nextRows*nextCols*sizeof(float), 32 );
  assert(FA!=NULL);
  for (int i=0; i<nextRows*nextCols; i++) FA[i]=0.0f;
  for (int l=0; l<F.rows; l++)
    for (int c=0; c<F.cols; c++) 
      FA[l*nextCols+c]=F(l,c);

  D.create(A.rows,A.cols);

  nthreads=8;
  vector<thread> threads(nthreads);
  for (int t=0; t<nthreads; t++) threads[t]=thread(filt,t);
  for (int t=0; t<nthreads; t++) threads[t].join();

  _aligned_free(FA);
  return D;
}

int diferenca(const Mat_<float>& A, const Mat_<float>& B)
{ if (A.size()!=B.size()) erro("Erro diferenca: Dimensoes diferentes");
  int conta=0;
  for (int l=0; l<A.rows; l++)
    for (int c=0; c<A.cols; c++)
      if (abs(A(l,c)-B(l,c))>abs(A(l,c)/100000.0)) conta++;
  return conta;
}

int main(int argc, char** argv) {
  if (argc!=4) {
    printf("Filtro3: Aplica filtro mxm na imagem nxn\n");
    printf("  Utiliza instrucoes AVX 256 bits e multi-thread\n");
    printf("Filtro3 n m seed\n");
    printf("  A imagem e' preenchida com numeros de 0 a 1\n");
    erro("Erro: Numero de argumentos invalido");
  }

  int n; convArg(n,argv[1]);
  int m; convArg(m,argv[2]);
  int seed; convArg(seed,argv[3]);

  Mat_<float> A,F;
  inicializa(A,n,seed);
  inicializa(F,m,3*seed);
  F = somatoriaUm(F);

  int t1=centseg();
  Mat_<float> D1=filtroAvx2(A,F);
  int t2=centseg()-t1;
  printf("Tempo gasto AVX: %d.%02ds\n",t2/100,t2%100);

  t1=centseg();
  Mat_<float> D2;
  filter2D(A,D2,-1,F,Point(-1,-1),0,BORDER_CONSTANT);
  t2=centseg()-t1;
  printf("Tempo gasto filter2D: %d.%02ds\n",t2/100,t2%100);

  t1=centseg();
  Mat_<float> D3=filtroConvencional(A,F);
  t2=centseg()-t1;
  printf("Tempo gasto convencional: %d.%02ds\n",t2/100,t2%100);

  int conta=diferenca(D1,D2);
  printf("Numero de elementos diferentes D1 D2: %d\n",conta);

  conta=diferenca(D1,D3);
  printf("Numero de elementos diferentes D1 D3: %d\n",conta);
}
