#include <cekeikon.h>
#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <MMSystem.h>
#include <windows.h>
#include <vector>

// #include "..\imagens2.h"

float raio=1; // raio da esfera (metros)
float lado2=5;  // lado dividido por dois da caixa (metros)
float coefRest=0.99;

int nbolas=3;
vector<Vec3f> so(nbolas); // posicoes iniciais
vector<Vec3f> vo(nbolas); // velocidade inicial em m/s
vector<Vec3f> ao(nbolas); // aceleracao da gravidade m2/s
vector<Vec3f> sa(nbolas); // posicao atual da esfera
vector<Vec3f> va(nbolas); // velocidade atual da esfera
vector<float> t(nbolas,0.0);

int fps=50;
float duracaoMs=1000.0/fps; //40ms - usado no timer
float duracaoS=(1.0/fps); //0.04s - usado para calcular posicao/velocidade

float olho=0;
float olhoM=0.2;

float vlim=0.5; //menor velocidade para continuar em m/s

void bateuCaixa(int i)
{ bool bateu=false;

  Vec3f vt=vo[i]+ao[i]*t[i]; // velocidade atual temporaria - pode mudar
  if (sa[i][0]<-lado2+raio && vt[0]<0)      { vt[0]=-vt[0]; bateu=true; }
  else if (lado2-raio<sa[i][0] && vt[0]>0)  { vt[0]=-vt[0]; bateu=true; } 

  if (sa[i][1]<-lado2+raio && vt[1]<0)      { vt[1]=-vt[1]; bateu=true; } 
  else if (lado2-raio<sa[i][1] && vt[1]>0)  { vt[1]=-vt[1]; bateu=true; } 

  if (sa[i][2]<-lado2+raio && vt[2]<0)      { vt[2]=-vt[2]; bateu=true; } 
  else if (lado2-raio<sa[i][2] && vt[2]>0)  { vt[2]=-vt[2]; bateu=true; } 

  if (bateu) {
    so[i]=sa[i];
    vo[i]=coefRest*vt;
    t[i]=0;
    //MessageBeep(0x00000010L);
    //printf("bateuCaixa\n");
  }
}

void bateuBola(int i, int j)
{ bool bateu=false;
  float dist=norm(sa[i]-sa[j]); // distancia atual
  if (dist<2*raio) { // pode ter batido
    float ti=t[i]-duracaoS;
    float tj=t[j]-duracaoS;
    Vec3f si = so[i] + vo[i]*ti + 0.5*elev2(ti)*ao[i]; // posicao instante anterior i
    Vec3f sj = so[j] + vo[j]*tj + 0.5*elev2(tj)*ao[j]; 
    float distAnt = norm(si-sj); // distancia anterior
    if (distAnt>dist) { // bateu com certeza, pois bolas estao aproximando
      Vec3f P=0.5*(sa[i]+sa[j]); // ponto de batida
      Vec3f Ni=versor(P-sa[i]); // normal a superficie da bola i - aponta em direcao a P
      Vec3f Nj=versor(P-sa[j]); 
      Vec3f Ui=va[i]; // velocidade i antes da colisao
      Vec3f Uj=va[j]; 
      Vec3f Wi=(Ui*Ni)*Ni; // velocidade i normal a superficie antes da colisao
      Vec3f Wj=(Uj*Nj)*Nj;
      Vec3f Vi=Ui-Wi; // velocidade i paralela a superficies antes da colisao
      Vec3f Vj=Uj-Wj; 
      Vec3f Xi = 0.5 * (Wi+Wj+coefRest*(Wj-Wi)); // velocidade normal a superficie depois da colisao
      Vec3f Xj = 0.5 * (Wj+Wi+coefRest*(Wi-Wj)); 

      vo[i] = Vi + Xi; // velocidade depois da colisao
      vo[j] = Vj + Xj;
      so[i] = sa[i];
      so[j] = sa[j];
      t[i] = 0;
      t[j] = 0;

      bateu=true;
    }
  }
  if (bateu) { 
    //MessageBeep(0x00000000L);
    //printf("bateuBola\n");
  }
}

void spinDisplay(int value)
{ sa[0] = so[0] + vo[0]*t[0] + 0.5*elev2(t[0])*ao[0];
  sa[1] = so[1] + vo[1]*t[1] + 0.5*elev2(t[1])*ao[1];
  sa[2] = so[2] + vo[2]*t[2] + 0.5*elev2(t[2])*ao[2];

  va[0] = vo[0] + t[0]*ao[0];
  va[1] = vo[1] + t[1]*ao[1];
  va[2] = vo[2] + t[2]*ao[2];

  bateuBola(0,1);
  bateuBola(0,2);
  bateuBola(1,2);

  bateuCaixa(0);
  bateuCaixa(1);
  bateuCaixa(2);

  t[0] += duracaoS;
  t[1] += duracaoS;
  t[2] += duracaoS;

  olho += olhoM;
  if (olho>90) olhoM=-olhoM;
  else if (olho<0) olhoM=-olhoM;

  if (norm(va[0])>vlim || norm(va[1])>vlim || norm(va[2])>vlim) { // condicao para continuar
    glutTimerFunc(round(duracaoMs),spinDisplay,0);
    glutPostRedisplay();
  } else exit(0);
}

void caixa(void)
{ 
  glutWireCube(10);
}

void esfera(void)
{ glutSolidSphere(raio,30,24); }

void desenha(void (&f)(void), float red, float gre, float blu)
{ 
  GLfloat mat_ambient[] = { red/2, gre/2, blu/2, 1.0 };
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  GLfloat mat_diffuse[] = { red, gre, blu, 1.0 };
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
  GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  GLfloat mat_shininess[] = { 50.0 };
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  f();
}

void display(void)
{ glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(3.5*lado2, 2*lado2, 0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
  glRotatef(olho,0,1,0);

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  desenha(caixa,0,0,0);

  glPushMatrix();
  glTranslatef(sa[0][0],sa[0][1],sa[0][2]); 
  desenha(esfera,0.5,1,1);
  glPopMatrix();

  glPushMatrix();
  glTranslatef(sa[1][0],sa[1][1],sa[1][2]); 
  desenha(esfera,1,0.5,1);
  glPopMatrix();

  glPushMatrix();
  glTranslatef(sa[2][0],sa[2][1],sa[2][2]); 
  desenha(esfera,1,1,0.5);
  glPopMatrix();

  glutSwapBuffers();
}

void init(void) 
{ so[0]=Vec3f(-lado2+raio, lado2-raio,  lado2-raio);
  so[1]=Vec3f( lado2-raio, lado2-raio,  lado2-raio);
  so[2]=Vec3f( lado2-raio, lado2-raio, -lado2+raio);

  vo[0]=12.0*Vec3f( 0.7,-0.2, -0.5); 
  vo[1]=12.0*Vec3f(-1.0,-0.8, -0.7); 
  vo[2]=12.0*Vec3f( 0.8,-0.5,  0.9); 

  ao[0]=Vec3f(0,-10,0);
  ao[1]=Vec3f(0,-10,0);
  ao[2]=Vec3f(0,-10,0);

  sa[0]=so[0]; 
  sa[1]=so[1]; 
  sa[2]=so[2]; 

  va[0]=vo[0]; 
  va[1]=vo[1]; 
  va[2]=vo[2]; 

  glClearColor(1.0, 1.0, 1.0, 1.0); 
  glShadeModel(GL_SMOOTH);
  
  GLfloat light_position[] = { 3.0, 5.0, 3.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE); // true e false estao invertidos
}

void reshape(int w, int h)
{ glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, (GLfloat) w/(GLfloat) h, lado2/2.0, 10.0*lado2); 
}

void keyboard(unsigned char key, int x, int y)
{
  if(key==27) exit(0);
}

int main(int argc, char** argv)
{ glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);
  glutInitWindowSize(640+320, 480+240); 
  glutInitWindowPosition(0, 0);
  glutCreateWindow(argv[0]);
  init();
  glutDisplayFunc(display); 
  glutReshapeFunc(reshape); 
  glutTimerFunc(40,spinDisplay,0);
  glutKeyboardFunc(keyboard);
  glutMainLoop();
  return 0;   
}
