// Os vertices e normais sao precalculados
// Cada retangulo e' renderizado como um quadrilatero
// com modelo de suavizacao de Gouraud
// Utiliza lista para acelerar - nao reparei nenhuma diferenca em usar lista.
#include <GL/glut.h>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cekeikon.h>

int fps=60;
float duracaoMs=1000.0/fps; //40ms
float duracaoS=1.0/fps; //0.04s
float olho=0;
float olhoM=0.5;

float lado2=10;  // lado dividido por dois. minx, maxx, miny, maxy.
float delta=0.1; // passo
float delta2=delta/2; // passo
int n=2*round(lado2/delta)+1;

GLuint theFunction;

Mat_<Vec3f> vertices(n,n,Vec3f(0,0,0));
Mat_<Vec3f> norquad(n-1,n-1,Vec3f(0,0,0));
Mat_<Vec3f> norvert(n,n,Vec3f(0,0,0));

void spinDisplay(int value)
{ olho += olhoM;
  if (olho>360) olho=olho-360;
  glutTimerFunc(round(duracaoMs),spinDisplay,0);
  glutPostRedisplay();
}

float f(float x, float y)
{ float r=sqrt(3*x*x+y*y);
  return 14*cos(1.2*r)/(r+3);
}

void solidFunction(void)
{ for (int xi=0; xi<n-1; xi++)
    for (int yi=0; yi<n-1; yi++) {
      glBegin(GL_QUADS);  
        glNormal3fv(&norvert(xi,yi)[0]); 
        glVertex3fv(&vertices(xi,yi)[0]); 

        glNormal3fv(&norvert(xi+1,yi)[0]); 
        glVertex3fv(&vertices(xi+1,yi)[0]); 

        glNormal3fv(&norvert(xi+1,yi+1)[0]); 
        glVertex3fv(&vertices(xi+1,yi+1)[0]); 

        glNormal3fv(&norvert(xi,yi+1)[0]); 
        glVertex3fv(&vertices(xi,yi+1)[0]); 
      glEnd();
    }
}

void display(void)
{ 
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glCallList(theFunction);
  glutSwapBuffers();
}

void init(void) 
{ 
  int xi,yi; float x,y;
  for (x=-lado2, xi=0; xi<n; x+=delta, xi++) {
    for (y=-lado2, yi=0; yi<n; y+=delta, yi++) {
      vertices(xi,yi)=Vec3f(x,y,f(x,y));
    }
  }

  for (xi=0; xi<n-1; xi++)
    for (yi=0; yi<n-1; yi++) {
      Vec3f a=vertices(xi,yi);
      Vec3f b=vertices(xi+1,yi);
      Vec3f c=vertices(xi+1,yi+1);
      Vec3f d=vertices(xi,yi+1);
      Vec3f n=normalTri(a,b,d);
      Vec3f m=normalTri(b,c,d);
      n = 0.5 * (n+m);
      norquad(xi,yi)=n;
    }

  //   v v v v
  //    q q q
  //   v v v v
  //    q q q
  //   v v v v

  for (xi=1; xi<n-1; xi++)
    for (yi=1; yi<n-1; yi++) 
      norvert(xi,yi) = 0.25*( norquad(xi-1,yi-1)+norquad(xi,yi-1)+norquad(xi,yi)+norquad(xi-1,yi) );
      
  for (xi=1; xi<n-1; xi++) {
    norvert(xi,0) = 0.5*( norquad(xi-1,0)+norquad(xi,0) );
    norvert(xi,n-1) = 0.5*( norquad(xi-1,n-2)+norquad(xi,n-2) );
  }

  for (yi=1; yi<n-1; yi++) {
    norvert(0,yi) = 0.5*( norquad(0,yi-1)+norquad(0,yi) );
    norvert(n-1,yi) = 0.5*( norquad(n-2,yi-1)+norquad(n-2,yi) );
  }

  norvert(0,0) = norquad(0,0);
  norvert(n-1,0) = norquad(n-2,0);
  norvert(n-1,n-1) = norquad(n-2,n-2);
  norvert(0,n-1) = norquad(0,n-2);

  //----------------------------

  theFunction = glGenLists (1);
  glNewList(theFunction, GL_COMPILE);
    glRotatef(olhoM,0,0,1);
    solidFunction();
  glEndList();
  vertices.create(0,0);
  norvert.create(0,0);
  norquad.create(0,0);
  //----------------------------

  glClearColor(1.0, 1.0, 1.0, 1.0); 
  glShadeModel(GL_SMOOTH);
  
  GLfloat light_position[] = { 5.0, 5.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);

  //---------------------------------
  glPolygonMode(GL_FRONT, GL_FILL);
  float red=1; float gre=0.6; float blu=0.6;
  GLfloat mat_ambient[] = { red/2, gre/2, blu/2, 1.0 };
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  GLfloat mat_diffuse[] = { red, gre, blu, 1.0 };
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
  GLfloat mat_specular[] = { 0.3, 0.3, 0.3, 1.0 };
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  GLfloat mat_shininess[] = { 10.0 };
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
}

void reshape(int w, int h)
{ glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, (GLfloat) w/(GLfloat) h, 3, 50); 
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0, 7, 15, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
  glRotatef(-90,1,0,0);
}

void keyboard(unsigned char key, int x, int y)
{
  if (key==27) exit(0);
}

int main(int argc, char** argv)
{ glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);
  glutInitWindowSize(640+320, 480+240); 
  glutInitWindowPosition(0, 0);
  glutCreateWindow(argv[0]);
  init();
  glutDisplayFunc(display); 
  glutReshapeFunc(reshape); 
  glutTimerFunc(40,spinDisplay,0);
  glutKeyboardFunc(keyboard);
  glutMainLoop();
  glutMainLoop();
  return 0;   
}

