#include "bib.h"

//Estou levando em conta as variaveis:
//_WIN64, __arm__ e __unix__

int main(int argc, char** argv) {
  const char* keys =
    " curto| longo    |default| explicacao\n"
    "  -cv | -ocv     | false | linka OpenCV\n"
    "  -v2 | -ver2    | true  | linka OpenCV2 CPU/GPU (default)\n"
    "  -v3 | -ver3    | false | linka OpenCV3 CPU/GPU\n"
    "  -c  | -cek     | false | linka Cekeikon e OpenCV\n"
    #if !defined(__arm__)
    "  -t  | -tiny    | false | linka Tiny-dnn. Liga pthread, AVX e tbb.\n"
    #endif
    #if defined(__arm__)
    "  -r  | -raspicam| false | linka raspicam\n"
    "  -w  | -wiringpi| false | linka wiringPi e pthread\n"
    #endif
    #if defined(__WIN64)
    "  -p  | -pro     | false | linka Proeikon\n"
    "  -pv | -procv   | false | linka Proeikon e OpenCV\n"
    "  -cl | -ocl     | false | linka OpenCL\n"
    #endif
    #if defined(__unix__)
    "  -dl | -dlib    | false | linka dlib\n"
    "  -ca | -caffe   | false | linka caffe\n"
    #endif
    "  -gl | -ogl     | false | linka Glut/OpenGL\n"
    "  -mp | -omp     | false | linka OpenMP\n"
    "  -pt | -pthread | false | Posix Threads\n"
    "  -tb | -tbb     | false | Linka Intel's TBB\n"
    "  -f  | -fltk    | false | linka FLTK\n"
    #if defined(__WIN64)
    "  -w  | -win     | false | modo windows\n"
    "  -re | -res     | false | modo resource (liga windows)\n"
    #endif
    "  -de | -debug   | false | modo debug gdb\n"
    "  -pe | -perm    | false | permissive\n"
    "  -ss | -sse     | false | instrucoes sse2\n"
    "  -av | -avx     | false | instrucoes avx\n"
    "  -an | -ansi    | false | ansi c++14 (default=gnu++14)\n"
    "  -W  | -warn    | false | -Wall\n"
    "  -e  | -extra   |       | extra compiler parameters\n"
    "    Ex: compila prog -e=\"-lm\"\n"
    "  -st | -stack   |       | tamanho do stack\n"
    "    Ex: compila prog -st=1000000\n"
    "  -v  | -verbose | false | expande pkg-config\n";

  if (argc<2) {
    printf("Compila: Compila programa .c ou .cc ou .cxx ou .cpp (default)\n");
    printf("  v1.14 (agosto 2018). Hae Yong Kim.\n");
    printf("Compila arq.cpp [opcoes]\n");
    #if defined(__arm__)
      printf("Processador ARM/Raspberry: __arm__\n");
    #elif defined(__WIN64)
      printf("Sistema Windows 64 bits: __WIN64\n");
    #elif defined(__unix__)
      printf("Sistema Linux: __unix__\n");
    #endif
    printf("%s",keys);
    erro("Erro: Numero de argumentos invalido");
  }

  ArgComando cmd(argc,argv);
  string nomearq=cmd.getCommand(0);
  if (nomearq=="") erro("Erro: Nao especificou arquivo a compilar");
  if (sufixo(nomearq)=="") {
    if (existeArq(nomearq+".cpp")) nomearq+=".cpp";
    else if (existeArq(nomearq+".cc")) nomearq+=".cc";
    else if (existeArq(nomearq+".cxx")) nomearq+=".cxx";
    else if (existeArq(nomearq+".c")) nomearq+=".c";
  }
  if (!existeArq(nomearq)) erro("Erro: Arquivo inexistente ",nomearq);

  bool ocv=false;
  bool ver2=false;
  bool ver3=false;
  bool cek=false;
  bool tiny=false;
  #if defined(__arm__)
  bool raspicam=false;
  bool wiringpi=false;
  #endif
  bool pro=false;
  bool procv=false;
  bool ocl=false;
  #if defined(__unix__)
  bool dlib=false;
  bool caffe=false;
  #endif
  bool ogl=false;
  bool omp=false;
  bool pthread=false;
  bool tbb=false;
  bool fltk=false;
  bool win=false;
  bool res=false;
  bool debug=false;
  bool perm=false;
  bool sse=false;
  bool avx=false;
  bool ansi=false;
  bool warn=false;

  // opcoes que ligam/desligam outras opcoes:
  cek=cmd.getBool("-c","-cek",cek);
  if (cek) ocv=true;

  ver2=cmd.getBool("-v2","-ver2",ver2);
  ver3=cmd.getBool("-v3","-ver3",ver3);
  int qtde=0;
  if (ver2) qtde++;
  if (ver3) qtde++;
  if (qtde==0) { ver2=true; } 
  if (qtde>=2) erro("Erro: So pode escolher uma opcao: ver2 ou ver3");

  #if !defined(__arm__)
    tiny=cmd.getBool("-t","-tiny",tiny);
  #endif
  if (tiny) { pthread=true; avx=true; sse=false; tbb=true; }

  #if defined(__arm__)
    wiringpi=cmd.getBool("-w","-wiringpi",wiringpi);
    if (wiringpi) pthread=true;
  #endif
  
  #if defined(__unix__)
    dlib=cmd.getBool("-dl","-dlib",dlib);
    caffe=cmd.getBool("-ca","-caffe",caffe);
    if (caffe) pthread=true;
  #endif

  procv=cmd.getBool("-pv","-procv",procv);
  if (procv) { pro=true; ocv=true; }

  #if defined(__WIN64)
  res=cmd.getBool("-re","-res",res);
  if (res) win=true;
  #endif

  avx=cmd.getBool("-av","-avx",avx);
  if (avx) sse=false;

  // opcoes que não afetam outras opcoes
  ocv=cmd.getBool("-cv","-ocv",ocv);
  #if defined(__arm__)
    raspicam=cmd.getBool("-r","-raspicam",raspicam);
  #endif
  #if defined(__WIN64)
  pro=cmd.getBool("-p","-pro",pro);
  ocl=cmd.getBool("-cl","-ocl",ocl);
  #endif
  ogl=cmd.getBool("-gl","-ogl",ogl);
  omp=cmd.getBool("-mp","-omp",omp);
  pthread=cmd.getBool("-pt","-pthread",pthread);
  tbb=cmd.getBool("-tb","-tbb",tbb);
  fltk=cmd.getBool("-f","-fltk",fltk);
  #if defined(__WIN64)
  win=cmd.getBool("-w","-win",win);
  #endif
  debug=cmd.getBool("-de","-debug",debug);
  perm=cmd.getBool("-pe","-perm",perm);
  sse=cmd.getBool("-ss","-sse",sse);
  ansi=cmd.getBool("-an","-ansi",ansi);
  warn=cmd.getBool("-W","-warn",warn);

  //variaveis de ambiente
  string CEK5_PATH=getEnv("CEK5_PATH");
  if (CEK5_PATH=="") erro("Erro: CEK5_PATH indefinido");

  #if defined(__unix__)
  string HOME=getEnv("HOME");
  if (HOME=="") erro("Erro: HOME indefinido");
  string CAFFE_ROOT=getEnv("CAFFE_ROOT");
  if (CAFFE_ROOT=="") CAFFE_ROOT=HOME+"/caffe";
  #endif
  
  string CC=getEnv("CC");
  if (CC=="") CC="gcc";
  string CFLAGS=getEnv("CFLAGS");
  string CXX=getEnv("CXX");
  if (CXX=="") CXX="g++";
  string CXXFLAGS=getEnv("CXXFLAGS");

  //sufixo
  string suf=sufixo(nomearq);
  string linha;
  if (suf=="c") {
    linha=CC+" "+CFLAGS;
  } else if (suf=="cc" || suf=="cpp" || suf=="cxx") {
    if (!ansi) linha=CXX+" "+CXXFLAGS+" -std=gnu++14";
    else linha=CXX+" "+CXXFLAGS+" -std=c++14";
  } else erro("Erro: Sufixo deve ser .c .cc .cpp .cxx ou nulo");
  #if defined(__WIN64)
    linha = linha + " " + nomearq + " -o " + semSufixo(nomearq) + ".exe" + " -fmax-errors=2";
  #else
    linha = linha + " " + nomearq + " -o " + semSufixo(nomearq) + " -fmax-errors=2";
  #endif

  if (ocv) {
    #if defined(__WIN64)
      string OPENCV_BIN;
      if (ver3) {
        OPENCV_BIN=CEK5_PATH+"\\opencv\\build\\mingw64\\bin";
      } else {
        OPENCV_BIN=CEK5_PATH+"\\opencv2410\\build\\mingw64\\bin";
      }
      vector<string> vs;
      vsWildCard(OPENCV_BIN+"\\libopencv_*.dll",vs);
      for (unsigned i=0; i<vs.size(); i++) {
        string st=semDiret(vs[i]).erase(0,3);
        linha = linha + " -l" + st;
      }
      if (ver3) {
        linha = linha + " -L" + CEK5_PATH +"\\opencv\\build\\mingw64\\lib";
        linha = linha + " -I" + CEK5_PATH +"\\opencv\\build\\include";
      } else {
        linha = linha + " -L" + CEK5_PATH +"\\opencv2410\\build\\mingw64\\lib";
        linha = linha + " -I" + CEK5_PATH +"\\opencv2410\\build\\include";
      }
    #elif defined(__unix__) && !defined(__arm__)
      string CEK5_MODE=string(getenv("CEK5_MODE"));
      if (CEK5_MODE=="") erro("Erro: Variavel de ambiente CEK5_MODE indefinido");
      if (ver2) {
        linha = linha +
          " -L" + CEK5_PATH + "/opencv2"+CEK5_MODE+"/lib"
          " -I" + CEK5_PATH + "/opencv2"+CEK5_MODE+"/include"
          " `pkg-config " + CEK5_PATH + "/opencv2"+CEK5_MODE+"/lib/pkgconfig/opencv.pc"
          " --libs --cflags`";
      } else if (ver3) {
        linha = linha +
          " -L" + CEK5_PATH + "/opencv3"+CEK5_MODE+"/lib"
          " -I" + CEK5_PATH + "/opencv3"+CEK5_MODE+"/include"
          " `pkg-config " + CEK5_PATH + "/opencv3"+CEK5_MODE+"/lib/pkgconfig/opencv.pc"
          " --libs --cflags`";
      }
    #elif defined(__unix__) && defined(__arm__)
      if (ver2) {
        linha = linha +
          " `pkg-config opencv --libs --cflags`";
      } else erro("Erro: No Raspberry, so esta definida ver2");
    #endif
  }

  if (cek) {
    #if defined(__WIN64)
      if (ver3) {
        linha = linha + " -lcekeikon";
        linha = linha + " -L" + CEK5_PATH + "\\cekeikon\\bin";
        linha = linha + " -L" + CEK5_PATH + "\\cekeikon\\projs\\bin";
        linha = linha + " -I" + CEK5_PATH + "\\cekeikon\\src";
      } else {
        linha = linha + " -lcekeikon41";
        linha = linha + " -L" + CEK5_PATH + "\\cekeikon\\bin";
        linha = linha + " -L" + CEK5_PATH + "\\cekeikon\\projs\\bin";
        linha = linha + " -I" + CEK5_PATH + "\\cekeikon\\src";
      }
    #elif defined(__unix__) && !defined(__arm__)
      if (ver2) {
        linha = linha + " -lcekeikon2";
        linha = linha + " -L" + CEK5_PATH + "/cekeikon/bin";
        linha = linha + " -I" + CEK5_PATH + "/cekeikon/src";
      } else if (ver3) {
        linha = linha + " -lcekeikon3";
        linha = linha + " -L" + CEK5_PATH + "/cekeikon/bin";
        linha = linha + " -I" + CEK5_PATH + "/cekeikon/src";
      }
    #elif defined(__unix__) && defined(__arm__)
      if (ver2) {
        linha = linha + " -lcekeikon";
        linha = linha + " -L" + CEK5_PATH + "/cekeikon/bin";
        linha = linha + " -I" + CEK5_PATH + "/cekeikon/src";
      } else erro("Erro: No Raspberry, so esta definida ver2");
    #endif
  }

  if (tiny) {
    #if defined(__WIN64)
      linha = linha + " -I" + CEK5_PATH + "\\tiny_dnn";
    #else
      linha = linha + " -I" + CEK5_PATH + "/tiny_dnn";
    #endif
  }

  #if defined(__arm__)
    if (raspicam) {
      linha = linha + " -I/usr/local/include/";
      linha = linha + " -L/opt/vc/lib -lraspicam -lraspicam_cv -lmmal -lmmal_core -lmmal_util";
    }

    if (wiringpi) {
      linha = linha + " -I/usr/local/include/";
      linha = linha + " -lwiringPi";
    }
  #endif

  #if defined(__WIN64)
  if (pro) {
    linha = linha + " -lproeikon";
    linha = linha + " -L" + CEK5_PATH + "\\proeikon\\bin";
    linha = linha + " -I" + CEK5_PATH + "\\proeikon\\lib";
  }

  if (procv) {
    linha = linha + " -limgpv";
    linha = linha + " -I" + CEK5_PATH + "\\proeikon\\pvlib";
  }

  if (ocl) {
    string opencl=getEnv("OpenCL");
    linha = linha +  " -L\"" + opencl + "\\lib\\x64\"" +  " -I\"" + opencl + "\\include\"" + " -lOpenCL";
  }
  #endif
  
  #if defined(__unix__)
  if (dlib) {
    linha = linha + " `pkg-config --libs --cflags dlib-1` -D DLIB_USE_CUDA";
  }
  if (caffe) {
    linha = linha + " -I" + CAFFE_ROOT + "/distribute/include "
                  + " -I/usr/local/cuda-8.0/targets/x86_64-linux/include "
                  + " -L" + CAFFE_ROOT + "/distribute/lib -lcaffe -lglog -lboost_system";
  }
  #endif

  if (ogl) {
    #if defined(__WIN64)
      linha = linha + " -lfreeglut -lopengl32 -lglu32";
      linha = linha + " -I" + CEK5_PATH + "\\cekeikon\\src";
      linha = linha + " -L" + CEK5_PATH + "\\cekeikon\\bin";
    #else
      linha = linha + " -lGL -lGLU -lglut";
    #endif
  }

  if (omp) {
    linha = linha + " -fopenmp";
  }

  if (pthread) {
    linha = linha + " -pthread";
  }

  if (tbb) {
    #if defined(__WIN64)
      linha = linha + " " + CEK5_PATH +"\\tbb\\build\\windows_intel64_gcc_mingw6.2.0_release\\tbb.dll";
      linha = linha + " -I" + CEK5_PATH + "\\tbb\\include";
    #else
      linha = linha + " -ltbb";
    #endif
  }

  if (fltk) {
    linha = linha + " -lfltk -lfltk_forms -lfltk_gl -lfltk_images -Wno-int-to-pointer-cast";
    #if defined(__WIN64)
    linha = linha + " -I" + CEK5_PATH + "\\fltk\\include";
    linha = linha + " -L" + CEK5_PATH + "\\fltk\\lib";
    #endif
  }

  #if defined(__WIN64)
  if (win) {
    linha = linha + " -mwindows -lwinmm";
  }

  if (res) {
    linha = linha + " " + semSufixo(nomearq) + ".res.o";
  }
  #endif

  if (debug) {
    linha = linha + " -O0 -g";
  } else {
    linha = linha + " -O3 -s";
  }

  if (perm && suf!="c")
    linha = linha + " -fpermissive";

  if (sse) {
    linha = linha + " -msse2";
  }

  if (avx) {
    linha = linha + " -mavx";
  }

  if (warn) {
    linha = linha + " -Wall";
  }

  string extra=cmd.getString("-e","-extra");
  if (extra!="") linha = linha + " " + extra;

  int tstack=cmd.getInt("-st","-stack");
  if (tstack!=INT_MIN) linha = linha + " -Wl,-stack=" + to_string(tstack);

  bool verbose=cmd.getBool("-v","-verbose",false);

  cmd.leuTodos();

  if (verbose) {
    string st="echo "+linha;
    system(st.c_str());
  } else {
    cout << linha << endl;
  }
  system(linha.c_str());
}
