#include "bib.h"

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<< EdVar <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
string ExpandPath(const string& adir ) {
// Expand occurrences of %X% in the path name with the relevant environment
// variable setting.
  string rv, envname;
  bool inenv = false;
  unsigned int i = 0;
  while( i < adir.size() ) {
    char c = adir[i++];
    if ( c == '%' && inenv ) {
      rv += getEnv( envname );
      inenv = false;
      envname = "";
    } else if ( c == '%' && ! inenv ) {
      envname = "";
      inenv = true;
    } else if ( inenv ) {
      envname += c;
    } else {
      rv += c;
    }
  }
  if ( envname != "" ) {
    rv += getEnv( envname );
  }
  return rv;
}

void NotifyChanges() {
// Notify system that things have changed
  const char * what = "Environment";
  #if __x86_64__ || __ppc64__ //64bits
    long long unsigned int rv;
  #else //32bits
    DWORD rv;
  #endif
  SendMessageTimeout( HWND_BROADCAST, WM_SETTINGCHANGE, 0,
              (LPARAM) what, SMTO_ABORTIFHUNG, 5000, & rv );
}

class RegPath {
  public:
    bool existeDir(const string& st); //Existe o diretorio no computador
    bool dirIsInVar(const string& adir); //Diretorio ja esta no evar
    bool variavelExiste=true;
    RegPath(string _envName="", HKEY _root=HKEY_CURRENT_USER, 
            REGSAM _regsam=KEY_ALL_ACCESS, bool _check=false);
    ~RegPath();

    void deleteVar();
    void setReg(const string& st);
    void updateReg();
    void addTail(const string& adir);
    void addHead(const string& adir);
    void delReg(const string& adir);
    void imp(bool semicolon=false);
    void splitPath(const string& st);
    
    string envName;
    HKEY pathKey, root;
    bool check;
    vector<string> mPath;
};

bool RegPath::existeDir(const string& st) { //Existe o diretorio no computador
  DWORD attr = GetFileAttributes( st.c_str() );
  if ( attr==INVALID_FILE_ATTRIBUTES || !(attr & FILE_ATTRIBUTE_DIRECTORY ) ) return false;
  else return true;
}

bool RegPath::dirIsInVar(const string& adir) { //Diretorio ja esta no evar
  bool existe=false;
  for (unsigned i=0; i<mPath.size(); i++) {
    if (maiuscula(adir)==maiuscula(mPath[i])) {
      existe=true;
      break;
    }
  }
  return existe;
}

RegPath::RegPath(string _envName, HKEY _root, REGSAM _regsam, bool _check): 
   envName(_envName), pathKey(0), root(_root), check(_check) {
// Create a RegPath object from a registry key, which must be either
// HKEY_CURRENT_USER or HKEY_LOCAL_MACHINE.
// _regsam must be either KEY_ALL_ACCESS or KEY_READ
  if (_envName!="") {
    long res=0;
    if (root==HKEY_CURRENT_USER) 
      res = RegOpenKeyEx(root, "Environment", 0, _regsam, & pathKey);
    else if (root == HKEY_LOCAL_MACHINE)
      res = RegOpenKeyEx(root, 
                         "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment",
                         0, _regsam, &pathKey);
    else erro("Error: Only HKEY_CURRENT_USER or HKEY_LOCAL_MACHINE");

    if (res != ERROR_SUCCESS) erro("Error: Could not get registry key.");

    const int BUFSIZE = 2048;
    BYTE buffer[ BUFSIZE + 1 ];
    DWORD bufflen = BUFSIZE;
    DWORD type = 0;
    res = RegQueryValueEx( pathKey, envName.c_str(), NULL, &type, buffer, &bufflen   );
    if ( res != ERROR_SUCCESS ) {
      variavelExiste=false; 
      buffer[bufflen] = 0;
    } else {
      splitPath( (string)(const char *)buffer );
    }
  }
}

RegPath::~RegPath() {
// Close the path key opened in ctor.
  if (pathKey) RegCloseKey(pathKey);
}

void RegPath::deleteVar() { 
  long res = RegDeleteValue( pathKey, envName.c_str());
  if (res!=ERROR_SUCCESS) erro("Error: Could not delete registry.");
  NotifyChanges();
}

void RegPath::setReg(const string& st) {
  // Atribui st a VAR. Se nao existe VAR, cria.
  // Se st for "", apaga VAR se existe.
  if (st=="" && variavelExiste) deleteVar();
  //else if (st=="" && !variavelExiste) nao faz nada
  else if (st!="") { //variavel pode existir ou nao
    long res = RegSetValueEx( pathKey, envName.c_str(), 0, REG_EXPAND_SZ,
                  (BYTE *)st.c_str(), st.size() + 1 );
    if (res!=ERROR_SUCCESS) erro("Error: Could not update registry.");
    NotifyChanges();
  }
}

void RegPath::updateReg() {
  string newpath;
  for ( unsigned int i = 0; i < mPath.size(); i++ ) 
    newpath += mPath[i] + ";";
  setReg(newpath);// chama NotifyChanges
}

void RegPath::addTail(const string& adir) {
  if (check && !existeDir(adir)) erro("Error: Does not exist "+adir);
  if ( !dirIsInVar(adir) ) { // o diretorio nao esta no path
    if (variavelExiste) { // variavel ja existe - acrescentar
      mPath.push_back(adir);
      updateReg();
    } else { // variavel nao existe - criar - acho que nao precisa deste else
      setReg(adir);
    }
  }
}

void RegPath::addHead(const string& adir) {
  if (check && !existeDir(adir)) erro("Warning: Does not exist "+adir);
  if (!dirIsInVar(adir)) { // o diretorio nao esta no path
    if (variavelExiste) { // variavel ja existe - acrescentar
      mPath.insert(mPath.begin(),adir);
      updateReg();
    } else { // variavel nao existe - criar - acho que nao precisa deste else
      setReg(adir);
    }
  }
}

void RegPath::delReg(const string& adir) {
// Find string in path, ignoring case
  bool removi=false;
  for (auto it=mPath.begin(); it!=mPath.end(); ++it) {
    if ( maiuscula(adir)==maiuscula(*it) ) {
      mPath.erase(it);
      removi=true;
      break;
    }
  }
  if (removi) updateReg();
  else {
    if (check) erro("Warning: Could not find directory "+adir); 
  }
}

void RegPath::imp(bool semicolon) {
  if (mPath.size()==0) {
    cout << "Not defined" << endl;
    return;
  }
  if (semicolon) {
    for (unsigned i=0; i<mPath.size(); i++) 
      cout << mPath[i] << ";";
    cout << endl;
  } else {
    for (unsigned i=0; i<mPath.size(); i++) 
      cout << mPath[i] << endl;
  }
}

void RegPath::splitPath(const string& st) {
// Helper to split st at ';' character
  //if (st=="") return;
  string::size_type pos=0;
  while (pos!=string::npos) { //!=-1
    string::size_type fpos=st.find(";", pos);
    string s;
    if (fpos==string::npos) { //nao achou ";"
      s=st.substr(pos); //copia de pos ate fim
      pos=string::npos;
    } else { 
      s=st.substr(pos, fpos-pos);
      pos=fpos+1;
    } 
    if (s.find_first_not_of( " \t" )!=string::npos) {
      mPath.push_back(s);
    }
  }
}

int main(int argc, char** argv) {
  HKEY hkey=0;
  try {
    if (argc<2) {
      printf("EdVar: Edits 'path-like' environment variables in registry.\n");
      printf("  v1.06 (2015). Hae Yong Kim.\n");
      printf("EdVar COMMAND VAR [VALUE] [-m] [-s] [-c] [-v]\n");
      printf("  COMMAND=\n");
      printf("    ListReg VAR: List registry value (USER or MACHINE -m)\n");
      printf("    ListEnv VAR: List environment variable value\n");
      printf("    SetReg VAR VALUE: Set registry VAR to VALUE\n");
      printf("    DelReg VAR: Delete registry VAR\n");
      printf("    AddTail VAR DIR: Add DIR to tail of registry VAR.\n");
      printf("      DIR may not exist. Use -c to test if DIR exists.\n");
      printf("      Use !DIR! to insert %%DIR%%\n");
      printf("    AddHead VAR DIR: Add DIR to head of registry VAR.\n");
      printf("      AddTail and AddHead does not insert repeated dir.\n");
      printf("    DelDir VAR DIR: Delete DIR from registry VAR.\n");
      printf("  VAR=PATH, LIBRARY_PATH, C_INCLUDE_PATH,\n");
      printf("      CPLUS_INCLUDE_PATH, PKG_CONFIG_PATH, etc.\n");
      printf("  OPTIONS=\n"); 
      printf("    -m uses HKEY_LOCAL_MACHINE instead of HKEY_CURRENT_USER\n");
      printf("       requires administor privileges to change value\n");
      printf("    -s output separated by semicolon instead of newline\n");
      printf("    -c check if DIR exists. Insert DIR only if DIR exists\n");
      printf("    -v verbose\n");
      printf("Ex: EdVar ListReg PATH (user registry)\n");
      printf("    EdVar ListReg PATH -m (machine registry)\n");
      printf("    EdVar ListEnv PATH\n");
      printf("    EdVar AddTail PATH C:\\DIRET\n");
      printf("    EdVar AddHead TEST !C_INCLUDE_PATH! (insert %%C_INCLUDE_PATH%%)\n");
      return 1;
    }    
    
    ArgComando comando(argc,argv);

    bool machine=comando.getBool("-m","-machine");
    if (machine) hkey=HKEY_LOCAL_MACHINE;
    else hkey=HKEY_CURRENT_USER;
    
    bool semicolon=comando.getBool("-s","-semicolon");

    bool check=comando.getBool("-c","-check");

    bool verbose=comando.getBool("-v","-verbose");

    string command=maiuscula(comando.getCommand(0));
    string varName=comando.getCommand(1);

    if (command=="LISTREG") {
      if (verbose) cout << "ListReg " << varName << endl;
      RegPath reg(varName,hkey,KEY_READ,check);
      reg.imp(semicolon);      
    } else if (command=="LISTENV") {
      if (verbose) cout << "ListEnv " << varName << endl;
      RegPath reg("");
      string st=getEnv(varName);
      reg.splitPath(st);
      reg.imp(semicolon); 
    } else if (command=="SETREG") {
      RegPath reg(varName,hkey,KEY_ALL_ACCESS,check);
      string value=comando.getCommand(2);
      if (verbose) cout << "SetReg " << varName << " " << value << endl;
      reg.setReg(value);
    } else if (command=="DELREG") {
      RegPath reg(varName,hkey,KEY_ALL_ACCESS,check);
      if (verbose) cout << "DelReg " << varName << endl;
      reg.deleteVar();
    } else if (command=="ADDTAIL") {
      RegPath reg(varName,hkey,KEY_ALL_ACCESS,check);
      string value=comando.getCommand(2);
      if (verbose) cout << "AddTail " << varName << " " << value << endl;
      reg.addTail(value);
    } else if (command=="ADDHEAD") {
      RegPath reg(varName,hkey,KEY_ALL_ACCESS,check);
      string value=comando.getCommand(2);
      if (verbose) cout << "AddHead " << varName << " " << value << endl;
      reg.addHead(value);
    } else if (command=="DELDIR") {
      RegPath reg(varName,hkey,KEY_ALL_ACCESS,check);
      string value=comando.getCommand(2);
      if (verbose) cout << "DelDir " << varName << " " << value << endl;
      reg.delReg(value);
    } else erro("Erro: Comando inexistente ",command);

    comando.leuTodos();
    
  } catch (string st) {
    if ( (st=="Error: Could not get registry key." || st=="Error: Could not update registry.") 
          && hkey==HKEY_LOCAL_MACHINE ) 
      cout << "You must have administrator privileges to set machine registry.\n";
    cout << st << endl;
  } catch (...) {
    cout << "Unknown error" << endl;
  }
}

